/*
*  barremdlgs.cpp 
*
*  Copyright (C) 2010 Richard Christman 
*
*  This file is part of QuickSilver Lite.
*
*  QuickSilver Lite is free software; you can redistribute it and/or modify 
*  it under the terms of the GNU General Public License as published by the 
*  Free Software Foundation; either version 3 of the License, or (at your 
*  option) any later version.
* 
*  This program is distributed in the hope that it will be useful, but 
*  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
*  or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License 
*  for more details.
* 
*  You should have received a copy of the GNU General Public License along 
*  with this program; if not, see <http://www.gnu.org/licenses>.
* 
*  Additional permission under GNU GPL version 3 section 7
* 
*  If you modify QuickSilver Lite, or any covered work, by linking or 
*  combining it with CodeGear VCL (or a modified version of that library), 
*  containing parts covered by the terms of CodeGear's license, the licensors 
*  of QuickSilver Lite grant you additional permission to convey the 
*  resulting work.
*
*/

//---------------------------------------------------------------------------
#include <vcl.h>
#pragma hdrstop

#include "barremdlgs.h"
#include "mixmaster.h"
#include "utility.h"
#include "global.h"
#include "help.h"
#include "qsdbs.h"
//---------------------------------------------------------------------------
#pragma package(smart_init)
#pragma resource "*.dfm"
//TBarRemDlg *BarRemDlg;
//---------------------------------------------------------------------------

bool __fastcall ManageBarredRemailers(TForm *Owner)
{
   TBarRemDlg *barremdlg = new TBarRemDlg(Owner);
   int result = barremdlg->ShowModal();
   delete barremdlg;
   return result == IDOK;
}
//---------------------------------------------------------------------------

__fastcall TBarRemDlg::TBarRemDlg(TComponent* Owner)
        : TForm(Owner)
{
   PopupParent = (TForm*)Owner;
   TIniFile *ini = new TIniFile(AppIni());
   LoadWindowPosition(this, "barremdlg", ini);
   orgwin = IntToStr(Width)+IntToStr(Height);
   AssignFont(Font, ini->ReadString("settings", "dialogfont", OptsDefDialogFont));
   AssignFont(RemList->Font, ini->ReadString("settings", "windowfont", OptsDefWindowFont));
   delete ini;

   bars = new TStringList();
   if (FileExists(MixStarexFile()))
   {
      try
      {
         bars->LoadFromFile(MixStarexFile());
      }
      catch(Exception &e)
      {
         MsgBox(this,
            "Sorry, I failed to open the file, " + mixstarextxt + "! \n" +
             e.Message,
             MT_ERROR, MB_OK
         );
      }
   }
   orgbars = bars->CommaText;
   last = new TStringList();
   DB->LoadBars(BarLast, last);
   orglast = last->CommaText;
   direct = new TStringList();
   DB->LoadBars(BarDirect, direct);
   orgdirect = direct->CommaText;
   rems = new TStringList();
   rems->Sorted = true;
   if (FileExists(MixPubringRawFile()))
   {
      TStringList *ring = new TStringList();
      try
      {
         ring->LoadFromFile(MixPubringRawFile());
      }
      catch (Exception &e)
      {
         MsgBox(this,
            "Sorry, I failed to open the file " + mixpubringraw + "!   \n" +
             e.Message,
             MT_ERROR, MB_OK
         );
      }

      for (int i = 0; i < ring->Count; i++)
      {
         AnsiString line = ring->Strings[i];
         if (line.Pos("@") == 0)
            continue;

         bool mid = GetTok(4, line, ' ').Pos("M") > 0;
         rems->AddObject(GetTok(0, line, ' '), (TObject*)mid);
      }
      delete ring;
      ChainsBoxClick(ChainsBox);
   }
   else
   {
      MsgBox(this,
         "The file '" + mixpubringraw + "' is missing from my data directory. You'll   \n"
         "need to update and save stats before continuing here.",
         MT_INFO, MB_OK);
      Close();
      return;
   }
   //Timer1->Enabled = true;
}
//---------------------------------------------------------------------------

__fastcall TBarRemDlg::~TBarRemDlg()
{
   if (IntToStr(Width)+IntToStr(Height) != orgwin)
   {
      TIniFile *ini = new TIniFile(AppIni());
      SaveWindowPosition(this, "barremdlg", ini);
      delete ini;
   }
   delete rems;
   delete direct;
   delete last;
   delete bars;
}
//---------------------------------------------------------------------------

/*void __fastcall TBarRemDlg::LoadRemailers()
{
   TStringList *rems = new TStringList();
   rems->Sorted = true;
   if (FileExists(MixPubringMixFile()))
   {
      try
      {
         rems->LoadFromFile(MixPubringMixFile());
      }
      catch (Exception &e)
      {
         MsgBox(this,
            "Sorry, I failed to open the barred remailers file, " +
             mixstarextxt + "! \n" + e.Message,
             MT_ERROR, MB_OK
         );
      }

      for (int i = 0; i < rems->Count; i++)
      {
         if (rems->Strings[i].Pos("@") == 0)
            continue;

         AnsiString rem = GetTok(0, rems->Strings[i], ' ');

         int idx = RemList->Items->Add(rem);
         if (bars->IndexOf(rem) >= 0)
            RemList->Checked[idx] = true;
      }

      for (int i = 0; i < bars->Count; i++)
      {
         AnsiString bar = bars->Strings[i];
         if (RemList->Items->IndexOf(bar) < 0)
         {
            int idx = RemList->Items->Add(bar);
            RemList->Checked[idx] = true;
         }
      }
   }
   delete rems;
}*/
//---------------------------------------------------------------------------

void __fastcall TBarRemDlg::FormCloseQuery(TObject *Sender, bool &CanClose)
{
   GatherBarred();
   bool modified = bars->CommaText != orgbars || last->CommaText != orglast ||
      direct->CommaText != orgdirect;

   if (modified) // list is modified
   {
      if (ModalResult == mrCancel)
      {
         switch (MsgBox(this,
            "Your barred remailer list is modified. Save your changes? ",
             MT_CONFIRM, MB_YESNOCANCEL))
          {
             case IDCANCEL:
                CanClose = false;
             case IDNO:
                return;
          }
      }

      if (bars->CommaText != orgbars)
      {
         try
         {
            bars->SaveToFile(MixStarexFile());
         }
         catch(Exception &e)
         {
            MsgBox(this,
               "Sorry, I failed to save the barred remailers file, " +
                mixstarextxt + "! \n" + e.Message,
                MT_ERROR, MB_OK
            );
         }
      }
      if (last->CommaText != orglast)
         DB->SaveBars(BarLast, last);
      if (direct->CommaText != orgdirect)
         DB->SaveBars(BarDirect, direct);

      MsgBox(this,
         "Last remailer bars take effect next time you update pubring.mix.   "
         "All others are good to go. ",
          MT_INFO, MB_OK);
   }
}
//---------------------------------------------------------------------------

void __fastcall TBarRemDlg::Resizedialog1Click(TObject *Sender)
{
   BorderStyle = bsSizeable;
}
//---------------------------------------------------------------------------

void __fastcall TBarRemDlg::Locksize1Click(TObject *Sender)
{
   BorderStyle = bsDialog;
}
//---------------------------------------------------------------------------

void __fastcall TBarRemDlg::PopupMenuPopup(TObject *Sender)
{
   Resizedialog1->Enabled = BorderStyle == bsDialog;
   Locksize1->Enabled = BorderStyle == bsSizeable;
}
//---------------------------------------------------------------------------

void __fastcall TBarRemDlg::HelpBtnClick(TObject *Sender)
{
   DisplayHtmlHelpTopic(this, "::/html/barredrems.htm>barred");
}
//---------------------------------------------------------------------------

void __fastcall TBarRemDlg::GatherBarred()
{
   TStrings *list = NULL;
   switch (mode)
   {
      case brDirect: list = direct; break;
      case brLast: list = last; break;
      case brChains: list = bars; break;
   }
   if (!list)
      return;

   list->Clear();
   for (int i = 0; i < RemList->Count; i++)
   {
      if (RemList->Checked[i] && RemList->ItemEnabled[i])
      {
         list->Add(RemList->Items->Strings[i]);
      }
   }
}
//---------------------------------------------------------------------------

void __fastcall TBarRemDlg::ChainsBoxClick(TObject *Sender)
{
   if (mode == brChains)
      return;
   RemList->Items->BeginUpdate();
   GatherBarred();
   mode = brChains;
   RemList->Clear();
   Label1->Caption =
      "Tick remailers you want barred completely from random chains";
   for (int i = 0; i < rems->Count; i++)
   {
      AnsiString rem = rems->Strings[i];
      int x = RemList->Items->Add(rem);
      if (bars->IndexOf(rem) >= 0)
         RemList->State[x] = cbChecked; //Checked[x];
   }
   // any bars not in rems ?
   for (int i = 0; i < bars->Count; i++)
   {
      AnsiString bar = bars->Strings[i];
      if (rems->IndexOf(bar) < 0)
      {
         int x = RemList->Items->Add(bar);
         RemList->State[x] = cbChecked;
      }
   }
   RemList->Items->EndUpdate();
}
//---------------------------------------------------------------------------

void __fastcall TBarRemDlg::LastBoxClick(TObject *Sender)
{
   if (mode == brLast)
      return;
   RemList->Items->BeginUpdate();
   GatherBarred();
   mode = brLast;
   RemList->Clear();
   Label1->Caption =
      "Tick remailers you want barred from last position in random chains";

   for (int i = 0; i < rems->Count; i++)
   {
      //if (rems->Objects[i])
      //   continue;
      AnsiString rem = rems->Strings[i];
      if (bars->IndexOf(rem) >= 0)
      {
         int x = RemList->Items->Add(rem);
         RemList->Checked[x] = true;
         RemList->ItemEnabled[x] = false;
         continue;
      }

      int x = RemList->Items->Add(rem);
      if (last->IndexOf(rem) >= 0)
         RemList->State[x] = cbChecked;
      if (rems->Objects[i])
         RemList->ItemEnabled[x] = false;
   }
   RemList->Items->EndUpdate();
}
//---------------------------------------------------------------------------

void __fastcall TBarRemDlg::DirectBoxClick(TObject *Sender)
{
   if (mode == brDirect)
      return;
   RemList->Items->BeginUpdate();
   GatherBarred();
   mode = brDirect;
   RemList->Clear();
   Label1->Caption =
      "Tick remailers you don't want messages sent directly to, see Help";

   for (int i = 0; i < rems->Count; i++)
   {
      AnsiString rem = rems->Strings[i];
      if (bars->IndexOf(rem) >= 0)
      {
         int x = RemList->Items->Add(rem);
         //RemList->State[x] = cbChecked;
         RemList->ItemEnabled[x] = false;
         continue;
      }

      int x = RemList->Items->Add(rem);
      if (direct->IndexOf(rem) >= 0)
         RemList->State[x] = cbChecked;
   }
   RemList->Items->EndUpdate();
}
//---------------------------------------------------------------------------


