/*
*  clipboard.cpp 
*
*  Copyright (C) 2010 Richard Christman 
*
*  This file is part of QuickSilver Lite.
*
*  QuickSilver Lite is free software; you can redistribute it and/or modify 
*  it under the terms of the GNU General Public License as published by the 
*  Free Software Foundation; either version 3 of the License, or (at your 
*  option) any later version.
* 
*  This program is distributed in the hope that it will be useful, but 
*  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
*  or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License 
*  for more details.
* 
*  You should have received a copy of the GNU General Public License along 
*  with this program; if not, see <http://www.gnu.org/licenses>.
* 
*  Additional permission under GNU GPL version 3 section 7
* 
*  If you modify QuickSilver Lite, or any covered work, by linking or 
*  combining it with CodeGear VCL (or a modified version of that library), 
*  containing parts covered by the terms of CodeGear's license, the licensors 
*  of QuickSilver Lite grant you additional permission to convey the 
*  resulting work.
*
*/

//---------------------------------------------------------------------------
#include <vcl.h>
#pragma hdrstop
#include <clipbrd.hpp>

#include "clipboard.h"
#include "utility.h"
#include "msgmans.h";
#include "follow2postdlgs.h"
#include "qsdbs.h"
#include "composewins.h"
#include "templatedlgs.h"
#include "global.h"
#include "gpg.h"
#ifdef QS2
#include "qs2mainwins.h"
#else
#include "mainwins.h"
#endif
#include "pgpdlgs2.h"
#include "sigfaileddlgs.h"
#include "nymstuff.h"
//---------------------------------------------------------------------------
#pragma package(smart_init)
//---------------------------------------------------------------------------

bool __fastcall ClipboardFollowup(TForm *Owner, TStrings *Followup,
   AnsiString TemplateName)
{
   bool reply = false, follow = true;
   TStringList *article = new TStringList();
   article->Text = Clipboard()->AsText;

   TMsgManager *msg = new TMsgManager(article);
   if (msg->GetHeader("Nym:") != "")
   {
	  MsgBox(Owner,
		 "The clipboard has an incoming NYM message. Use Clipboard/Reply   ",
		  MT_INFO, MB_OK);
	  delete msg;
	  delete article;
	  return false;
   }
   msg->DelHeader("Xref: ");

   if (msg->GetHeader("Followup-To:").Trim().LowerCase() == "poster")
   {
	  if (FollowupToPoster(Owner, reply, follow))
	  {
		 if (reply && !follow)
		 {
			delete msg;
			delete article;
			return ClipboardReply(Owner, Followup);
		 }
	  }
	  else
		 return false;
   }

   if(!msg->HasHeader("Newsgroups:") && !msg->HasHeader("Followup-To:"))
   {
	  MsgBox(Owner,
        "The clipboard doesn't appear to contain a USENET article!    \n\n"
        "Chances are you didn't include the article headers when   \n"
        "you copied it. Please give it another try.",
         MT_INFO, MB_OK);
      delete msg;
      delete article;
      return false;
   }

   if (TemplateName.IsEmpty() &&
      (TemplateName = SelectTemplateFollowup(Owner)) == "")
   {
	  delete msg;
	  delete article;
      return false;
   }

   bool ok = FollowupArticle(Owner, article, TemplateName, reply, Followup);
   delete msg;
   delete article;
   return ok;
}
//---------------------------------------------------------------------------

bool __fastcall needsemptyline(TStrings *Template)
{
   bool needs = true; // in case the template is headers only.
   for (int i = 0; i < Template->Count; i++)
   {
	  if (Template->Strings[i].Trim().IsEmpty())
	  {
		 needs = false;
		 break;
	  }
   }
   return needs;
}
//---------------------------------------------------------------------------

bool __fastcall FollowupArticlePost(TForm *Owner, TStrings *Article,
   TStrings *Template, bool reply2, TStrings *Followup)
{
   AnsiString line, rep, adr, hdr, mID, pststr, blank;
   Followup->Clear();

   TIniFile *ini = new TIniFile(AppIni());
   pststr = ini->ReadString("settings", "quotemarks", "\"> \"");
   DropQuotes(pststr);
   delete ini;

   TMsgManager *tmpart = new TMsgManager(Template);
   TMsgManager *orgart = new TMsgManager(Article);
   TStringList *lines = new TStringList();

   // if followup to poster, delete the header
   if (orgart->GetHeader("Followup-To:").LowerCase().Trim() == "poster")
	  orgart->DelHeader("Followup-To:");

   // if followup-to in this message, drop any newsgroups header and
   // change followup-to to newsgroups and add back to article
   if (orgart->HasHeader("Followup-To: "))
   {
	  hdr = orgart->GetHeader("Followup-To:");
	  int idx = orgart->DelHeader("Newsgroups:");
	  if (idx < 0)
		 idx = orgart->EOH();
	  Article->Insert(idx, "Newsgroups: " + hdr);
   }

   // get special header positions from template if there
   while (!(hdr = tmpart->PopHeader()).IsEmpty())
   {
	  if (hdr.Pos("References:") == 1)
		 hdr = orgart->Get4Reply("References:");
	  else
	  if (hdr.Pos("Post:") == 1)
		 hdr = orgart->Get4Reply("Post:");
	  else
	  if (hdr.Pos("Subject:") == 1 && hdr.Trim().Length() < 10)
		 hdr = orgart->Get4Reply("Subject:");

	  lines->Text = lines->Text + hdr;
   }

   hdr = "";
   // add special headers to end of headers if not handled above
   if (!lines->Text.Pos("References: "))
	  hdr = orgart->Get4Reply("References:");
   if (!lines->Text.Pos("Subject:"))
	  hdr += orgart->Get4Reply("Subject:");
   if (!hdr.IsEmpty())
	  lines->Text = lines->Text + hdr;

   lines->Add("");
   lines->Add("In article " + orgart->GetHeader("Message-ID:"));
   lines->Add(orgart->GetHeader("From:") + " wrote:");

   // remove trailing nulls
   while (Article->Count && Article->Strings[Article->Count - 1].Trim().IsEmpty())
	  Article->Delete(Article->Count - 1);

   int i, eoh = orgart->EOH();
   // add quoted article text to new article
   for (i = eoh; i < Article->Count; i++)
	  lines->Add(pststr + Article->Strings[i]);

   // remove trailing nulls from template
   while (Template->Count && Template->Strings[Template->Count - 1].Trim().IsEmpty())
	  Template->Delete(Template->Count - 1);

   // copy template sig to lines
   for (i = 1; i < Template->Count; i++)
	  lines->Add(Template->Strings[i]);

   if (reply2) // user chooses to reply as well as followup
   {
	  TMsgManager *linman = new TMsgManager(lines);
	  rep = orgart->GetHeader("From:");
	  adr = linman->GetHeader("To:");

	  // make the authors address first in list
	  if (!adr.IsEmpty())
		 rep += ",\r\n ";
	  adr = rep + adr;

	  // add the header back at present location
	  int idx = linman->DelHeader("To:");
	  if (idx < 0)
		 idx = linman->EOH();
	  lines->Insert(idx, "To: " + adr);
	  delete linman;
   }

   //lines->Insert(0, "Followup using: " + TemplateName);
   Followup->AddStrings(lines);

   delete tmpart;
   delete orgart;
   delete lines;
   return true;
}
//---------------------------------------------------------------------------

bool __fastcall FollowupArticle(TForm *Owner, TStrings *Article,
   const AnsiString &TemplateName, bool reply2, TStrings *Followup)
{
   AnsiString line, rep, adr, hdr, mID, pststr, blank;
   Followup->Clear();
   
   TIniFile *ini = new TIniFile(AppIni());
   pststr = ini->ReadString("settings", "quotemarks", "\"> \"");
   DropQuotes(pststr);
   delete ini;

   // get the template and test for problems
   TStringList *Template = new TStringList();
   DB->LoadTemplate(TemplateName, Template);
   // in case the template is empty
   if (!Template->Count)
   {
	  MsgBox(Owner,
		 "Problem! The template you selected is empty! There aren't   \n"
		 "any lines of text at all! You'll need to fix that.   ",
		  MT_INFO, MB_OK);
	  delete Template;
	  return false;
   }
   // in case the template is headers only (no empty line after headers)
   // add an empty line
   if (needsemptyline(Template))
	  Template->Add("");

   TMsgManager *tmpart = new TMsgManager(Template);
   if (tmpart->HasHeader("Post: "))
   {
	  bool ok = FollowupArticlePost(Owner, Article, Template, reply2, Followup);
      delete tmpart;
      delete Template;
      return ok;
   }

   TStringList *lines = new TStringList();
   TMsgManager *orgart = new TMsgManager(Article);

   // if followup to poster, delete the header
   if (orgart->GetHeader("Followup-To:").LowerCase().Trim() == "poster")
	  orgart->DelHeader("Followup-To:");

   // if followup-to in this message, drop any newsgroups header and
   // change followup-to to newsgroups and add back to article
   if (orgart->HasHeader("Followup-To:"))
   {
	  hdr = orgart->GetHeader("Followup-To:");
	  int idx = orgart->DelHeader("Newsgroups:");
	  if (idx < 0)
         idx = orgart->EOH();
	  Article->Insert(idx, "Newsgroups: " + hdr);
   }

   // get special header positions from template if there
   while (!(hdr = tmpart->PopHeader()).IsEmpty())
   {
      if (hdr.Pos("References:") == 1)
		 hdr = orgart->Get4Reply("References:");
	  else
      if (hdr.Pos("Newsgroups:") == 1)
         hdr = orgart->Get4Reply("Newsgroups:");
      else
	  if (hdr.Pos("Subject:") == 1 && hdr.Trim().Length() < 10)
		 hdr = orgart->Get4Reply("Subject: ");

	  lines->Text = lines->Text + hdr;
   }

   hdr = "";
   // add special headers to end of headers if not handled above
   if (!lines->Text.Pos("References:"))
	  hdr = orgart->Get4Reply("References:");
   if (!lines->Text.Pos("Newsgroups:"))
	  hdr += orgart->Get4Reply("Newsgroups:");
   if (!lines->Text.Pos("Subject:"))
	  hdr += orgart->Get4Reply("Subject:");
   if (!hdr.IsEmpty())
	  lines->Text = lines->Text + hdr;

   lines->Add("");
   lines->Add("In article " + orgart->GetHeader("Message-ID:"));
   lines->Add(orgart->GetHeader("From:") + " wrote:");

   // remove trailing nulls
   while (Article->Count && Article->Strings[Article->Count - 1].Trim().IsEmpty())
	  Article->Delete(Article->Count - 1);

   int i, eoh = orgart->EOH();
   // add quoted article text to new article
   for (i = eoh; i < Article->Count; i++)
	  lines->Add(pststr + Article->Strings[i]);

   // remove trailing nulls from template
   while (Template->Count && Template->Strings[Template->Count - 1].Trim().IsEmpty())
	  Template->Delete(Template->Count - 1);

   // copy template sig to lines
   for (i = 1; i < Template->Count; i++)
	  lines->Add(Template->Strings[i]);

   if (reply2) // user chooses to reply as well as followup
   {
	  TMsgManager *linman = new TMsgManager(lines);
	  rep = orgart->GetHeader("From:");
	  adr = linman->GetHeader("To:");

	  // make the authors address first in list
	  if (!adr.IsEmpty())
		 rep += ",\r\n ";
	  adr = rep + adr;

	  // add the header back at present location
	  int idx = linman->DelHeader("To:");
	  if (idx < 0)
		 idx = linman->EOH();
	  lines->Insert(idx, "To: " + adr);
	  delete linman;
   }

   //lines->Insert(0, "Followup using: " + TemplateName);
   Followup->AddStrings(lines);

   delete tmpart;
   delete orgart;
   delete lines;
   delete Template;
   return true;
}
//---------------------------------------------------------------------------

bool __fastcall ClipboardReply(TForm *Owner, TStrings *Reply,
   AnsiString TemplateName, LPSTR ReplyMessage)
{
   TStringList *sig = new TStringList();
   TStringList *message = new TStringList();

   if (!ReplyMessage)
	  message->Text = Clipboard()->AsText;
   else // use replymessage instead of clipboard text
	  message->SetText(ReplyMessage);

   TMsgManager *msg = new TMsgManager(message);
   AnsiString nymhdr = msg->GetHeader("Nym:");
   bool isnymmsg = nymhdr != "";
   if(!msg->HasHeader("Reply-To: ") && !msg->HasHeader("From: "))
   {
	  MsgBox(Owner,
		"The clipboard doesn't appear to contain an email message!    \n\n"
		"Chances are you didn't include the message headers   \n"
		"when you copied it. Please give it another try.",
		 MT_INFO, MB_OK);
	  delete msg;
	  delete message;
	  return false;
   }

   if (isnymmsg)
   {
	  if (GPGIsPGPMessage(message) && !DecryptNymMessage(Owner, message))
	  {
		 delete msg;
		 delete message;
		 return false;
	  }

	  if (GPGIsPGPMessage(message) && !DecryptMessage(Owner, message, true))
	  {
		 delete msg;
		 delete message;
		 return false;
	  }
	  ProcessNymMessage(Owner, message, sig, nymhdr);
   }

   if (GPGIsPGPMessage(message) && !DecryptMessage(Owner, message, false))
   {
	  delete msg;
	  delete message;
	  return false;
   }

   if (!nymhdr.IsEmpty())
	  message->Insert(0, "Nym: " + nymhdr);
   bool ok = true, badtemplate = true;
   while (badtemplate)
   {
	  if (TemplateName.IsEmpty() &&
		 (TemplateName = SelectTemplateReply(Owner)) == "")
	  {
		 delete msg;
		 delete message;
		 return false;
	  }
	  ok = ReplyToMessage(Owner, message, TemplateName, Reply, badtemplate);
	  if (ok)
		 break;
	  TemplateName = "";
   }

   //if (ok && isnymmsg)
   //{
   //	  InsertNymserverVerification(Reply, sig);
   //}
   delete msg;
   delete message;
   delete sig;
   return ok;
}
//---------------------------------------------------------------------------

bool __fastcall ReplyToMessage(TForm *Owner, TStrings *Message,
   AnsiString &TemplateName, TStrings *Reply, bool &badtemplate)
{
   AnsiString nymhdr, hdr, sub, mID, pststr, replyto, blank;
   Reply->Clear();

   TIniFile *ini = new TIniFile(AppIni());
   pststr = ini->ReadString("settings", "quotemarks", "\"> \"");
   DropQuotes(pststr);
   delete ini;

   // get the template and test for problems
   TStringList *Template = new TStringList();
   DB->LoadTemplate(TemplateName, Template);

   // in case the template is empty
   if (!Template->Count)
   {
	  MsgBox(Owner,
		 "Problem! The template you selected is empty! There aren't   \n"
		 "any lines of text at all! You'll need to fix that.   ",
		  MT_INFO, MB_OK);
	  delete Template;
	  return false;
   }
   // in case the template is headers only (no empty line after headers)
   // add an empty line
   if (needsemptyline(Template))
	  Template->Add("");

   TStringList *lines = new TStringList();
   TMsgManager *orgmsg = new TMsgManager(Message);
   TMsgManager *tmpmsg = new TMsgManager(Template);
   TMsgManager *linmsg = new TMsgManager(lines);
   bool isnymmsg = orgmsg->GetHeader("Nym:") != "";

   bool quote = tmpmsg->GetHeader("Quote-Original: ").LowerCase() != "no";
   tmpmsg->DelHeader("Quote-Original: ");

   // get special header positions from template if there
   while ((hdr = tmpmsg->PopHeader()) != "")
   {
	  if (hdr.Pos("References:") == 1 && hdr.Length() < 13)
		 hdr = orgmsg->Get4Reply("References:");
	  else
	  if (hdr.Pos("To:") == 1 && hdr.Length() < 5)
		 hdr = orgmsg->Get4Reply("To:");
	  else
	  if (hdr.Pos("Subject:") == 1 && hdr.Trim().Length() < 10)
		 hdr = orgmsg->Get4Reply("Subject:");

	  lines->Text = lines->Text + hdr;
   }

   hdr = "";
   // add special headers to end of headers if not handled above
   if (!lines->Text.Pos("To: "))
	  hdr = orgmsg->Get4Reply("To: ");
   if (!lines->Text.Pos("Subject: ") && !orgmsg->GetHeader("Subject: ").IsEmpty())
	  hdr += orgmsg->Get4Reply("Subject: ");
   if (!hdr.IsEmpty())
	  lines->Text = lines->Text + hdr;

   lines->Add("");
   if (quote)
	  lines->Add("On " + orgmsg->GetHeader("Date: ") + ", you wrote:");

   int i, eoh = orgmsg->EOH();
   // remove trailing nulls
   while (Message->Count && Message->Strings[Message->Count-1].Trim().IsEmpty())
		 Message->Delete(Message->Count-1);

   // add quoted message text to new message
   if (quote)
	  for (i = eoh; i < Message->Count; i++)
		 lines->Add(pststr + Message->Strings[i]);
   lines->Add("");

   // remove trailing nulls from template
   while (Template->Count && Template->Strings[Template->Count-1].IsEmpty())
	  Template->Delete(Template->Count-1);

   // copy template sig to lines
   for (i = 1; i < Template->Count; i++)
	  lines->Add(Template->Strings[i]);

   // nym functionality
   if (isnymmsg && !CheckFormatNymMessage(Owner, orgmsg, linmsg, badtemplate))
   {
	  delete linmsg;
	  delete tmpmsg;
	  delete orgmsg;
	  delete lines;
	  delete Template;
	  return false;
   }

   // finished
   Reply->AddStrings(lines);

   delete linmsg;
   delete tmpmsg;
   delete orgmsg;
   delete lines;
   delete Template;
   return true;
}
//---------------------------------------------------------------------------
#ifdef QS2
void __fastcall ClipboardReceive(TForm *Owner)
{
   TStringList *msg = new TStringList();
   msg->Text = Clipboard()->AsText;
   TMsgManager *msgman = new TMsgManager(msg);
   if (msgman->HasHeader("From:") || msgman->HasHeader("Reply-To:"))
   {
	  AnsiString key = msgman->receivekey;
	  msgman->DelHeader("Xref:");
	  DB->SaveMessage("inbox", key, msg);
	  QS2MainWin->index->Add(key);
	  QS2MainWin->IndexPane->Items->Count++;
	  QS2MainWin->OpenMessage("inbox", key);
   }
   else
   {
	  MsgBox(Owner,
		"The clipboard doesn't appear to contain an email message!    \n\n"
		"Chances are you didn't include the message headers   \n"
		"when you copied it. Please give it another try.",
		 MT_INFO, MB_OK
	  );
   }
   delete msgman;
   delete msg;
}
#endif
//---------------------------------------------------------------------------
