/*
*  gpg.cpp
*
*  Copyright (C) 2010 Richard Christman                                 
*
*  This file is part of QuickSilver Lite.
*
*  QuickSilver Lite is free software; you can redistribute it and/or modify 
*  it under the terms of the GNU General Public License as published by the 
*  Free Software Foundation; either version 3 of the License, or (at your
*  option) any later version.
*
*  This program is distributed in the hope that it will be useful, but 
*  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
*  or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
*  for more details.
*
*  You should have received a copy of the GNU General Public License along 
*  with this program; if not, see <http://www.gnu.org/licenses>.
* 
*  Additional permission under GNU GPL version 3 section 7
*
*  If you modify QuickSilver Lite, or any covered work, by linking or 
*  combining it with CodeGear VCL (or a modified version of that library),
*  containing parts covered by the terms of CodeGear's license, the licensors
*  of QuickSilver Lite grant you additional permission to convey the 
*  resulting work.
*
*/

//---------------------------------------------------------------------------
#include <vcl.h>
#pragma hdrstop
#include <sys\types.h>

#include "gpg.h"
#include "gpg2.h"
#include "qsdbs.h"
#include "global.h"
#include "utility.h"
#include "msgmans.h"
#include "pgpspassdlgs.h"
#include "pgpcpassdlgs.h"
#include "pgpdpassdlgs2.h"
#include "pgprecipdlgs.h"
#include "pgpcache.h"
#include "mimeencs.h"
#include "prevdlgs.h"
#include "eraser.h"
#include "dsadlgs.h"
#include "blenddlgs.h"
//---------------------------------------------------------------------------
#pragma package(smart_init)
#pragma resource "*.dfm"
// TGPGerrDlg *GPGerrDlg;
AnsiString gpgdatadir;
AnsiString gpgdir;
//---------------------------------------------------------------------------

AnsiString __fastcall GPGDir()
{
   //return GetShortName(AppDir()).LowerCase();
   return gpgdir;  // has already had GetShortName called
}
//---------------------------------------------------------------------------

AnsiString __fastcall GPGDataDir()
{
   //return GetShortName(AppDataDir()).LowerCase();
   return gpgdatadir; // has already had GetShortNameApplied
}
//---------------------------------------------------------------------------

AnsiString __fastcall GPGExe()
{
   return GPGDir() + "gpg.exe";
}
//---------------------------------------------------------------------------

AnsiString __fastcall GPGConf()
{
   return GPGDataDir() + "gpg.conf";
}
//---------------------------------------------------------------------------

AnsiString __fastcall GPGIconvDll()
{
   return GPGDir() + "iconv.dll";
}
//---------------------------------------------------------------------------

AnsiString __fastcall GPGTempFile()
{
   static int c = 1;
   AnsiString filename = GPGTempFileBase();
   if (FileExists(filename + IntToStr(c)))
      c++;
   return filename + IntToStr(c);
}
//---------------------------------------------------------------------------

AnsiString __fastcall GPGTempFileBase()
{
   return GPGDataDir() + "gpg.tmp";
}
//---------------------------------------------------------------------------

bool __fastcall GPGIsPGPMessage(TStrings *Message)
{
   int x;
   x = Message->IndexOf(BeginPGPMessage);
   if (x >= 0 && Message->IndexOf(EndPGPMessage) > x)
	  return true;
   x = Message->IndexOf(BeginPGPSignedMessage);
   if (x >= 0 && Message->IndexOf(EndPGPSignature) > x)
	  return true;
   return false;
}
//---------------------------------------------------------------------------

bool __fastcall GPGIsPGPSignedMessage(TStrings *Message)
{
   int x = Message->IndexOf(BeginPGPSignedMessage);
   return x >= 0 && Message->IndexOf(EndPGPSignature) > x;
}
//---------------------------------------------------------------------------

bool __fastcall GPGHasPGPSignatureVerification(TStrings *Message)
{
   int x = Message->IndexOf(PGPSignatureVerification);
   return x >= 0;
}
//---------------------------------------------------------------------------

AnsiString __fastcall GPGDefaultKeyID()
{
   AnsiString line, defkey;
   if (FileExists(GPGConf()))
   {
	  TStringList *cfg = new TStringList();
	  cfg->LoadFromFile(GPGConf());
	  for (int i = 0; i < cfg->Count; i++)
	  {
		 line = cfg->Strings[i].Trim();
		 if (line.LowerCase().Pos("default-key") == 1)
		 {
			defkey = line.SubString(12, 18);
			break;
		 }
	  }
   }
   return defkey.Trim();
}
//---------------------------------------------------------------------------

AnsiString __fastcall GPGTrustModel()
{
   AnsiString line, model;
   if (FileExists(GPGConf()))
   {
	  TStringList *cfg = new TStringList();
	  cfg->LoadFromFile(GPGConf());
	  for (int i = 0; i < cfg->Count; i++)
	  {
		 line = cfg->Strings[i].Trim();
		 if (line.LowerCase().Pos("trust-model") == 1)
		 {
			model = line.SubString(12, 18).Trim();
			break;
		 }
	  }
   }
   return model;
}
//---------------------------------------------------------------------------

AnsiString __fastcall GPGGetKeyID8(const AnsiString &keystr)
{
   AnsiString longid = GetTok(lkfKeyID, keystr, ':');
   return "0x"+longid.SubString(longid.Length()-7, 8);
}
//---------------------------------------------------------------------------

AnsiString __fastcall GPGGetUserID(const AnsiString &keyid)
{
   return GetTok(lkfKeyUID, GPGGetPublicKeyStr(keyid), ':');
}

//---------------------------------------------------------------------------
// main function
//---------------------------------------------------------------------------

bool __fastcall GPGProcessMessage(TForm *Owner, TStrings *Message)
{
   TGPG *pgp = new TGPG();
   if (pgp->enabled && !pgp->CheckInitSuccess())
   {
	  delete pgp;
	  return false;
   }

   TIniFile *ini = new TIniFile(AppIni());
   bool preview = ini->ReadBool("settings", "previewsend", false);
#ifndef QS2
   bool expertmode = ini->ReadBool("settings", "expertmode", false);
#else
   bool expertmode = true;
#endif
   delete ini;
   bool ok, enc, sig, sym = false;
   AnsiString hdrname, header, pass, enckeys, sigkey;
   AnsiString trustmodel = GPGTrustModel().LowerCase();

   while (true)
   {
	  TMsgManager *msg = new TMsgManager(Message);
	  hdrname = "PgpBtn:";
	  if ((header = msg->GetHeader(hdrname)) == "")
	  {
		 hdrname = "Pgp:";
		 header = msg->GetHeader(hdrname);
	  }

	  if (header.IsEmpty())
	  {
		 delete msg;
		 break;
	  }
	  assert(header != "");
	  msg->DelHeader(hdrname);

	  AnalysePGPHeader(header, enc, sig, enckeys, sigkey);
	  if (!pgp->enabled && (sig || enckeys != ""))
	  {
		 MsgBox(Owner,
			"Sorry, you're attempting to use PGP public key encryption,  \n"
			"but you don't have PGP enabled! You'll need fix that in  \n"
			"the Options Dialog.",
			MT_INFO, MB_OK);
		 delete pgp;
		 return false;
	  }
	  if (sig && sigkey.IsEmpty())
	  {
		 AnsiString keystr, signer = msg->GetHeader("Nym:");
		 if (!signer.IsEmpty())
		 {
			keystr = GPGGetPublicKeyStr(GetEAddress(signer));
			sigkey = GPGGetKeyID8(keystr);
		 }
		 else
		 {
			sigkey = GPGDefaultKeyID();
		 }
	  }
	  /*if (sig && sigkey.IsEmpty())
	  {
		 AnsiString keystr, signer = msg->GetHeader("Nym:");
		 if (signer.IsEmpty())
			signer = msg->GetHeader("From:");
		 keystr = GPGGetPublicKeyStr(GetEAddress(signer));
		 sigkey = GPGGetKeyID8(keystr);
	  }*/
	  delete msg;

	  if (enc && (!expertmode || !pgp->enabled)) // really just novice mode
	  {
		 AnsiString symkey = GetCEncryptionPassword(Owner);
		 if (symkey.IsEmpty())
		 {
			delete pgp;
			return false;
		 }
		 if (!pgp->CEncryptMessage(Message, symkey))
		 {
			delete pgp;
			return false;
		 }
		 continue;
	  }

	  if (enc)
	  {
		 if (enckeys.IsEmpty() &&
			!GetRecipientPGPPublicKeys(Owner, Message, enckeys, sym, sig))
		 {
			delete pgp;
			return false;
		 }

		 if (sym)
		 {
			AnsiString symkey = enckeys;
			if (!pgp->CEncryptMessage(Message, symkey))
			{
			   delete pgp;
			   return false;
			}
			delete pgp;
			return true;
		 }

		 if (trustmodel == "pgp" || trustmodel == "classic" || trustmodel == "direct")
		 {
			AnsiString keyid;
			int x = 0; while ((keyid = GetTok(x, enckeys, ',')) != "")
			{
			   AnsiString ctrust, keystr = GPGGetPublicKeyStr(keyid);
			   if (keystr.IsEmpty())
			   {
				  MsgBox(Owner,
					 "Problem! You can't encrypt your message to key:   \n" +
					 keyid + "   \n\n"
					 "The key DOESN'T EXIST!",
					 MT_INFO, MB_OK);
				  delete pgp;
				  return false;
			   }
			   ctrust = GetTok(1, keystr, ':');
			   assert(ctrust.Length());
			   if (ctrust[1] == 'e')
			   {
				  MsgBox(Owner,
					 "Problem! You can't encrypt your message to key:   \n" +
					 keyid + " " + GetTok(9, keystr, ':') + "   \n\n"
					 "The key is EXPIRED!",
					 MT_INFO, MB_OK);
				  delete pgp;
				  return false;
			   }
			   if (ctrust[1] != 'u' && ctrust[1] != 'f')
			   {
				  if (trustmodel == "pgp" || trustmodel == "classic")
				  {
					 MsgBox(Owner,
						"Key, " + keyid + " " + GetTok(9, keystr, ':') + ",   \n"
						"isn't fully trusted! Before you can encrypt your message    \n"
						"to this key you need to SIGN the key with your default    \n"
						"signing key!",
						MT_INFO, MB_OK);
					 delete pgp;
					 return false;
				  }
				  else
				  {
					 MsgBox(Owner,
						"Key, " + keyid + " " + GetTok(9, keystr, ':') + ",   \n"
						"isn't fully trusted! Before you can encrypt your message    \n"
						"to this key you need to set OWNER-TRUST for the key to   \n"
						"FULLY trusted!",
						MT_INFO, MB_OK);
					 delete pgp;
					 return false;
				  }
			   }
			   x++;
			}
		 }
	  }

	  if (sig)
	  {
		 if (!PGPCache->Get(sigkey, pass))
			pass = GetSecretKeyPassword(Owner, sigkey);
		 if (pass.IsEmpty())
		 {
			delete pgp;
			return false;
		 }

		 if (trustmodel == "pgp" || trustmodel == "classic" || trustmodel == "direct")
		 {
			AnsiString keystr = GPGGetPublicKeyStr(sigkey),
			   ctrust = GetTok(1, keystr, ':');
			assert(ctrust.Length());
			if (ctrust[1] == 'e')
			{
			   MsgBox(Owner,
				  "Problem! You can't SIGN your message with key:   \n" +
				  sigkey + " " + GetTok(9, keystr, ':') + "   \n\n"
				  "The key is EXPIRED!",
				  MT_INFO, MB_OK);
				  delete pgp;
				  return false;
			}
		 }
	  }

	  if (HasAttachments(Message) && !EncodeMIMEAttachments(Owner, Message, true))
	  {
		 MsgBox(Owner,
		   "Sorry, I've failed to MIME encode your attachments! ",
			MT_ERROR, MB_OK);
		 delete pgp;
		 return false;
	  }

	  if (sig && !enc)
	  {
		 if (!pgp->SignMessage(Message, sigkey, pass))
		 {
			ZeroString(pass);
			delete pgp;
			return false;
		 }
		 PGPCache->Add(sigkey, pass);
		 //ZeroString(pass);

		 if (preview && PreviewMessage(Owner, "message PGP signed",
			Message) == IDCANCEL)
		 {
			Owner->Perform(WM_SHOWSTATUS, 0, (LPARAM)"User canceled operation!");
			delete pgp;
			return false;
		 }
	  }
	  else if (enc && !sig)
	  {
		 if (sym)
		 {
			ok = pgp->CEncryptMessage(Message, enckeys);
		 }
		 else
		 {
			ok = pgp->EncryptMessage(Message, enckeys);
		 }
		 if (!ok)
		 {
			delete pgp;
			return false;
		 }

		 if (preview && PreviewMessage(Owner, "message PGP encrypted",
			Message) == IDCANCEL)
		 {
			Owner->Perform(WM_SHOWSTATUS, 0, (LPARAM)"User canceled operation!");
			delete pgp;
			return false;
		 }
	  }
	  else // sign && encrypt
	  {
		 if (sym)
		 {
		  //ok = pgp->CEncryptSignMessage(Message, enckeys, sigkey, pass);
			MsgBox(Owner,
			   "Sorry, QS doesn't support signing conventionally encrypted   \n"
			   "messages. You can clearsign this message BEFORE encryption   \n"
			   "by using two 'Pgp:' headers with signing done in the first   \n"
			   "and encryption in the second.",
				MT_INFO, MB_OK);
			ok = false;
		 }
		 else
		 {
			ok = pgp->EncryptSignMessage(Message, enckeys, sigkey, pass);
			if (ok)
			   PGPCache->Add(sigkey, pass);
		 }
		 if (!ok)
		 {
			delete pgp;
			return false;
		 }

		 if (preview && PreviewMessage(Owner, "message PGP signed/encrypted", Message) == IDCANCEL)
		 {
			Owner->Perform(WM_SHOWSTATUS, 0, (LPARAM)"User canceled operation!");
			delete pgp;
			return false;
		 }
	  }
   }
   delete pgp;
   return true;
}
//---------------------------------------------------------------------------

/*bool __fastcall GPGProcessMessage(TForm *Owner, TStrings *Message)
{
   TGPG *pgp = new TGPG();
   if (pgp->enabled && !pgp->CheckInitSuccess())
   {
	  delete pgp;
	  return false;
   }

   TIniFile *ini = new TIniFile(AppIni());
   bool preview = ini->ReadBool("settings", "previewsend", false);
#ifndef QS2
   bool expertmode = ini->ReadBool("settings", "expertmode", false);
#else
   bool expertmode = true;
#endif
   delete ini;
   bool ok, enc, sig, sym = false;
   AnsiString hdrname, header, pass, enckeys, sigkey;
   AnsiString trustmodel = GPGTrustModel().LowerCase();

   while (true)
   {
	  TMsgManager *msg = new TMsgManager(Message);
	  hdrname = "PgpBtn:";
	  if ((header = msg->GetHeader(hdrname)) == "")
	  {
		 hdrname = "Pgp:";
		 header = msg->GetHeader(hdrname);
	  }

	  if (header.IsEmpty())
	  {
		 delete msg;
		 break;
	  }
	  assert(header != "");
	  msg->DelHeader(hdrname);

	  AnalysePGPHeader(header, enc, sig, enckeys, sigkey);
	  if (!pgp->enabled && (sig || enckeys != ""))
	  {
		 MsgBox(Owner,
			"Sorry, you're attempting to use PGP public key encryption,  \n"
			"but you don't have PGP enabled! You'll need fix that in  \n"
			"the Options Dialog.",
			MT_INFO, MB_OK);
		 delete pgp;
		 return false;
	  }
	  if (sig && sigkey.IsEmpty())
	  {
		 AnsiString keystr, signer = msg->GetHeader("Nym:");
		 if (signer.IsEmpty())
			signer = msg->GetHeader("From:");
		 keystr = GPGGetPublicKeyStr(GetEAddress(signer));
		 sigkey = GPGGetKeyID8(keystr);
	  }
	  delete msg;

	  if (enc && (!expertmode || !pgp->enabled)) // really just novice mode
	  {
		 AnsiString symkey = GetCEncryptionPassword(Owner);
		 if (symkey.IsEmpty())
		 {
			delete pgp;
			return false;
		 }
		 if (!pgp->CEncryptMessage(Message, symkey))
		 {
			delete pgp;
			return false;
		 }
		 continue;
	  }

	  if (enc)
	  {
		 if (enckeys.IsEmpty() &&
			!GetRecipientPGPPublicKeys(Owner, Message, enckeys, sym, sig))
		 {
			delete pgp;
			return false;
		 }

		 if (sym)
		 {
			AnsiString symkey = enckeys;
			if (!pgp->CEncryptMessage(Message, symkey))
			{
			   delete pgp;
			   return false;
			}
			delete pgp;
			return true;
		 }

		 if (trustmodel == "pgp" || trustmodel == "classic" || trustmodel == "direct")
		 {
			AnsiString keyid;
			int x = 0; while ((keyid = GetTok(x, enckeys, ',')) != "")
			{
			   AnsiString ctrust, keystr = GPGGetPublicKeyStr(keyid);
			   if (keystr.IsEmpty())
			   {
				  MsgBox(Owner,
					 "Problem! You can't encrypt your message to key:   \n" +
					 keyid + "   \n\n"
					 "The key DOESN'T EXIST!",
					 MT_INFO, MB_OK);
				  delete pgp;
				  return false;
			   }
			   ctrust = GetTok(1, keystr, ':');
			   assert(ctrust.Length());
			   if (ctrust[1] == 'e')
			   {
				  MsgBox(Owner,
					 "Problem! You can't encrypt your message to key:   \n" +
					 keyid + " " + GetTok(9, keystr, ':') + "   \n\n"
					 "The key is EXPIRED!",
					 MT_INFO, MB_OK);
				  delete pgp;
				  return false;
			   }
			   if (ctrust[1] != 'u' && ctrust[1] != 'f')
			   {
				  if (trustmodel == "pgp" || trustmodel == "classic")
				  {
					 MsgBox(Owner,
						"Key, " + keyid + " " + GetTok(9, keystr, ':') + ",   \n"
						"isn't fully trusted! Before you can encrypt your message    \n"
						"to this key you need to SIGN the key with your default    \n"
						"signing key!",
						MT_INFO, MB_OK);
					 delete pgp;
					 return false;
				  }
				  else
				  {
					 MsgBox(Owner,
						"Key, " + keyid + " " + GetTok(9, keystr, ':') + ",   \n"
						"isn't fully trusted! Before you can encrypt your message    \n"
						"to this key you need to set OWNER-TRUST for the key to   \n"
						"FULLY trusted!",
						MT_INFO, MB_OK);
					 delete pgp;
					 return false;
				  }
			   }
			   x++;
			}
		 }
	  }

	  if (sig)
	  {
		 if (!PGPCache->Get(sigkey, pass))
			pass = GetSecretKeyPassword(Owner, sigkey);
		 if (pass.IsEmpty())
		 {
			delete pgp;
			return false;
		 }

		 if (trustmodel == "pgp" || trustmodel == "classic" || trustmodel == "direct")
		 {
			AnsiString keystr = GPGGetPublicKeyStr(sigkey),
			   ctrust = GetTok(1, keystr, ':');
			assert(ctrust.Length());
			if (ctrust[1] == 'e')
			{
			   MsgBox(Owner,
				  "Problem! You can't SIGN your message with key:   \n" +
				  sigkey + " " + GetTok(9, keystr, ':') + "   \n\n"
				  "The key is EXPIRED!",
				  MT_INFO, MB_OK);
				  delete pgp;
				  return false;
			}
		 }
	  }

	  if (HasAttachments(Message) && !EncodeMIMEAttachments(Owner, Message, true))
	  {
		 MsgBox(Owner,
		   "Sorry, I've failed to MIME encode your attachments! ",
			MT_ERROR, MB_OK);
		 delete pgp;
		 return false;
	  }

	  if (sig && !enc)
	  {
		 if (!pgp->SignMessage(Message, sigkey, pass))
		 {
			ZeroString(pass);
			delete pgp;
			return false;
		 }
		 PGPCache->Add(sigkey, pass);
		 //ZeroString(pass);

		 if (preview && PreviewMessage(Owner, "message PGP signed",
			Message) == IDCANCEL)
		 {
			Owner->Perform(WM_SHOWSTATUS, 0, (LPARAM)"User canceled operation!");
			delete pgp;
			return false;
		 }
	  }
	  else if (enc && !sig)
	  {
		 if (sym)
		 {
			ok = pgp->CEncryptMessage(Message, enckeys);
		 }
		 else
		 {
			ok = pgp->EncryptMessage(Message, enckeys);
		 }
		 if (!ok)
		 {
			delete pgp;
			return false;
		 }

		 if (preview && PreviewMessage(Owner, "message PGP encrypted",
			Message) == IDCANCEL)
		 {
			Owner->Perform(WM_SHOWSTATUS, 0, (LPARAM)"User canceled operation!");
			delete pgp;
			return false;
		 }
	  }
	  else // sign && encrypt
	  {
		 if (sym)
		 {
		  //ok = pgp->CEncryptSignMessage(Message, enckeys, sigkey, pass);
			MsgBox(Owner,
			   "Sorry, QS doesn't support signing conventionally encrypted   \n"
			   "messages. You can clearsign this message BEFORE encryption   \n"
			   "by using two 'Pgp:' headers with signing done in the first   \n"
			   "and encryption in the second.",
				MT_INFO, MB_OK);
			ok = false;
		 }
		 else
		 {
			ok = pgp->EncryptSignMessage(Message, enckeys, sigkey, pass);
			if (ok)
			   PGPCache->Add(sigkey, pass);
		 }
		 if (!ok)
		 {
			delete pgp;
			return false;
		 }

		 if (preview && PreviewMessage(Owner, "message PGP signed/encrypted",
			Message) == IDCANCEL)
		 {
			Owner->Perform(WM_SHOWSTATUS, 0, (LPARAM)"User canceled operation!");
			delete pgp;
			return false;
		 }
	  }
   }

   delete pgp;
   return true;
}*/

//---------------------------------------------------------------------------
// main decrypy/verify function
//---------------------------------------------------------------------------

bool __fastcall GPGProcessMessageDecryption(TForm *Owner, TStrings *Message)
{
   assert(GPGIsPGPMessage(Message));

   TGPG *pgp = new TGPG();
   if (Message->IndexOf(BeginPGPSignedMessage) >= 0)
   {
	   if (!pgp->enabled)
	   {
		  DSABox(Owner, "pgpdsadlg1",
			 "I can't verify this signed message. PGP Public Key "
			 "Encryption is not enabled.",
			 "PGP Information",
			  MB_OK);
		  delete pgp;
		  return false;
	   }
	   bool ok = pgp->VerifyClearSigned(Message);
	   delete pgp;
	   return ok;
   }

   DecodeQuotedPrintableInArmor(Message);
   TStringList *keys = new TStringList();
   pgp->GetEncryptionKeys(Message, keys);

   assert(keys->Count);
   if (!pgp->enabled)
   {
	  if (!keys->Text.Pos("passphrase"))
	  {
		 DSABox(Owner, "pgpdsadlg2",
			"I can't decrypt this message. PGP Public Key "
			"Encryption is not enabled.",
			"PGP Information",
			 MB_OK);
		 delete keys;
		 delete pgp;
		 return false;
	  }
	  // delete the actual keys leaving passphrase line
	  int x = 0; while (x < keys->Count)
	  {
		 if (!keys->Strings[x].Pos("passphrase"))
			keys->Delete(x);
		 else
			x++;
	  }
   }

   AnsiString key, pass;
   for (int i = 0; i < keys->Count; i++)
   {
	  key = keys->Strings[i];
	  if (!key.Pos("passphrase") && PGPCache->Get(key, pass))
		 break;
   }

   TIniFile *ini = new TIniFile(AppIni());
   int tries, maxtries = ini->ReadInteger("settings", "decrypttrys", 3);
   delete ini;
   if (maxtries < 1)
	  maxtries = 1;
   tries = 1;

   for (int i = 0; i < keys->Count; i++)
	  if (!keys->Strings[i].LowerCase().Pos("passphrase"))
		 keys->Strings[i] = keys->Strings[i] + " " + GPGGetUserID(keys->Strings[i]);

   while (true)
   {
	  if (pass.IsEmpty())
		 pass = GetDecryptionPassphrase(Owner, keys);
	  if (pass.IsEmpty())
	  {
		 delete keys;
		 delete pgp;
		 return false;
	  }

	  if (!pgp->DecryptMessage(Message, pass))
	  {
		 pass = "";
		 tries++;
		 if (tries < maxtries)
			continue;
	  }
	  else
		 PGPCache->Add(key, pass);
	  break;
   }
   delete keys;
   delete pgp;

   return true;
}

//---------------------------------------------------------------------------

/*void __fastcall ReportMultipleKeys(TForm *Owner, TStrings *Keys)
{
   MsgBox(Owner,
	  "There's a problem with this nym message. It's encrypted    \n"
	  "with these keys:\n\n"+
	   Keys->Text+"\n"
	  "This layer of encryption should report a single key.   \n\n"
	  "Take care!",
	   MT_WARN, MB_OK);
}
//---------------------------------------------------------------------------

void __fastcall ReportSuccessAndFurtherPGP(TForm *Owner, TStrings *Message)
{
   AnsiString text = "Nym message successfuly decrypted!   \n";
   if (Message->IndexOf(BeginPGPMessage) >= 0)
	  text += "\n"
		 "I see there's another layer of encryption. That   \n"
		 "layer must have been added by the sender.   \n";
	MsgBox(Owner, text, MT_INFO, MB_OK);
}
//---------------------------------------------------------------------------

AnsiString __fastcall AsymmetricLayerStr()
{
   return "This is the public key encrypted, nymserver signed layer "
	  "of your NYM message.";
}
//---------------------------------------------------------------------------

AnsiString __fastcall SymmetricLayerStr(int layer)
{
   return "This is symmetric encryption layer "+IntToStr(layer)+" "
	  "of your incoming NYM message.";
}
//---------------------------------------------------------------------------

void __fastcall CheckReportBlendingAttack(TForm *Owner, const AnsiString &msgkeystr,
   const AnsiString &nymkeystr)
{
   if (msgkeystr == nymkeystr)
	  return;

   ShowPossibleBlendingAttack(Owner, msgkeystr, nymkeystr);
}
//---------------------------------------------------------------------------

AnsiString __fastcall msgkeystr(TStrings *feedback)
{
   AnsiString keystr, keyid16, keyid8;
   for (int i = feedback->Count - 1; i > 0; i--)
   {
	  AnsiString line = feedback->Strings[i];
	  if (line.Pos("[GNUPG:] NEED_PASSPHRASE") > 0)
	  {
		 keyid16 = GetTok(2, line, ' ');
		 break;
	  }
   }
   keystr = GPGGetPublicKeyStr(keyid16);
   keyid16 = GetTok(4, keystr, ':');
   keyid8 = keyid16.SubString(keyid16.Length() - 7, 8);
   return "0x" + keyid8 + " " + GetTok(0, keystr, ':');
}*/
//---------------------------------------------------------------------------

/*bool __fastcall GPGDecryptNymMessage2(TForm *Owner, TStrings *Message)
{
   assert(Message != NULL);
   assert(Message->IndexOf(BeginPGPMessage) >= 0);
   TMsgManager *msg = new TMsgManager(Message);
   AnsiString nymid, keyid, ntype, nymkeystr, nymhdr = msg->GetHeader("Nym:");
   assert(nymhdr != "");
   nymid = GetTok(0, nymhdr, ';');
   keyid = GetTok(1, nymhdr, ';').TrimLeft();
   keyid = keyid.SubString(7, keyid.Length());
   ntype = GetTok(2, nymhdr, ';').TrimLeft();
   ntype = ntype.SubString(6, ntype.Length());
   assert(nymid != "" && keyid != "" && ntype != "");
   nymkeystr = keyid + " " + nymid;
   delete msg;

   int layer = 0;
   bool sym, isthrowkey;
   AnsiString keystr, passphrase;
   TStringList *keys = new TStringList();
   TStringList *feedback = new TStringList();
   while (GPGIsPGPMessage(Message))
   {
	  layer++;
	  sym = false;
	  isthrowkey = false;
	  if (!GPGGetEncryptionKeys2(Message, keys))
	  {
		 delete feedback;
		 delete keys;
		 return false;
	  }
	  if (keys->Count > 1)
		 ReportMultipleKeys(Owner, keys);

	  for (int i = 0; i < keys->Count; i++)
	  {
		 AnsiString key = keys->Strings[i];
		 if (key.LowerCase().Pos("passphrase"))
		 {
			sym = true;
			continue;
		 }
		 if (key.Pos("00000000"))
		 {
			keys->Strings[i] = nymkeystr;
			isthrowkey = true;
			continue;
		 }
		 keys->Strings[i] = key + " " + GetTok(0, GPGGetPublicKeyStr(key), ':');
	  }

	  int tries = 0;
	  while (true)
	  {
		 tries++;
		 if (sym)
		 {
			passphrase = GetNymDecryptionPassphrase(Owner, keys,
			   "Decrypting NYM Message Symmetric Layer",
				SymmetricLayerStr(layer));
		 }
		 else
		 {
			if (isthrowkey)
			{
			   passphrase = GetNymDecryptionPassphrase2(Owner, keys,
				  "Decrypting NYM Message Public-Key Layer");
			}
			else
			{
			   passphrase = GetNymDecryptionPassphrase(Owner, keys,
				   "Decrypting NYM Message Public-Key Layer",
					AsymmetricLayerStr());
			}
		 }
		 if (passphrase.IsEmpty())
		 {
			delete feedback;
			delete keys;
			return false;
		 }

		 bool ok = GPGDecryptMessage2(Owner, Message, passphrase, true, feedback);

		 if ((feedback->Text.Pos("decryption failed: bad key") ||
			  feedback->Text.Pos("DECRYPTION_FAILED")) && tries < 3)
			continue;

		 if (!ok)
		 {
			delete feedback;
			delete keys;
			return false;
		 }

		 if (isthrowkey)
		 {
			CheckReportBlendingAttack(Owner, msgkeystr(feedback), nymkeystr);
		 }
		 else
		 {
			CheckReportBlendingAttack(Owner, keys->Strings[0], nymkeystr);
		 }

		 break;
	  }

	  if (!sym)
	  {
		 if (GPGIsPGPMessage(Message))
			ReportSuccessAndFurtherPGP(Owner, Message);

		 break;
	  }
   }

   delete feedback;
   delete keys;
   return true;
}*/
//---------------------------------------------------------------------------

/*bool __fastcall GPGDecryptNymMessage(TForm *Owner, TStrings *Message)
{
   assert(Message != NULL);
   assert(Message->IndexOf(BeginPGPMessage) >= 0);
   TMsgManager *msg = new TMsgManager(Message);
   assert(msg->HasHeader("Nym:"));
   AnsiString nymid, keyid, ntype, nymhdr = msg->GetHeader("Nym:");
   assert(nymhdr != "");
   nymid = GetTok(0, nymhdr, ';');
   keyid = nymid.SubString(1,10);
   nymid = nymid.SubString(11, nymid.Length()).TrimLeft();
   ntype = GetTok(1, nymhdr, ';').TrimLeft();
   ntype = ntype.SubString(6, 32);
   assert(nymid != "" && keyid != "" && ntype != "");
   delete msg;

   int layer = 0;
   bool sym, isthrowkey;
   AnsiString passphrase;
   TStringList *keys = new TStringList();
   TStringList *feedback = new TStringList();
   while (Message->IndexOf(BeginPGPMessage) >= 0)
   {
	  layer++;
	  if (!GPGGetEncryptionKeys2(Message, keys))
	  {
		 delete keys;
		 return false;
	  }
	  if (keys->Count > 1)
		 ReportMultipleKeys(Owner, keys);

	  sym = keys->Strings[0].LowerCase().Pos("passphrase") > 0;
	  isthrowkey = keys->Strings[0].LowerCase().Pos("00000000") > 0;

	  int tries = 0;
	  while (true)
	  {
		 tries++;
		 passphrase = GetNymDecryptionPassphrase(Owner, keys,
			(sym ? "Decrypting NYM Message Symmetricl Layer "+IntToStr(layer) :
				   AnsiString("Decrypting NYM Message Public-Key Layer")),
			(sym ? SymmetricLayerStr(layer) : AsymmetricLayerStr()));
		 if (passphrase.IsEmpty())
		 {
			delete feedback;
			delete keys;
			return false;
		 }
		 bool ok = GPGDecryptMessage2(Owner, Message, passphrase, true, feedback);
//look(feedback);
		 if ((feedback->Text.Pos("decryption failed: bad key") > 0 ||
			  feedback->Text.Pos("DECRYPTION_FAILED")) && tries < 3)
			continue;

		 if (!ok)
		 {
			delete feedback;
			delete keys;
			return false;
		 }
		 break;
	  }

look(Message);
	  if (!sym)
	  {
		 ReportSuccess(Owner, Message);
		 break;
	  }
   }

   delete feedback;
   delete keys;
   return true;
}*/
//---------------------------------------------------------------------------

void __fastcall DecodeQuotedPrintableInArmor(TStrings *msg)
{
   bool gotbegin = false;
   for (int i = 0; i < msg->Count; i++)
   {
	  AnsiString line = msg->Strings[i];
	  if (i < msg->Count-1 && msg->Strings[i + 1] == EndPGPMessage)
		 break;
	  if (line == EndPGPMessage)
		 break;
	  if (line == BeginPGPMessage)
		 gotbegin = true;
	  if (!gotbegin)
		 continue;
	  int x = line.Pos("=");
	  if (x > 0 && x < line.Length())
		 msg->Strings[i] = DecodeQuotedPrintable(line);
   }
}
//---------------------------------------------------------------------------

AnsiString __fastcall getcommandline(const AnsiString &item, const AnsiString defvalue)
{
   TIniFile *ini = new TIniFile(AppIni());
   AnsiString cmdline = ini->ReadString("gpg", item, defvalue);
   delete ini;
   return GPGExe() + " " + cmdline.TrimLeft();
}
//---------------------------------------------------------------------------

bool __fastcall setfield(const AnsiString &field, const AnsiString &value,
   AnsiString &line)
{
   int x = line.Pos(field);
   if (x == 0)
      return false;
   line.Delete(x, field.Length());
   line.Insert(value, x);
   return true;
}
//---------------------------------------------------------------------------

bool __fastcall checkgpgavailable()
{
   bool ok = true;
   if (!FileExists(GPGExe()))
   {
	  MsgBox(Screen->ActiveForm,
		 "Problem! The file 'gpg.exe' is missing from my home directory.  \n"
         "I can't complete the PGP operation! Reinstalling the program   \n"
		 "should fix this problem.",
		  MT_ERROR, MB_OK);
	  ok = false;
   }
   if (!FileExists(GPGIconvDll()))
   {
      MsgBox(Screen->ActiveForm,
         "Problem! The gpg file 'iconv.dll' is missing from my home directory.  \n"
         "I can't complete the PGP operation! Reinstalling the program   \n"
         "should fix this problem.",
		  MT_ERROR, MB_OK);
      ok = false;
   }

   return ok;
}
//---------------------------------------------------------------------------

AnsiString __fastcall recipkeystr(AnsiString CommaKeyIDs)
{
   int x; while ((x = CommaKeyIDs.Pos(",")) > 0)
   {
      CommaKeyIDs.Delete(x, 1);
      CommaKeyIDs.Insert(" -r ", x);
   }
   return CommaKeyIDs;
}
//---------------------------------------------------------------------------

void __fastcall trimbodyright(TStrings *msgbody)
{
   for (int i = 0; i < msgbody->Count; i++)
   {
	  AnsiString line = msgbody->Strings[i].TrimRight();
	  if (line != msgbody->Strings[i])
		 msgbody->Strings[i] = line;
   }
}
//---------------------------------------------------------------------------

void __fastcall trimbodybottom(TStrings *msgbody)
{
   while (msgbody->Count && msgbody->Strings[msgbody->Count - 1].Trim().IsEmpty())
	  msgbody->Delete(msgbody->Count - 1);
   msgbody->Add("");
}
//---------------------------------------------------------------------------

void __fastcall getheaders(TStrings *headers, TStrings *message)
{
   assert(headers != NULL);
   assert(message != NULL);
   while (message->Count)
   {
	  AnsiString line = message->Strings[0];
	  headers->Add(line);
	  message->Delete(0);
	  if (line.IsEmpty())
		  break;
   }
   assert(message->Count > 0);
}
//---------------------------------------------------------------------------

void __fastcall putheaders(TStrings *message, TStrings *headers)
{
   for (int i = 0; i < headers->Count; i++)
	  message->Insert(i, headers->Strings[i]);
}
//---------------------------------------------------------------------------

void __fastcall assemblemessage(TStrings *message, TStream *output,
   TStrings *headers, bool maskversion)
{
   message->LoadFromStream(output);
   if (maskversion) // do the masking of the Version Line gpg emits
   {
	  for (int i = 0; i < message->Count; i++)
	  {
		 if (message->Strings[i].Pos("Version:") == 1)
		 {
			message->Strings[i] = "Version: N/A";
			break;
		 }
	  }
   }
   for (int i = 0; i < headers->Count; i++)
	  message->Insert(i, headers->Strings[i]);
}
//---------------------------------------------------------------------------

bool __fastcall loadmessage(TForm *Owner, TStrings *message, TStrings *headers,
   const AnsiString &filename, bool maskversion)
{
   int t = 0;
   Snooze(500);

again:
   try
   {
	  t++;
	  message->LoadFromFile(filename);
	  if (maskversion) // do the masking of the Version Line gpg emits
	  {
		 for (int i = 0; i < message->Count; i++)
		 {
			if (message->Strings[i].Pos("Version:") == 1)
			{
			   message->Strings[i] = "Version: N/A";
			   break;
			}
		 }
	  }
   }
   catch (Exception &e)
   {
	  if (t == 10)
	  {
		 MsgBox(Owner,
			"I was unable to load the encrypted temporary file: " + filename + " \n" +
			 e.Message,
			 MT_ERROR, MB_OK);
		 return false;
	  }
	  else
	  {
		 Snooze(500);
		 goto again;
	  }
   }

   putheaders(message, headers);
   return true;
}
//---------------------------------------------------------------------------
// deprecated: decided not to do leave the empties be
void __fastcall checkremovekeyringempties()
{
   AnsiString pubring = GPGDir()+"pubring.gpg", secring = GPGDir()+"secring.gpg";
   if (FileExists(pubring) && GetFileSize(pubring) == 0)
	  DeleteFile(pubring);
   if (FileExists(secring) && GetFileSize(secring) == 0)
	  DeleteFile(secring);
}
//---------------------------------------------------------------------------

AnsiString __fastcall GPGVersion()
{
   if (!checkgpgavailable())
	  return "GPG not installed!";

   AnsiString cmdline = getcommandline(GPGversion, GPGVersionCmd);
   //setfield("%homedir%", GPGDataDir(), cmdline);

   TStringList *output = new TStringList();
   TGPGConsole *con = new TGPGConsole(NULL);
   bool ok = con->Run(cmdline, output);

   if (ok)
	  ok = con->error->Text.LowerCase().Pos("no valid openpgp data found") == 0;
   delete con;

   if (!ok)
   {
	  delete output;
	  return "failed";
   }

   assert(output->Count > 0);
   AnsiString ver = output->Strings[0];
   ver = "GnuPG\t\t" + GetTok(2, ver, ' ');
   delete output;

   //checkremovekeyringempties();
   return ver;
}
//---------------------------------------------------------------------------

bool __fastcall GPGListPublicKeysWithColonsNE(const AnsiString &KeyIDS, TStrings *Keys)
{
   assert(KeyIDS.Length() && Keys != NULL);
   Keys->Clear();

   if (!checkgpgavailable())
	  return false;

   AnsiString cmdline = getcommandline(GPGlistpublickeyswithcolons,
	  GPGListPublicKeysWithColonsCmd);
   cmdline += " " + KeyIDS.Trim();

   TGPGConsole *con = new TGPGConsole(Screen->ActiveForm);
   bool ok = con->Run(cmdline, Keys);
   delete con;
   return ok;
}
//---------------------------------------------------------------------------

AnsiString __fastcall GPGGetPublicKeyStr(const AnsiString &Keyid)
{
   AnsiString keystr;
   TStringList *details = new TStringList();
   GPGListPublicKeysWithColonsNE(Keyid, details);

   for (int i = 0; i < details->Count; i++)
   {
	  if (details->Strings[i].Pos("pub:") == 1)
	  {
		 keystr = details->Strings[i];
		 break;
	  }
   }
   delete details;
   if (!keystr.IsEmpty())
	  keystr = GetTok(9, keystr, ':') + ":" + keystr.SubString(5, keystr.Length());
   return keystr;
}
//---------------------------------------------------------------------------

bool __fastcall GPGIsFileEncrypted(const AnsiString &FileName)
{
   if (!FileExists(FileName))
	  return false;

   if (!checkgpgavailable())
	  return false;

   AnsiString cmdline = getcommandline(GPGlistencryptionkeysfromfile,
	  GPGListEncryptionKeysFromFileCmd);
   setfield("%filename%", GetShortName(FileName), cmdline);

   TStringList *output = new TStringList();
   TGPGConsole *con = new TGPGConsole(NULL);
//sm(cmdline);
   bool ok = con->Run(cmdline, output);
   if (ok)
   {
	  ok = con->error->Text.LowerCase().Pos("gpg: no valid openpgp data found") == 0;
   }
//
//sm(con->error->Text+con->status->Text);
//
   delete con;
   delete output;

   //checkremovekeyringempties();
   return ok;
}
//---------------------------------------------------------------------------

bool __fastcall wipeencryptedDB()
{
   return GetAdvancedConfigItem(GPGwipeencrypteddb, GPGWipeEncryptedDB);
}
//---------------------------------------------------------------------------

bool __fastcall GPGEncryptDatabase(TQSDB *DB, AnsiString &Passphrase)
{
   if (!checkgpgavailable())
	  return false;

   AnsiString cmdline = getcommandline(GPGcencrypttofile, GPGCEncryptToFileCmdNA),
	  tmpname = AppDBTmp();
   if (FileExists(tmpname))
	  DeleteFile(tmpname);

   setfield("%filename%", tmpname, cmdline);

   TStringList *output = new TStringList();
   TGPGConsole *con = new TGPGConsole(Screen->ActiveForm);
   bool ok = con->Run(cmdline, Passphrase.c_str(), DB->stream, output);
   if (ok)
   {
	  // append the size of the unencrypted database to end of encrypted database
	  __int64 dbSize = DB->stream->Size;
	  TFileStream *strm = new TFileStream(tmpname, fmOpenWrite | fmShareExclusive);
	  strm->Seek(0, soFromEnd);
	  strm->Write(&dbSize, sizeof(__int64));
	  delete strm;

	  // delete original db and rename tmpname to AppDB()
	  if (DB->filename != AppDBTmp()) // only wipe/rename if DB isn't AppDBTmp()
	  {                               // (when packing database)
		 if (FileExists(AppDB()))
		 {
			if (!GPGIsFileEncrypted(AppDB()) || wipeencryptedDB())
			   eraserWipeFile(AppDB(), false);
			else
			   DeleteFile(AppDB());
		 }
		 RenameFile(tmpname, AppDB());
	  }
   }
   else
   {
	  Snooze(200);
	  ShowGPGError(Screen->ActiveForm, con->error, con->status);
   }
   ZeroString(Passphrase);
   delete con;
   delete output;

   //checkremovekeyringempties();
   return ok;
}
//---------------------------------------------------------------------------
/* This function is ready to replacy the original below

bool __fastcall GPGDecryptDatabase(TQSDB *DB, AnsiString &Passphrase)
{
   if (!checkgpgavailable())
	  return false;

   AnsiString cmdline = getcommandline(GPGdecrypt, GPGDecryptCmd),
	  filename = AppDB();

   // get unencrypted file size from end of encrypted file
   __int64 dbSize;
   int size, size64 = sizeof(__int64);
   TFileStream *diskstrm = new TFileStream(filename, fmOpenRead | fmShareExclusive);
   diskstrm->Seek(-size64, soFromEnd);
   diskstrm->Read(&dbSize, size64);
   size = dbSize > 2147483648 ? 2147483648 : (int)dbSize;
   DB->memstream->SetSize(size);
   diskstrm->Seek(0, soFromBeginning);

   TGPGConsole4 *con = new TGPGConsole4(Screen->ActiveForm);
   bool ok = con->Run(cmdline, Passphrase.c_str(), diskstrm, DB->stream);
   //ZeroString(Passphrase);   need the passphrase after exit to save in masterpassword
   if (con->status->Text.Pos("DECRYPTION_OKAY") && !con->status->Text.Pos("NODATA"))
   {
	  ok = true;
   }
   else
   {
	  Snooze(200);
	  ShowGPGError(Screen->ActiveForm, con->error, con->status);
	  ok = false;
   }
   delete con;
   delete diskstrm;

   return ok;
}*/
//---------------------------------------------------------------------------

bool __fastcall GPGDecryptDatabase(TQSDB *DB, AnsiString &Passphrase)
{
   if (!checkgpgavailable())
	  return false;

   AnsiString cmdline = getcommandline(GPGdecryptfromfile, GPGDecryptFromFileCmd),
	  filename = AppDB();

   // get unencrypted file size from end of encrypted file
   __int64 dbSize;
   int size, size64 = sizeof(__int64);
   TFileStream *strm = new TFileStream(filename, fmOpenRead | fmShareExclusive);
   strm->Seek(-size64, soFromEnd);
   strm->Read(&dbSize, size64);
   size = dbSize > 2147483648 ? 2147483648 : (int)dbSize;
   DB->memstream->SetSize(size);
   delete strm;

   setfield("%filename%", GetShortName(filename), cmdline);

   TGPGConsole *con = new TGPGConsole(Screen->ActiveForm);
   bool ok = con->Run(cmdline, Passphrase.c_str(), DB->stream);
 //ZeroString(Passphrase);   need the passphrase after exit to save in masterpassword
   if (!ok)
   {
	  Snooze(200);
	  ShowGPGError(Screen->ActiveForm, con->error, con->status);
   }
   delete con;

   return ok;
}
//---------------------------------------------------------------------------

bool __fastcall GPGLoadMasterPassword(TQSDB *db, AnsiString &passphrase)
{
   if (!checkgpgavailable())
      return false;

   AnsiString key;
   db->LoadSessionKey(key);
   return GPGDecryptStreamToString(key, db->MasterPassword, passphrase);
}
//---------------------------------------------------------------------------

bool __fastcall GPGSaveMasterPassword(TQSDB *db, AnsiString &passphrase)
{
   if (!checkgpgavailable())
      return false;

   AnsiString key;
   db->LoadSessionKey(key);
   bool ok = GPGEncryptStringToStream(key, passphrase, db->MasterPassword);
   ZeroString(passphrase);
   return ok;
}
//---------------------------------------------------------------------------

bool __fastcall GPGEncryptStringToStream(AnsiString &Passphrase, AnsiString &Input,
   TStream *Output)
{
   // max size of data = 32k
   if (!checkgpgavailable())
      return false;

   AnsiString cmdline = getcommandline(GPGencryptstringtostream, GPGEncryptStringToStreamCmd),
	  filename = AppDB();

   //setfield("%homedir%", GPGDataDir(), cmdline);

   TGPGConsole *con = new TGPGConsole(Screen->ActiveForm);
   bool ok = con->Run(cmdline, Passphrase, Input, Output);
   ZeroString(Passphrase);
   if (!ok)
   {
      Snooze(200);
	  ShowGPGError(Screen->ActiveForm, con->error, con->status);
   }
   delete con;

   Snooze(200);

   //checkremovekeyringempties();
   return ok;
}
//---------------------------------------------------------------------------

bool __fastcall GPGDecryptStreamToString(AnsiString &Passphrase, TStream *Input,
   AnsiString &Output)
{
   // max size of data = 32k
   if (!checkgpgavailable())
	  return false;

   AnsiString cmdline = getcommandline(GPGdecryptstreamtostring, GPGDecryptStreamToStringCmd),
	  filename = AppDB();

   //setfield("%homedir%", GPGDataDir(), cmdline);

   TGPGConsole *con = new TGPGConsole(Screen->ActiveForm);
   bool ok = con->Run(cmdline, Passphrase, Input, Output);
   ZeroString(Passphrase);
   if (!ok)
   {
	  Snooze(200);
	  ShowGPGError(Screen->ActiveForm, con->error, con->status);
   }
   delete con;

   Snooze(200);

   //checkremovekeyringempties();
   return ok;
}
//---------------------------------------------------------------------------

bool __fastcall GPGListSecretKeysWithColonsRingName(const AnsiString &ringname,
   TStrings *keys)
{
   if (!checkgpgavailable())
	  return false;

   keys->Clear();
   assert(FileExists(ringname));
   AnsiString cmdline = getcommandline(GPGlistsecretkeyswithcolonsringname,
	  GPGListSecretKeysWithColonsRingNameCmd);
   setfield("%secring%", GetShortName(ringname), cmdline);

   TMemoryStream *output = new TMemoryStream();

   TGPGConsole4 *con = new TGPGConsole4(Screen->ActiveForm);
   bool ok = con->Run(cmdline, NULL, NULL, output);
   if (ok)
   {
	  ok = output->Size > 0;
	  if (ok)
	  {
		 keys->LoadFromStream(output);
	  }
	  else
	  {
		 Snooze(200);
		 ShowGPGError(Screen->ActiveForm, con->error, con->status);
	  }
   }
   delete con;
   delete output;

   //checkremovekeyringempties();
   return ok;
}
//---------------------------------------------------------------------------

/*bool __fastcall GPGListSecretKeysWithColonsRingName(const AnsiString &ringname,
   TStrings *keys)
{
   if (!checkgpgavailable())
	  return false;

   assert(FileExists(ringname));
   AnsiString cmdline = getcommandline(GPGlistsecretkeyswithcolonsringname,
	  GPGListSecretKeysWithColonsRingNameCmd);
   //setfield("%homedir%", GPGDataDir(), cmdline);
   setfield("%secring%", GetShortName(ringname), cmdline);

   TGPGConsole *con = new TGPGConsole(Screen->ActiveForm);
   bool ok = con->Run(cmdline, keys);
   if (!ok)
   {
	  Snooze(200);
	  ShowGPGError(Screen->ActiveForm, con->error, con->status);
   }
   delete con;

   //checkremovekeyringempties();
   return ok;
}*/
//---------------------------------------------------------------------------

bool __fastcall GPGListSecretKeysWithColonsNE(const AnsiString &keyids,
   TStrings *keys)
{
   assert(keys != NULL);

   if (!checkgpgavailable())
	  return false;

   AnsiString cmdline = getcommandline(GPGlistsecretkeyswithcolons,
	  GPGListSecretKeysWithColonsCmd);
   if (keyids != "")
	  cmdline += " " + keyids;

   TGPGConsole *con = new TGPGConsole(Screen->ActiveForm);
   bool ok = con->Run(cmdline, keys);
   //if (!ok)
   //{
   //	  Snooze(200);
   //	  ShowGPGError(Screen->ActiveForm, con->error, con->status);
   //}
   delete con;

   //checkremovekeyringempties();
   return ok;
}
//---------------------------------------------------------------------------

bool __fastcall GPGGenRandomToFile(int count, int quality, const AnsiString &filename)
{
   if (!checkgpgavailable())
	  return false;

   assert(quality >= 0 && quality <= 2);
   assert(count > 0);
   assert(filename != "");

   AnsiString cmdline = getcommandline(GPGgenrandom, GPGGenRandomCmd);

   //setfield("%homedir%", GPGDataDir(), cmdline);
   //setfield("%fileame%", filename, cmdline);
   setfield("%quality%", IntToStr(quality), cmdline);
   setfield("%count%", IntToStr(count), cmdline);

   TMemoryStream *output = new TMemoryStream();
   TGPGConsole *con = new TGPGConsole(Screen->ActiveForm);
   bool ok = con->Run(cmdline, output);
   if (!ok)
   {
	  Snooze(200);
      ShowGPGError(Screen->ActiveForm, con->error, con->status);
   }
   else
      output->SaveToFile(filename);
   delete con;
   delete output;


   //checkremovekeyringempties();
   return ok;
}

//---------------------------------------------------------------------------
// GPG
//---------------------------------------------------------------------------

__fastcall TGPG::TGPG()
{
   cfg = new TStringList();
   DB->LoadConfig(PGPConfig, cfg);
   Owner = Screen->ActiveForm;
   error = new TStringList();
   status = new TStringList();
}
//---------------------------------------------------------------------------

__fastcall TGPG::~TGPG()
{
   delete status;
   delete error;
   delete cfg;
}
//---------------------------------------------------------------------------

bool __fastcall TGPG::GetEnabled()
{
   return cfg->Values["pgp"] == "1";
}
//---------------------------------------------------------------------------

AnsiString __fastcall TGPG::GetSecring()
{
   return cfg->Values["secring"];
}
//---------------------------------------------------------------------------

AnsiString __fastcall TGPG::GetPubring()
{
   return cfg->Values["pubring"];
}
//---------------------------------------------------------------------------

AnsiString __fastcall TGPG::GetDefKey()
{
   return cfg->Values["defkey"];
}
//---------------------------------------------------------------------------

bool __fastcall TGPG::GetEncDef()
{
   return cfg->Values["encdef"].ToIntDef(0);
}
//---------------------------------------------------------------------------

bool __fastcall TGPG::GetMaskVer()
{
   return cfg->Values["maskver"].ToIntDef(1);
}
//---------------------------------------------------------------------------

void __fastcall TGPG::erasefile(const AnsiString &filename)
{
   if (FileExists(filename))
	  eraserWipeFile(filename, false);
}
//---------------------------------------------------------------------------

bool __fastcall TGPG::SignMessage(TStrings *Message, const AnsiString &SignKey,
	  AnsiString &Passphrase)
{
   return GPGClearSignMessage2(Owner, Message, SignKey, Passphrase, maskversion);

   /*ShowStatus("PGP signing...");
   AnsiString cmdline = getcommandline(GPGclearsigntofile, GPGClearSignToFileCmd),
	  filename = GPGTempFile();

   TStringList *out = new TStringList();
   TStringList *headers = new TStringList();
   getheaders(headers, Message);
   trimbodyright(Message);
   trimbodybottom(Message);

   //setfield("%homedir%", GPGDataDir(), cmdline);
   setfield("%signkey%", SignKey, cmdline);
   setfield("%filename%", filename, cmdline);

   TGPGConsole *con = new TGPGConsole(Owner);
   bool ok = con->Run(cmdline, Passphrase.c_str(), Message, out);
   //ZeroString(Passphrase);  passphrase is zeroed after saving in PGPcache
   if (ok)
	  ok = loadmessage(Owner, Message, headers, filename, maskversion);
   else
	  ShowGPGError(Owner, con->error, con->status);
   delete con;

   if (FileExists(filename))
	  erasefile(filename);

   delete headers;
   delete out;

   ShowStatus("Done!");
   Snooze(200);
   return ok;*/
}
//---------------------------------------------------------------------------

bool __fastcall TGPG::GetEncryptionKeys(TStrings *Message, TStrings *Keys)
{
   return GPGGetEncryptionKeys2(Message, Keys);
   /*assert(Message->IndexOf(BeginPGPMessage) >= 0);
   AnsiString cmdline = getcommandline(GPGlistencryptionkeys, GPGListEncryptionKeysCmd);

   Keys->Clear();
   TStringList *msg = new TStringList();
   TMemoryStream *input = new TMemoryStream();
   TMemoryStream *output = new TMemoryStream();
   msg->AddStrings(Message);
   while (msg->Strings[0] != BeginPGPMessage)
	  msg->Delete(0);
   assert(msg->Count);

   TGPGConsole4 *con = new TGPGConsole4(NULL);
   bool ok = con->Run(cmdline, NULL, input, output);
   if (ok)
   {
	  AnsiString id, line;
	  for (int i = 0; i < con->status->Count; i++)
	  {
		 line = con->status->Strings[i];
		 if (line.Pos("ENC_TO") > 0)
		 {
			id = GetTok(2, line, ' ');
			assert(id.Length() == 16);
			Keys->Add("0x" + id.SubString(id.Length() - 7, 8));
		 }
	  }
	  for (int i = 0; i < con->error->Count; i++)
	  {
		 line = con->error->Strings[i];
		 if (line.Pos("passphrase") > 0)
			Keys->Add(line.SubString(20, 64).Trim());
	  }
   }
   else
   {
	  Snooze(200);
	  ShowGPGError(Screen->ActiveForm, con->error, con->status);
   }
   delete con;
   delete output;
   delete input;
   delete msg;

   //checkremovekeyringempties();
   return ok;*/
}
//---------------------------------------------------------------------------

void __fastcall FormatSigVerification(TStrings *verified)
{
   assert(verified->Count);
   while (verified->Count && !verified->Strings[0].Pos("Signature made"))
	  verified->Delete(0);
   if (!verified->Count)
	  return;

   TStringList *veri = new TStringList();
   veri->Add("*** PGP SIGNATURE VERIFICATION ***");
   for (int i = 0; i < verified->Count; i++)
   {
	  AnsiString line = verified->Strings[i];
	  if (line.Pos("gpg:") == 1)
		 veri->Add("*** " + line.SubString(6, 128));
   }
   veri->Add("*** Verified " + FormatDateTime("MM/DD/YY HH:NN:SS", Now()));
   verified->Text = veri->Text;
   delete veri;
}
//---------------------------------------------------------------------------

bool __fastcall TGPG::VerifyClearSigned(TStrings *Message)
{
   return GPGVerifyClearSigned2(Owner, Message);

   /*AnsiString cmdline = getcommandline(GPGverifyclearsigned, GPGVerifyClearSignedCmd),
	  tmpname = GPGTempFile();

   TStringList *output = new TStringList();
   TStringList *headers = new TStringList();
   setfield("%filename%", tmpname, cmdline);
   getheaders(headers, Message);
   Message->SaveToFile(tmpname);

   TGPGConsole *con = new TGPGConsole(Screen->ActiveForm);
   bool ok = con->Run(cmdline, output);
   if (ok)
   {
	  FormatSigVerification(con->error);
	  Message->Clear();
	  Message->AddStrings(headers);
	  Message->AddStrings(con->error);
	  Message->Add("*** BEGIN PGP VERIFIED MESSAGE ***");
	  Message->Add("");
	  Message->AddStrings(output);
	  Message->Add("*** END PGP VERIFIED MESSAGE ***");

   }
   else
   {
	  ShowGPGError(Screen->ActiveForm, con->error, con->status);
	  putheaders(Message, headers);
   }
   delete con;

   if (FileExists(tmpname))
	  eraserWipeFile(tmpname, false);

   delete headers;
   delete output;
   ShowStatus("Done!");
   Snooze(200);

   return ok;*/
}
//---------------------------------------------------------------------------

bool __fastcall TGPG::DecryptMessage(TStrings *Message, AnsiString &Passphrase)
{
   return GPGDecryptMessage2(Owner, Message, Passphrase);

 /*assert(Message->IndexOf(BeginPGPMessage) >= 0);
   ShowStatus("PGP decrypting...");
   AnsiString cmdline = getcommandline(GPGdecryptfromfile, GPGDecryptFromFileCmd),
	  tmpname = GPGTempFile();

   TStringList *output = new TStringList();
   TStringList *headers = new TStringList();
   setfield("%filename%", tmpname, cmdline);
   getheaders(headers, Message);
   Message->SaveToFile(tmpname);

   TGPGConsole *con = new TGPGConsole(Owner);
   bool ok = con->Run(cmdline, Passphrase.c_str(), (TStrings*)NULL, output);
   if (ok || con->status->Text.Pos("DECRYPTION_OKAY"))
   {
	  FormatSigVerification(con->error);
	  Message->Clear();
	  Message->AddStrings(headers);
	  if (con->error->Count)
	  {
		 Message->AddStrings(con->error);
		 Message->Add("*** BEGIN PGP DECRYPTED/VERIFIED MESSAGE ***");
		 Message->Add("");
	  }
	  else
	  {
		 Message->Add("*** BEGIN PGP DECRYPTED MESSAGE ***");
	  }
	  Message->AddStrings(output);
	  if (con->error->Count)
	  {
		 Message->Add("*** END PGP DECRYPTED/VERIFIED MESSAGE ***");
	  }
	  else
	  {
		 Message->Add("*** END PGP DECRYPTED MESSAGE ***");
	  }
	  ok = true;
   }
   else
   {
	  ShowGPGError(Owner, con->error, con->status);
	  putheaders(Message, headers);
   }
   delete con;

   if (FileExists(tmpname))
	  erasefile(tmpname);

   delete headers;
   delete output;
   ShowStatus("Done!");
   Snooze(200);

   return ok; */
}
//---------------------------------------------------------------------------

bool __fastcall TGPG::EncryptMessage(TStrings *Message, const AnsiString &CommaEncKeys)
{
   return GPGEncryptMessage2(Owner, Message, CommaEncKeys, maskversion);

   /*ShowStatus("PGP encrypting...");
   AnsiString cmdline = getcommandline(GPGencrypt, GPGEncryptCmd),
	  filename = GPGTempFile();

   TStringList *output = new TStringList();
   TStringList *headers = new TStringList();
   getheaders(headers, Message);
   trimbodybottom(Message);

   //setfield("%homedir%", GPGDataDir(), cmdline);
   setfield("%recipkey%", recipkeystr(CommaEncKeys), cmdline);
   setfield("%filename%", filename, cmdline);

   TGPGConsole *con = new TGPGConsole(Owner);
   bool ok = con->Run(cmdline, NULL, Message, output);
   if (ok)
	  ok = loadmessage(Owner, Message, headers, filename, maskversion);
   else
	  ShowGPGError(Owner, con->error, con->status);
   delete con;

   if (FileExists(filename))
	  erasefile(filename);

   delete headers;
   delete output;
   ShowStatus("Done!");
   Snooze(200);

   //checkremovekeyringempties();
   return ok; */
}
//---------------------------------------------------------------------------

bool __fastcall TGPG::EncryptSignMessage(TStrings *Message, const AnsiString &CommaEncKeys,
   const AnsiString &SignKey, AnsiString &Passphrase)
{
   return GPGEncryptSignMessage2(Owner, Message, CommaEncKeys, SignKey, Passphrase, maskversion);

   /*ShowStatus("PGP encrypting/signing...");
   AnsiString cmdline = getcommandline(GPGencryptsigntofile, GPGEncryptSignToFileCmd),
	  filename = GPGTempFile();

   TStringList *out = new TStringList();
   TStringList *headers = new TStringList();
   getheaders(headers, Message);
   trimbodyright(Message);
   trimbodybottom(Message);

   //setfield("%homedir%", GPGDataDir(), cmdline);
   setfield("%signkey%", SignKey, cmdline);
   setfield("%recipkey%", recipkeystr(CommaEncKeys), cmdline);
   setfield("%filename%", filename, cmdline);

   TGPGConsole *con = new TGPGConsole(Owner);
   bool ok = con->Run(cmdline, Passphrase.c_str(), Message, out);
   //ZeroString(Passphrase);  passphrase is zeroed after saving in PGPCache
   if (ok)
	  ok = loadmessage(Owner, Message, headers, filename, maskversion);
   else
	  ShowGPGError(Owner, con->error, con->status);
   delete con;

   if (FileExists(filename))
	  erasefile(filename);

   delete headers;
   delete out;

   ShowStatus("Done!");
   Snooze(200);

   //checkremovekeyringempties();
   return ok;*/
}
//---------------------------------------------------------------------------

bool __fastcall TGPG::CEncryptMessage(TStrings *Message, AnsiString &Passphrase)
{
   return GPGCEncryptMessage2(Owner, Message, Passphrase, maskversion);
   /*ShowStatus("PGP encrypting...");
   AnsiString cmdline = getcommandline(GPGcencrypttofilena, GPGCEncryptToFileCmd),
	  filename = GPGTempFile();

   TStringList *out = new TStringList();
   TStringList *headers = new TStringList();
   getheaders(headers, Message);

   //setfield("%homedir%", GPGDataDir(), cmdline);
   setfield("%filename%", filename, cmdline);

   TGPGConsole *con = new TGPGConsole(Owner);
   bool ok = con->Run(cmdline, Passphrase.c_str(), Message, out);
   ZeroString(Passphrase);
   if (ok)
	  ok = loadmessage(Owner, Message, headers, filename, maskversion);
   else
	  ShowGPGError(Owner, con->error, con->status);
   delete con;

   if (FileExists(filename))
	  erasefile(filename);

   delete headers;
   delete out;

   ShowStatus("Done!");
   Snooze(200);

   //checkremovekeyringempties();
   return ok;*/
}
//---------------------------------------------------------------------------

int __fastcall TGPG::ListSecretKeysWithColons(TStrings *keys)
{
   AnsiString cmdline = getcommandline(GPGlistsecretkeyswithcolons,
	  GPGListSecretKeysWithColonsCmd);
   //setfield("%homedir%", GPGDataDir(), cmdline);
   setfield("%secring%", GetShortName(secring), cmdline);

   TGPGConsole *con = new TGPGConsole(Owner);
   bool ok = con->Run(cmdline, keys);
   status->Text = con->status->Text;
   error->Text = con->error->Text;
   delete con;

   //checkremovekeyringempties();
   return ok;
}
//---------------------------------------------------------------------------

int __fastcall TGPG::ListSecretKeysWithColons(TStrings *keys, TStrings *keyids)
{
   AnsiString cmdline = getcommandline(GPGlistsecretkeyswithcolons,
	  GPGListSecretKeysWithColonsCmd);
 //setfield("%secring%", GetShortName(secring), cmdline);
   for (int i = 0; i < keyids->Count; i++)
	  cmdline += " " + keyids->Strings[i];

   TGPGConsole *con = new TGPGConsole(Owner);
   bool ok = con->Run(cmdline, keys);
   status->Text = con->status->Text;
   error->Text = con->error->Text;
   delete con;

   //checkremovekeyringempties();
   return ok;
}
//---------------------------------------------------------------------------

int __fastcall TGPG::ListSecretKeysWithColons(TStrings *keys, const AnsiString &commakeyids)
{
   TStringList *keyids = new TStringList();
   keyids->CommaText = commakeyids;
   bool ok = ListSecretKeysWithColons(keys, keyids);
   delete keyids;
   return ok;
}
//---------------------------------------------------------------------------

int __fastcall TGPG::ListPublicKeysWithColons(TStrings *keys)
{
   AnsiString cmdline = getcommandline(GPGlistpublickeyswithcolons,
	  GPGListPublicKeysWithColonsCmd);
   setfield("%pubring%", GetShortName(pubring), cmdline);

   TGPGConsole *con = new TGPGConsole(Owner);
   bool ok = con->Run(cmdline, keys);
   status->Text = con->status->Text;
   error->Text = con->error->Text;
   delete con;

   //checkremovekeyringempties();
   return ok;
}
//---------------------------------------------------------------------------

int __fastcall TGPG::ListPublicKeysWithColons(TStrings *keys, TStrings *keyids)
{
   AnsiString cmdline = getcommandline(GPGlistpublickeyswithcolons,
	  GPGListPublicKeysWithColonsCmd);
 //setfield("%pubring%", GetShortName(pubring), cmdline);
   for (int i = 0; i < keyids->Count; i++)
	  cmdline += " " + keyids->Strings[i];

   TGPGConsole *con = new TGPGConsole(Owner);
   bool ok = con->Run(cmdline, keys);
   status->Text = con->status->Text;
   error->Text = con->error->Text;
   delete con;

   return ok;
}
//---------------------------------------------------------------------------

int __fastcall TGPG::ListPublicKeysWithColons(TStrings *keys, const AnsiString &commakeyids)
{
   TStringList *keyids = new TStringList();
   keyids->CommaText = commakeyids;
   bool ok = ListPublicKeysWithColons(keys, keyids);
   delete keyids;
   return ok;
}
//---------------------------------------------------------------------------

void __fastcall TGPG::ShowStatus(AnsiString statusmsg)
{
   statusmsg = " " + statusmsg + " ";
   Owner->Perform(WM_SHOWSTATUS, 0, (LPARAM)statusmsg.c_str());
}
//---------------------------------------------------------------------------

void __fastcall AnalysePGPHeader(const AnsiString &header, bool &enc, bool &sig,
   AnsiString &enckey, AnsiString &sigkey)
{
   enc = false;
   sig = false;
   enckey = "";
   sigkey = "";

   AnsiString clause;
   int x = 0; while ((clause = GetTok(x, header, ';').Trim()) != "")
   {
	  if (clause.LowerCase().Pos("encrypt") > 0)
	  {
		 enckey = GetTok(1, clause, '=');
		 if (!enckey.IsEmpty())
		 {
			int x;
			if (enckey[1] == ',')
			   enckey = enckey.SubString(2, enckey.Length() - 1);
			if (enckey[enckey.Length()] == ',')
			   enckey = enckey.SubString(1, enckey.Length() - 1);
			while ((x = enckey.Pos("\t")) > 0)
			   enckey.Delete(x, 1);
			while ((x = enckey.Pos("\r")) > 0)
			   enckey.Delete(x, 1);
			while ((x = enckey.Pos("\n")) > 0)
			   enckey.Delete(x, 1);
			while ((x = enckey.Pos(" ")) > 0)
			   enckey.Delete(x, 1);
		 }
		 enc = true;
	  }
	  else
	  if (clause.LowerCase().Pos("sign") > 0)
	  {
		 sigkey = GetTok(1, clause, '=');
		 sig = true;
	  }
	  x++;
   }
}
//---------------------------------------------------------------------------

bool __fastcall TGPG::CheckInitSuccess()
{
   if (!enabled)
   {
	  MsgBox(Owner,
		 "Problem! You are trying to use PGP but PGP isn't enabled! You'll  \n"
		 "need to fix that in the Options Dialog. ",
		 MT_INFO, MB_OK);
	  return false;
   }

   if (secring.IsEmpty())
   {
	  MsgBox(Owner,
		 "Problem! You haven't assigned your PGP PRIVATE keyring.  \n"
		 "To use PGP, you need to fix that in the Options Dialog.  \n",
		  MT_INFO, MB_OK);
	  return false;
   }

   if (pubring.IsEmpty())
   {
	  MsgBox(Owner,
		 "Problem! You haven't assigned your PGP PUBLIC keyring.  \n"
		 "To use PGP, you need to fix that in the Options Dialog.  \n",
		  MT_INFO, MB_OK);
	  return false;
   }

   if (defaultkey.IsEmpty())
   {
	  MsgBox(Owner,
		 "Problem! You haven't assigned your default signing key.  \n"
		 "To use PGP, you need to fix that in the Options Dialog.  \n",
		  MT_INFO, MB_OK);
	  return false;
   }

   if (!FileExists(secring))
   {
	  MsgBox(Owner,
		"Problem! You secret keyring, " + secring + ", doesn't exist!  \n"
		"To use PGP, You'll have to fix that in the Options Dialog. ",
		 MT_INFO, MB_OK);
	  return false;
   }

   if (!FileExists(pubring))
   {
	  MsgBox(Owner,
		"Problem! You public keyring, " + pubring + ", doesn't exist!  \n"
		"To use PGP, you'll have to fix that in the Options Dialog. ",
		 MT_INFO, MB_OK);
	  return false;
   }

   if (!FileExists(GPGExe()))
   {
	  MsgBox(Owner,
		 "Problem! The program 'gpg.exe' is missing. It should be in   \n"
		 "my home directory. Reinstalling the prohram should fix this.",
		  MT_ERROR, MB_OK);
	  return false;
   }

   if (!SetGPGConfValues())
	  return false;

   return true;
}
//---------------------------------------------------------------------------

/*bool __fastcall TGPG::SetConfValues()
{
   TStringList *cfg = new TStringList();
   if (FileExists(GPGConf()))
	  cfg->LoadFromFile(GPGConf());
   if (!cfg->Count)
   {
	  cfg->Add("no-greeting");
	  cfg->Add("no-mdc-warning");
	  cfg->Add("trust-model always");
	  cfg->Add("no-default-keyring");
   }

   int i = 0; while (i < cfg->Count)
   {
	  AnsiString line = cfg->Strings[i].TrimLeft().LowerCase();
	  if (line.IsEmpty())
	  {
		 cfg->Delete(i);
		 continue;
	  }
      if (line.Pos("keyring") == 1 || line.Pos("secret-keyring") == 1 ||
		 line.Pos("default-key") == 1)
            break;
      i++;
   }
   while (cfg->Count > i)
	  cfg->Delete(i);

   while (cfg->Count && cfg->Strings[0].TrimLeft().Pos("#") == 1)
      cfg->Delete(0);

   cfg->Insert(0, "# QS maintains the values for keyring, secret-keyring,");
   cfg->Insert(1, "# and default-key. If you add or modify values ABOVE these");
   cfg->Insert(2, "# entries, do so with extreme care. Items added below these");
   cfg->Insert(3, "# will be deleted.");
   cfg->Add("keyring " + GetShortName(pubring).LowerCase());
   cfg->Add("secret-keyring " + GetShortName(secring).LowerCase());
   cfg->Add("default-key " + defaultkey);

   try
   {
      cfg->SaveToFile(GPGConf());
   }
   catch (Exception &e)
   {
      MsgBox(Owner,
         "I failed to create the gpg config file for this PGP operation. I   \n"
		 "can't provide PGP! The exception reported was:   \n" +
		 e.Message,
         MT_ERROR, MB_OK);
      delete cfg;
	  return false;
   }

   delete cfg;
   return true;
}*/
//---------------------------------------------------------------------------

bool __fastcall SetGPGConfValues()
{
   TStringList *pgp = new TStringList();
   DB->LoadConfig(PGPConfig, pgp);

   TStringList *cfg = new TStringList();
   if (FileExists(GPGConf()))
	  cfg->LoadFromFile(GPGConf());
   if (!cfg->Count)
   {
      cfg->Add("display-charset iso-8859-1");
	  cfg->Add("no-greeting");
	  cfg->Add("no-mdc-warning");
	  cfg->Add("trust-model always");
	  cfg->Add("no-default-keyring");
   }

   int i = 0; while (i < cfg->Count)
   {
	  AnsiString line = cfg->Strings[i].TrimLeft().LowerCase();
	  if (line.IsEmpty())
	  {
		 cfg->Delete(i);
		 continue;
	  }
	  if (line.Pos("keyring") == 1 || line.Pos("secret-keyring") == 1 ||
		 line.Pos("default-key") == 1)
			break;
	  i++;
   }
   while (cfg->Count > i)
	  cfg->Delete(i);

   while (cfg->Count && cfg->Strings[0].TrimLeft().Pos("#") == 1)
	  cfg->Delete(0);

   cfg->Insert(0, "# QS maintains the values for keyring, secret-keyring,");
   cfg->Insert(1, "# and default-key. If you add or modify values ABOVE these");
   cfg->Insert(2, "# entries, do so with extreme care. Items added below these");
   cfg->Insert(3, "# will be deleted.");
   cfg->Add("keyring " + GetShortName(pgp->Values["pubring"]).LowerCase());
   cfg->Add("secret-keyring " + GetShortName(pgp->Values["secring"]).LowerCase());
   cfg->Add("default-key " + pgp->Values["defkey"]);

   try
   {
	  cfg->SaveToFile(GPGConf());
   }
   catch (Exception &e)
   {
	  MsgBox(Screen->ActiveForm,
		 "I failed to create the gpg config file for this PGP operation. I   \n"
		 "can't provide PGP! The exception reported was:   \n" +
		 e.Message,
		 MT_ERROR, MB_OK);
	  delete cfg;
	  return false;
   }

   delete cfg;
   delete pgp;
   return true;
}
//---------------------------------------------------------------------------

AnsiString __fastcall TGPG::KeyType(const AnsiString &keystr)
{
   return GetTok(lkfKeyType, keystr, ':');
}
//---------------------------------------------------------------------------

AnsiString __fastcall TGPG::KeyValid(const AnsiString &keystr)
{
   return GetTok(lkfKeyValid, keystr, ':');
}
//---------------------------------------------------------------------------

AnsiString __fastcall TGPG::KeyID(const AnsiString &keystr)
{
   return GetTok(lkfKeyID, keystr, ':');
}
//---------------------------------------------------------------------------

AnsiString __fastcall TGPG::KeyID8(const AnsiString &keystr)
{
   AnsiString id = GetTok(lkfKeyID, keystr, ':');
   return id.SubString(id.Length() - 7, 8);
}
//---------------------------------------------------------------------------

AnsiString __fastcall TGPG::KeyUID(const AnsiString &keystr)
{
   return GetTok(lkfKeyUID, keystr, ':');
}

//---------------------------------------------------------------------------
//   GPGConsole
//---------------------------------------------------------------------------

__fastcall TGPGConsole::TGPGConsole(TForm *owner)
   : TObject()
{
   Owner = owner;
   hInputRead = NULL;
   hInputWrite = NULL;
   hOutputRead = NULL;
   hOutputWrite = NULL;
   hErrorRead = NULL;
   hErrorWrite = NULL;
   hPassRead = NULL;
   hPassWrite = NULL;
   hAttrRead = NULL;
   hAttrWrite = NULL;
   hStatRead = NULL;
   hStatWrite = NULL;

   bstatus = false;
   battrib = false;

   exitwait = false;
   exitcode = 0;

   attrib = new TStringList();
   status = new TStringList();
   error = new TStringList();
}
//---------------------------------------------------------------------------

__fastcall TGPGConsole::~TGPGConsole()
{
   delete error;
   delete status;
   delete attrib;

   DestroyHandles();
}
//---------------------------------------------------------------------------

void __fastcall TGPGConsole::SetupFileDescriptors(AnsiString &cmdline)
{
   int p;
   AnsiString fd;

   p = cmdline.Pos("%stat%");
   if (p)
   {
      fd = AnsiString((int)hStatWrite);
      setfield("%stat%", fd, cmdline);
   }

   p = cmdline.Pos("%attr%");
   if (p)
   {
      fd = AnsiString((int)hAttrWrite);
      setfield("%attr%", fd, cmdline);
   }

   p = cmdline.Pos("%pass%");
   if (p)
   {
      fd = AnsiString((int)hPassRead);
      setfield("%pass%", fd, cmdline);
   }
}
//---------------------------------------------------------------------------

bool __fastcall TGPGConsole::CreateHandles()
{
   TSecurityAttributes sa;
   HANDLE hOutputReadTmp, hErrorReadTmp, hInputWriteTmp, hPassWriteTmp,
	  hAttrReadTmp, hStatReadTmp;

   DestroyHandles();
   sa.nLength = sizeof(sa);
   sa.lpSecurityDescriptor = NULL;
   sa.bInheritHandle = true;

   if (!CreatePipe(&hInputRead, &hInputWriteTmp, &sa, 0))
   {
	  ShowError("Pipe create error: \n");
      return false;
   }

   if (!CreatePipe(&hOutputReadTmp, &hOutputWrite, &sa, 0))
   {
      ShowError("Pipe create error: \n");
      return false;
   }

   if (!CreatePipe(&hErrorReadTmp, &hErrorWrite, &sa, 0))
   {
      ShowError("Pipe create error: \n");
      return false;
   }

   if (bstatus && !CreatePipe(&hStatReadTmp, &hStatWrite, &sa, 0))
   {
	  ShowError("Pipe create error: \n");
      return false;
   }

   if (battrib && !CreatePipe(&hAttrReadTmp, &hAttrWrite, &sa, 0))
   {
      ShowError("Pipe create error: \n");
	  return false;
   }

   if (bpasswd && !CreatePipe(&hPassRead, &hPassWriteTmp, &sa, 0))
   {
      ShowError("Pipe create error: \n");
      return false;
   }

   if (!DuplicateHandle(GetCurrentProcess(), hInputWriteTmp,
      GetCurrentProcess(), &hInputWrite, 0, FALSE, DUPLICATE_SAME_ACCESS))
   {
	  ShowError("Duplicate handle error: \n");
      return false;
   }

   if (!DuplicateHandle(GetCurrentProcess(), hOutputReadTmp,
      GetCurrentProcess(), &hOutputRead, 0, FALSE, DUPLICATE_SAME_ACCESS))
   {
      ShowError("Duplicate handle error: \n");
      return false;
   }

   if (!DuplicateHandle(GetCurrentProcess(), hErrorReadTmp,
      GetCurrentProcess(), &hErrorRead, 0, FALSE, DUPLICATE_SAME_ACCESS))
   {
	  ShowError("Duplicate handle error: \n");
      return false;
   }

   if (bstatus && !DuplicateHandle(GetCurrentProcess(), hStatReadTmp,
      GetCurrentProcess(), &hStatRead, 0, FALSE, DUPLICATE_SAME_ACCESS))
   {
	  ShowError("Duplicate handle error: \n");
      return false;
   }

   if (battrib && !DuplicateHandle(GetCurrentProcess(), hAttrReadTmp,
      GetCurrentProcess(), &hAttrRead, 0, FALSE, DUPLICATE_SAME_ACCESS))
   {
	  ShowError("Duplicate handle error: \n");
      return false;
   }

   if (bpasswd && !DuplicateHandle(GetCurrentProcess(), hPassWriteTmp,
	  GetCurrentProcess(), &hPassWrite, 0, FALSE, DUPLICATE_SAME_ACCESS))
   {
      ShowError("Duplicate handle error: \n");
      return false;
   }

   if (battrib) CloseHandle(hAttrReadTmp);
   if (bstatus) CloseHandle(hStatReadTmp);
   if (bpasswd) CloseHandle(hPassWriteTmp);
   CloseHandle(hOutputReadTmp);
   CloseHandle(hErrorReadTmp);
   CloseHandle(hInputWriteTmp);

   return true;
}
//---------------------------------------------------------------------------

void __fastcall TGPGConsole::DestroyHandles()
{
   if (hInputRead) CloseHandle(hInputRead);
   if (hOutputRead) CloseHandle(hOutputRead);
   if (hInputWrite) CloseHandle(hInputWrite);
   if (hOutputWrite) CloseHandle(hOutputWrite);
   if (hErrorRead) CloseHandle(hErrorRead);
   if (hErrorWrite) CloseHandle(hErrorWrite);
   if (hPassRead) CloseHandle(hPassRead);
   if (hPassWrite) CloseHandle(hPassWrite);
   if (hAttrRead) CloseHandle(hAttrRead);
   if (hAttrWrite) CloseHandle(hAttrWrite);
   if (hStatRead) CloseHandle(hStatRead);
   if (hStatWrite) CloseHandle(hStatWrite);

   hInputRead = NULL;
   hInputWrite = NULL;
   hOutputRead = NULL;
   hOutputWrite = NULL;
   hErrorRead = NULL;
   hErrorWrite = NULL;
   hPassRead = NULL;
   hPassWrite = NULL;
   hAttrRead = NULL;
   hAttrWrite = NULL;
   hStatRead = NULL;
   hStatWrite = NULL;
}
//---------------------------------------------------------------------------

void __fastcall TGPGConsole::PrepareStartupInfo()
{
   memset(&startupinfo, 0, sizeof(STARTUPINFO));
   startupinfo.cb = sizeof (STARTUPINFO);
   startupinfo.lpReserved = NULL;
   startupinfo.lpDesktop = NULL;
   startupinfo.lpTitle = NULL;
   startupinfo.lpReserved2 = NULL;
   startupinfo.cbReserved2 = 0;
   startupinfo.dwFlags = STARTF_USESTDHANDLES;

   startupinfo.hStdInput = hInputRead;
   startupinfo.hStdOutput = hOutputWrite;
   startupinfo.hStdError = hErrorWrite;
}
//---------------------------------------------------------------------------

bool __fastcall TGPGConsole::LaunchChildProcess(const AnsiString &cmdLine)
{
   if (!CreateProcess(
	  NULL,             // pointer to filename
	  cmdLine.c_str(),  // pointer to command line string
	  NULL, 		// pointer to process security attributes, we don't need to inherit the handle
	  NULL, 		// pointer to thread security attributes
	  true,  		// handle inheritance flag
	  CREATE_NO_WINDOW, // creation flags
	  NULL, 		// pointer to new environment block
	  NULL, 		// pointer to current directory name
	 &startupinfo,      // pointer to STARTUPINFO
	 &processinfo       // pointer to PROCESS_INFORMATION
   ))
   {
	  ShowError("CreateProcess error: \n");
	  return false;
   }

   //CloseHandle(processinfo.hThread);

   return true;
}
//---------------------------------------------------------------------------

bool __fastcall TGPGConsole::SendPassphrase(char *passphrase)
{
   DWORD byteswrote;
   if (!WriteFile(hPassWrite, passphrase, strlen(passphrase), &byteswrote, NULL))
   {
	  ShowError("Error sending the passphrase to GPG: \n");
	  return false;
   }

   CloseHandle(hPassWrite);
   hPassWrite = NULL;
   return true;
}
//---------------------------------------------------------------------------

bool __fastcall TGPGConsole::Run(AnsiString cmdLine, char *Passphrase,
   TStrings *Input, TStrings *Output)
{
   bstatus = cmdLine.Pos("%stat%");
   battrib = cmdLine.Pos("%attr%");
   bpasswd = cmdLine.Pos("%pass%");

   if (!CreateHandles())
	  return false;

   SetupFileDescriptors(cmdLine);
   PrepareStartupInfo();

   if (!LaunchChildProcess(cmdLine))
	  return false;

   if (Passphrase && !SendPassphrase(Passphrase))
	  return false;

   if (!SendInput(Input))
	  return false;

   if (!ReadOutput(Output))
	  return false;

   return true;
}
//---------------------------------------------------------------------------

bool __fastcall TGPGConsole::Run(AnsiString cmdLine, char *Passphrase,
   TStream *Input, TStrings *Output)
{
   assert(Passphrase && strlen(Passphrase) > 0);
   bstatus = cmdLine.Pos("%stat%");
   battrib = cmdLine.Pos("%attr%");
   bpasswd = cmdLine.Pos("%pass%");

   if (!CreateHandles())
	  return false;

   SetupFileDescriptors(cmdLine);
   PrepareStartupInfo();

   if (!LaunchChildProcess(cmdLine))
	  return false;

   if (Passphrase && !SendPassphrase(Passphrase))
	  return false;

   if (!SendInput(Input))
	  return false;

   if (!ReadOutput(Output))
	  return false;

   return true;
}
//---------------------------------------------------------------------------

bool __fastcall TGPGConsole::Run(AnsiString cmdLine, AnsiString &Passphrase,
	  const AnsiString &Input, TStream *Output)
{
   assert(Passphrase.Length() > 0);
   bstatus = cmdLine.Pos("%stat%");
   battrib = cmdLine.Pos("%attr%");
   bpasswd = cmdLine.Pos("%pass%");

   if (!CreateHandles())
	  return false;

   SetupFileDescriptors(cmdLine);
   PrepareStartupInfo();

   if (!LaunchChildProcess(cmdLine))
	  return false;

   if (!SendPassphrase(Passphrase.c_str()))
	  return false;

   if (!SendInput(Input))
	  return false;

   if (!ReadOutput(Output))
	  return false;

   return true;
}
//---------------------------------------------------------------------------

bool __fastcall TGPGConsole::Run(AnsiString cmdLine, AnsiString &Passphrase,
	  TStream *Input, AnsiString &Output)
{
   assert(Passphrase.Length() > 0);
   bstatus = cmdLine.Pos("%stat%");
   battrib = cmdLine.Pos("%attr%");
   bpasswd = cmdLine.Pos("%pass%");

   if (!CreateHandles())
	  return false;

   SetupFileDescriptors(cmdLine);
   PrepareStartupInfo();

   if (!LaunchChildProcess(cmdLine))
	  return false;

   if (!SendPassphrase(Passphrase.c_str()))
	  return false;

   if (!SendInput(Input))
      return false;

   if (!ReadOutput(Output))
	  return false;

   return true;
}
//---------------------------------------------------------------------------

bool __fastcall TGPGConsole::Run(AnsiString cmdLine, TStrings *Output)
{
   bstatus = cmdLine.Pos("%stat%");
   battrib = cmdLine.Pos("%attr%");

   if (!CreateHandles())
	  return false;

   SetupFileDescriptors(cmdLine);
   PrepareStartupInfo();
//sm(cmdLine);

   if (!LaunchChildProcess(cmdLine))
	  return false;

   if (!ReadOutput(Output))
	  return false;

   return true;
}
//---------------------------------------------------------------------------

bool __fastcall TGPGConsole::Run(AnsiString cmdLine, char *Passphrase, TStream *Output)
{
   bstatus = cmdLine.Pos("%stat%");
   battrib = cmdLine.Pos("%attr%");
   bpasswd = cmdLine.Pos("%pass%");

   if (!CreateHandles())
	  return false;

   SetupFileDescriptors(cmdLine);
   PrepareStartupInfo();

   if (!LaunchChildProcess(cmdLine))
	  return false;

   if (!SendPassphrase(Passphrase))
	  return false;

   if (!ReadOutput(Output))
	  return false;

   return true;
}
//---------------------------------------------------------------------------

bool __fastcall TGPGConsole::Run(AnsiString cmdLine, TStream *Output)
{
   bstatus = cmdLine.Pos("%stat%");
   battrib = cmdLine.Pos("%attr%");

   if (!CreateHandles())
	  return false;

   SetupFileDescriptors(cmdLine);
   PrepareStartupInfo();

   if (!LaunchChildProcess(cmdLine))
	  return false;

   if (!ReadOutput(Output))
	  return false;

   return true;
}
//---------------------------------------------------------------------------

bool __fastcall TGPGConsole::SendInput(TStrings *input)
{
   DWORD byteswrote;
   if (input)
   {
	  int i = 0;
	  while (i < input->Count)
	  {
		 AnsiString line = input->Strings[i] + "\r\n";
		 if (!WriteFile(hInputWrite, line.c_str(), line.Length(), &byteswrote, NULL))
		 {
			ShowError("Error sending the data to GPG: \n");
			return false;
		 }
		 i++;
	  }
	  CloseHandle(hInputWrite);
	  hInputWrite = NULL;
   }

   return true;
}
//---------------------------------------------------------------------------

bool __fastcall TGPGConsole::SendInput(const AnsiString &input)
{
   DWORD byteswrote;
   if (!WriteFile(hInputWrite, input.c_str(), input.Length(), &byteswrote, NULL))
   {
	  ShowError("Error sending the data to GPG: \n");
	  return false;
   }
   CloseHandle(hInputWrite);
   hInputWrite = NULL;

   return true;
}
//---------------------------------------------------------------------------

bool __fastcall TGPGConsole::SendInput(TStream *input)
{
   char buf[4096];
   DWORD bufsize = sizeof(buf), byteswrote, bytesread;// = bufsize;

   if (input)
   {
      assert(input->Size > 0);
      input->Seek(0, soFromBeginning);

      while ((bytesread = input->Read(buf, bufsize)) > 0)
      {
         if (!WriteFile(hInputWrite, buf, bytesread, &byteswrote, NULL))
		 {
            ShowError("Error sending the data to GPG: \n");
            return false;
         }
      }

      CloseHandle(hInputWrite);
      hInputWrite = NULL;
   }

   return true;
}
//---------------------------------------------------------------------------
// exitwait: waitforsingleobject: WAIT_ABANDON, WAIT_OBJECT_0, WAIT_TIMEOUT
// exitcode: this is the exec'd programs exit code

bool __fastcall TGPGConsole::ReadOutput(TStrings *output)
{
   DWORD bt;
   int ms = 100;
   time_t marktime = time(NULL);

   while (time(NULL) < marktime + 30)
   {
      exitwait = WaitForSingleObject(processinfo.hProcess, ms);
      if (exitwait == WAIT_FAILED)
      {
		 ShowError("Error occured while waiting for GPG operation to finish: \n");
         return false;
      }

      while (PeekNamedPipe(hOutputRead, NULL, NULL, NULL, &bt, NULL) && bt)
      {
         if (!GetWaiting(hOutputRead, output))
         {
            ShowError("Error reading GPG output: \n");
            return false;
		 }
	  }

      while (PeekNamedPipe(hErrorRead, NULL, NULL, NULL, &bt, NULL) && bt)
      {
         if (!GetWaiting(hErrorRead, error))
         {
            ShowError("Error reading GPG error output: \n");
            return false;
         }
      }

	  if (bstatus)
      while (PeekNamedPipe(hStatRead, NULL, NULL, NULL, &bt, NULL) && bt)
      {
         if (!GetWaiting(hStatRead, status))
         {
            ShowError("Error reading GPG status output: \n");
            return false;
         }
      }

      if (battrib)
      while (PeekNamedPipe(hAttrRead, NULL, NULL, NULL, &bt, NULL) && bt)
      {
         if (!GetWaiting(hAttrRead, attrib))
         {
			ShowError("Error reading GPG logger output: \n");
            return false;
         }
      }

      if (!GetExitCodeProcess(processinfo.hProcess, &exitcode))
      {
         ShowError("Error getting GPG exit code: \n");
         return false;
      }

      if (exitwait != WAIT_TIMEOUT)
         break;

      marktime = time(NULL);
   }

   /*if (exitcode)
   {
	  // 1 = GPG processing error
	  // 2 = GPG command error
	  return false;
   }*/

   return true;
}
//---------------------------------------------------------------------------
// exitwait: waitforsingleobject: WAIT_ABANDON, WAIT_OBJECT_0, WAIT_TIMEOUT
// exitcode: this is the exec'd programs exit code

bool __fastcall TGPGConsole::ReadOutput(AnsiString &output)
{
   DWORD bt;
   int ms = 100;
   time_t marktime = time(NULL);

   while (time(NULL) < marktime + 30)
   {
	  exitwait = WaitForSingleObject(processinfo.hProcess, ms);
	  if (exitwait == WAIT_FAILED)
	  {
		 ShowError("Error occured while waiting for GPG operation to finish: \n");
		 return false;
	  }

	  while (PeekNamedPipe(hOutputRead, NULL, NULL, NULL, &bt, NULL) && bt)
	  {
		 if (!GetWaiting(hOutputRead, output))
		 {
			ShowError("Error reading GPG output: \n");
			return false;
		 }
	  }

	  while (PeekNamedPipe(hErrorRead, NULL, NULL, NULL, &bt, NULL) && bt)
	  {
		 if (!GetWaiting(hErrorRead, error))
		 {
			ShowError("Error reading GPG error output: \n");
			return false;
		 }
	  }

	  if (bstatus)
	  while (PeekNamedPipe(hStatRead, NULL, NULL, NULL, &bt, NULL) && bt)
	  {
		 if (!GetWaiting(hStatRead, status))
		 {
			ShowError("Error reading GPG status output: \n");
			return false;
		 }
	  }

	  if (battrib)
	  while (PeekNamedPipe(hAttrRead, NULL, NULL, NULL, &bt, NULL) && bt)
	  {
		 if (!GetWaiting(hAttrRead, attrib))
		 {
			ShowError("Error reading GPG logger output: \n");
			return false;
		 }
	  }

	  if (!GetExitCodeProcess(processinfo.hProcess, &exitcode))
	  {
		 ShowError("Error getting GPG exit code: \n");
		 return false;
	  }

	  if (exitwait != WAIT_TIMEOUT)
		 break;

	  marktime = time(NULL);
   }

   /*if (exitcode)
   {
	  // 1 = GPG processing error
	  // 2 = GPG command error
	  return false;
   }*/

   return true;
}
//---------------------------------------------------------------------------
// exitwait: waitforsingleobject: WAIT_ABANDON, WAIT_OBJECT_0, WAIT_TIMEOUT
// exitcode: this is the exec'd programs exit code

bool __fastcall TGPGConsole::ReadOutput(TStream *output)
{
   DWORD bt;
   int ms = 100;
   time_t marktime = time(NULL);

   while (time(NULL) < marktime + 30)
   {
	  exitwait = WaitForSingleObject(processinfo.hProcess, ms);
	  if (exitwait == WAIT_FAILED)
	  {
		 ShowError("Error occured while waiting for GPG operation to finish: \n");
		 return false;
	  }

	  while (PeekNamedPipe(hOutputRead, NULL, NULL, NULL, &bt, NULL) && bt)
	  {
		 if (!GetWaiting(hOutputRead, output))
		 {
			ShowError("Error reading GPG output: \n");
			return false;
		 }
	  }

	  while (PeekNamedPipe(hErrorRead, NULL, NULL, NULL, &bt, NULL) && bt)
	  {
		 if (!GetWaiting(hErrorRead, error))
		 {
			ShowError("Error reading GPG error output: \n");
			return false;
		 }
	  }

	  if (status)
	  while (PeekNamedPipe(hStatRead, NULL, NULL, NULL, &bt, NULL) && bt)
	  {
		 if (!GetWaiting(hStatRead, status))
		 {
			ShowError("Error reading GPG status output: \n");
			return false;
		 }
	  }

	  if (attrib)
	  while (PeekNamedPipe(hAttrRead, NULL, NULL, NULL, &bt, NULL) && bt)
	  {
		 if (!GetWaiting(hAttrRead, attrib))
		 {
			ShowError("Error reading GPG logger output: \n");
			return false;
		 }
	  }

	  if (!GetExitCodeProcess(processinfo.hProcess, &exitcode))
	  {
		 ShowError("Error getting GPG exit code: \n");
		 return false;
	  }

	  if (exitwait != WAIT_TIMEOUT)
		 break;

	  marktime = time(NULL);
   }

   /*if (exitcode)
   {
	  // 1 = GPG processing error
	  // 2 = GPG command error
	  return false;
   }*/
   return true;
}
//---------------------------------------------------------------------------

bool __fastcall TGPGConsole::GetWaiting(HANDLE hReadHandle, TStrings *output)
{
   static char buf[1024];
   static AnsiString incomingline;
   unsigned long bytesread, bufsize = sizeof(buf);

   if (!ReadFile(hReadHandle, buf, bufsize, &bytesread, NULL))
      return false;

   if (bytesread > 0)
   {
      for (unsigned i = 0; i < bytesread; i++)
	  {
         if (buf[i] != '\0' && buf[i] != '\n' && buf[i] != '\r')
            incomingline += String(buf[i]);
         if (buf[i] == '\n')
         {
            output->Add(incomingline);
            incomingline.SetLength(0);
         }
      }
   }
   return true;
}
//---------------------------------------------------------------------------

bool __fastcall TGPGConsole::GetWaiting(HANDLE hReadHandle, TStream *output)
{
   static char buf[32768];
   unsigned long bytesread, bufsize = sizeof(buf);

   if (!ReadFile(hReadHandle, buf, bufsize, &bytesread, NULL))
      return false;

   if (bytesread > 0)
   {
      output->Write(buf, bytesread);
   }

   return true;
}
//---------------------------------------------------------------------------

bool __fastcall TGPGConsole::GetWaiting(HANDLE hReadHandle, AnsiString &output)
{
   static char buf[32768] = { '\0' };
   unsigned long bytesread, bufsize = sizeof(buf);

   output = "";
   if (!ReadFile(hReadHandle, buf, bufsize, &bytesread, NULL))
      return false;

   if (bytesread > 0)
   {
      output = buf;
   }

   return true;
}
//---------------------------------------------------------------------------

void __fastcall TGPGConsole::ShowError(const AnsiString &errormsg)
{
   MsgBox(Owner, errormsg + GetLastErrorStr(), MT_ERROR, MB_OK);
}

//---------------------------------------------------------------------------
// GPGerrDlg
//---------------------------------------------------------------------------

void __fastcall ShowGPGError(TForm *Owner, TStrings *Error, TStrings *Status)
{
   TGPGerrDlg *errdlg = new TGPGerrDlg(Owner, Error, Status);
   errdlg->ShowModal();
   delete errdlg;
}
//---------------------------------------------------------------------------

void __fastcall ShowGPGError(TForm *Owner, TStrings *ErrorStatus)
{
   bool inerror = true;
   TStringList *error = new TStringList();
   TStringList *status = new TStringList();
   for (int i = 0; i < ErrorStatus->Count; i++)
   {
	  AnsiString line = ErrorStatus->Strings[i];
	  if (inerror && line == "-----")
	  {
		 inerror = false;
		 continue;
	  }
	  if (!inerror)
		 status->Add(line);
	  else
		 error->Add(line);
   }
   ShowGPGError(Owner, error, status);
   delete status;
   delete error;
}
//---------------------------------------------------------------------------

__fastcall TGPGerrDlg::TGPGerrDlg(TComponent* Owner)
        : TForm(Owner)
{
   ErrorBox->Clear();
   TIniFile *ini = new TIniFile(AppIni());
   AssignFont(Font, ini->ReadString("settings", "dialogfont", OptsDefDialogFont));
   AssignFont(ErrorBox->Font, ini->ReadString("settings", "editorfont",
      OptsDefEditorFont));
   delete ini;
}
//---------------------------------------------------------------------------

__fastcall TGPGerrDlg::TGPGerrDlg(TComponent* Owner, TStrings *GPGerror,
   TStrings *GPGstatus) : TForm(Owner)
{
   PopupParent = (TForm*)Owner;
   ErrorBox->Clear();
   ErrorBox->Lines->Text = GPGerror->Text;
   TIniFile *ini = new TIniFile(AppIni());
   AssignFont(Font, ini->ReadString("settings", "dialogfont", OptsDefDialogFont));
   AssignFont(ErrorBox->Font, ini->ReadString("settings", "editorfont",
	  OptsDefEditorFont));
   delete ini;
   
   status = new TStringList();
   status->Text = GPGstatus->Text;
   error = new TStringList();
   error->Text = GPGerror->Text;

   StatusBtn->Visible = GPGstatus->Count;
}
//---------------------------------------------------------------------------

__fastcall TGPGerrDlg::~TGPGerrDlg()
{
   delete status;
}
//---------------------------------------------------------------------------

void __fastcall TGPGerrDlg::FormShow(TObject *Sender)
{
   HWND DC;
   TCanvas *canvas = new TCanvas();
   canvas->Handle = GetDeviceContext(DC);
   canvas->Font->Assign(ErrorBox->Font);

   int high = canvas->TextHeight("yY") * (ErrorBox->Lines->Count + 2);
   ClientHeight = high + (ClientHeight - ErrorBox->Height);
   int w, wide = 0;
   for (int i = 0; i < ErrorBox->Lines->Count; i++)
   {
      w = canvas->TextWidth(ErrorBox->Lines->Strings[i]);
      if (w > wide)
         wide = w;
   }
   wide = wide + (ClientWidth - ErrorBox->Width) + 40;
   ClientWidth = wide < 400 ? 400 : wide;

   OkBtn->Left = (Width / 2) - (OkBtn->Width / 2);

   delete canvas;
}

//---------------------------------------------------------------------------

void __fastcall TGPGerrDlg::StatusBtnClick(TObject *Sender)
{
   if (StatusBtn->Caption == "More...")
   {
      ErrorBox->Lines->Add("");
	  ErrorBox->Lines->Add("-------------------");
      ErrorBox->Lines->Add("");
      ErrorBox->Lines->AddStrings(status);
      StatusBtn->Caption = "Less...";
   }
   else
   {
      ErrorBox->Text = error->Text;
      StatusBtn->Caption = "More...";
   }
   FormShow(NULL);
}
//---------------------------------------------------------------------------

void __fastcall TGPGerrDlg::PopupMenu1Popup(TObject *Sender)
{
   Copy1->Enabled = ErrorBox->SelLength >= 0;        
}
//---------------------------------------------------------------------------

void __fastcall TGPGerrDlg::Copy1Click(TObject *Sender)
{
   ErrorBox->CopyToClipboard();        
}
//---------------------------------------------------------------------------

