/*
*  gpg.h 
*
*  Copyright (C) 2010 Richard Christman 
*
*  This file is part of QuickSilver Lite.
*
*  QuickSilver Lite is free software; you can redistribute it and/or modify 
*  it under the terms of the GNU General Public License as published by the 
*  Free Software Foundation; either version 3 of the License, or (at your 
*  option) any later version.
* 
*  This program is distributed in the hope that it will be useful, but 
*  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
*  or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License 
*  for more details.
*
*  You should have received a copy of the GNU General Public License along 
*  with this program; if not, see <http://www.gnu.org/licenses>.
* 
*  Additional permission under GNU GPL version 3 section 7
* 
*  If you modify QuickSilver Lite, or any covered work, by linking or 
*  combining it with CodeGear VCL (or a modified version of that library), 
*  containing parts covered by the terms of CodeGear's license, the licensors 
*  of QuickSilver Lite grant you additional permission to convey the 
*  resulting work.
*
*/

//---------------------------------------------------------------------------
#ifndef gpgH
#define gpgH
//---------------------------------------------------------------------------
#include <Classes.hpp>
#include <Controls.hpp>
#include <StdCtrls.hpp>
#include <Forms.hpp>
#include <ComCtrls.hpp>
#include "qsdbs.h"
#include <Menus.hpp>
//---------------------------------------------------------------------------
AnsiString __fastcall GPGExe();
AnsiString __fastcall GPGDir();
AnsiString __fastcall GPGDataDir();
AnsiString __fastcall GPGConf();
AnsiString __fastcall GPGIconvDll();
AnsiString __fastcall GPGTempFileBase();
AnsiString __fastcall GPGTempFile();
AnsiString __fastcall GPGDefaultKeyID();
bool __fastcall GPGIsPGPMessage(TStrings *Message);
bool __fastcall GPGIsPGPSignedMessage(TStrings *Message);
bool __fastcall GPGHasPGPSignatureVerification(TStrings *Message);

AnsiString __fastcall GPGGetUserID(const AnsiString &keyid);
// main function completely processes multple PGP headers in a message
bool __fastcall GPGProcessMessage(TForm *Owner, TStrings *Message);
// main decrypt/verify function processes PGP message
bool __fastcall GPGProcessMessageDecryption(TStrings *Message);
//bool __fastcall GPGDecryptNymMessage(TForm *Owner, TStrings *Message);
// these functions are called whether pubkey is enabled or not
bool __fastcall GPGEncryptDatabase(TQSDB *DB, AnsiString &Passphrase);
bool __fastcall GPGDecryptDatabase(TQSDB *DB, AnsiString &Passphrase);
bool __fastcall GPGLoadMasterPassword(TQSDB *DB, AnsiString &Passphrase);
bool __fastcall GPGSaveMasterPassword(TQSDB *DB, AnsiString &Passphrase);
bool __fastcall GPGIsFileEncrypted(const AnsiString &FileName);
//bool __fastcall GPGGetEncryptionKeys(TStrings *Message, TStrings *Keys);
//bool __fastcall GPGVerifyClearSigned(TStrings *Message);
bool __fastcall GPGListSecretKeysWithColonsRingName(const AnsiString &Ringname, TStrings *Keys);
bool __fastcall GPGListSecretKeysWithColonsNE(const AnsiString &keyids, TStrings *keys);
bool __fastcall GPGEncryptStringToStream(AnsiString &Passphrase, AnsiString &Input, TStream *Output);
bool __fastcall GPGDecryptStreamToString(AnsiString &Passphrase, TStream *Input, AnsiString &Output);
bool __fastcall GPGGenRandomToFile(int count, int quality, const AnsiString &filename);
AnsiString __fastcall GPGVersion();
bool __fastcall SetGPGConfValues();


AnsiString __fastcall GPGGetPublicKeyStr(const AnsiString &Keyid);
bool __fastcall GPGListPublicKeysWithColonsNE(const AnsiString &KeyIDs, TStrings *Keys); // NE = no errors reported

// this class handles public key crypto function for GPGProcessMessage
class TGPG : public TObject
{
   TForm *Owner;
   bool __fastcall GetEnabled();
   AnsiString __fastcall GetSecring();
   AnsiString __fastcall GetPubring();
   AnsiString __fastcall GetDefKey();
   bool __fastcall GetEncDef();
   bool __fastcall GetMaskVer();
   void __fastcall erasefile(const AnsiString &filename);
   void __fastcall ShowStatus(AnsiString statusmsg);
   //bool __fastcall SetConfValues();

public:
   __property bool enabled                        = { read=GetEnabled };
   __property AnsiString secring                  = { read=GetSecring };
   __property AnsiString pubring                  = { read=GetPubring };
   __property AnsiString defaultkey               = { read=GetDefKey  };
   __property bool encrypttodef                   = { read=GetEncDef  };
   __property bool maskversion                    = { read=GetMaskVer };
   TStringList *cfg, *error, *status;
   __fastcall TGPG();
   __fastcall virtual ~TGPG();
   bool __fastcall SignMessage(TStrings *Message, const AnsiString &SignKey,
	  AnsiString &Passphrase);
   bool __fastcall VerifyClearSigned(TStrings *Message);
   bool __fastcall GetEncryptionKeys(TStrings *Message, TStrings *Keys);
   bool __fastcall DecryptMessage(TStrings *Message, AnsiString &Passphrase);
   bool __fastcall EncryptMessage(TStrings *Message, const AnsiString &CommaEncKeys);
   bool __fastcall EncryptSignMessage(TStrings *Message, const AnsiString &CommaEncKeys,
	  const AnsiString &SignKey, AnsiString &Passphrase);
   bool __fastcall CEncryptMessage(TStrings *Message, AnsiString &Passphrase);
   int __fastcall ListSecretKeysWithColons(TStrings *keys);
   int __fastcall ListSecretKeysWithColons(TStrings *keys, TStrings *keyids);
   int __fastcall ListSecretKeysWithColons(TStrings *keys, const AnsiString &commakeyids);

   int __fastcall ListPublicKeysWithColons(TStrings *keys);
   int __fastcall ListPublicKeysWithColons(TStrings *keys, TStrings *keyids);
   int __fastcall ListPublicKeysWithColons(TStrings *keys, const AnsiString &commakeyids);
   //void __fastcall AnalysePGPHeader(const AnsiString &header, bool &enc, bool &sig,
   //	  AnsiString &enckey, AnsiString &sigkey);
   bool __fastcall CheckInitSuccess();

   AnsiString __fastcall KeyType(const AnsiString &keystr);
   AnsiString __fastcall KeyValid(const AnsiString &keystr);
   AnsiString __fastcall KeyID(const AnsiString &keystr);
   AnsiString __fastcall KeyID8(const AnsiString &keystr);
   AnsiString __fastcall KeyUID(const AnsiString &keystr);
};

// console class for all GPG calls
class TGPGConsole : public TObject
{
protected:
   STARTUPINFO startupinfo;
   PROCESS_INFORMATION processinfo;
   HANDLE hOutputRead, hOutputWrite, hInputRead, hInputWrite, hPassRead,
	  hPassWrite, hErrorRead, hErrorWrite, hAttrRead, hAttrWrite,
      hStatRead, hStatWrite;
   bool bstatus, battrib, bpasswd;
   DWORD exitwait;
   TForm *Owner;

   bool __fastcall CreateHandles();
   void __fastcall DestroyHandles();
   void __fastcall PrepareStartupInfo();
   void __fastcall SetupFileDescriptors(AnsiString &cmdline);
   bool __fastcall LaunchChildProcess(const AnsiString &cmdline);
   bool __fastcall ProcessExists();
   bool __fastcall SendPassphrase(char *passphrase);
   bool __fastcall SendInput(const AnsiString &input);
   bool __fastcall SendInput(TStrings *input);
   bool __fastcall SendInput(TStream *input);
   bool __fastcall ReadOutput(TStrings *output);
   bool __fastcall ReadOutput(TStream *output);
   bool __fastcall ReadOutput(AnsiString &output);
   bool __fastcall GetWaiting(HANDLE hReadHandle, TStream *output);
   bool __fastcall GetWaiting(HANDLE hReadHandle, TStrings *output);
   bool __fastcall GetWaiting(HANDLE hReadHandle, AnsiString &output);
   void __fastcall ShowError(const AnsiString &errormsg);
public:
   DWORD exitcode;
   TStringList *error, *status, *attrib;
   AnsiString filename;

   __fastcall TGPGConsole(TForm *Owner);
   __fastcall virtual ~TGPGConsole();
   bool __fastcall Run(AnsiString cmdLine, AnsiString &Passphrase, const AnsiString &Input, TStream *Output);
   bool __fastcall Run(AnsiString cmdLine, AnsiString &Passphrase, TStream *Input, AnsiString &Output);
   bool __fastcall Run(AnsiString cmdLine, char *Passphrase, TStrings *Input, TStrings *Output);
   bool __fastcall Run(AnsiString cmdLine, char *Passphrase, TStream *Input, TStrings *Output);
   bool __fastcall Run(AnsiString cmdLine, char *Passphrase, TStream *Output);
   bool __fastcall Run(AnsiString cmdLine, TStrings *Output);
   bool __fastcall Run(AnsiString cmdLine, TStream *Output);
};

class TGPGerrDlg : public TForm
{
__published:	// IDE-managed Components
        TLabel *Label1;
        TRichEdit *ErrorBox;
        TButton *StatusBtn;
        TButton *OkBtn;
        TPopupMenu *PopupMenu1;
        TMenuItem *Copy1;
        void __fastcall FormShow(TObject *Sender);
        void __fastcall StatusBtnClick(TObject *Sender);
		void __fastcall PopupMenu1Popup(TObject *Sender);
		void __fastcall Copy1Click(TObject *Sender);
private:	// User declarations
public:		// User declarations
		TStrings *status, *error;
		__fastcall virtual TGPGerrDlg(TComponent* Owner);
		__fastcall virtual TGPGerrDlg(TComponent* Owner, TStrings *GPGerror,
				TStrings *GPGstatus);
		__fastcall virtual ~TGPGerrDlg();
};


const AnsiString BeginREMAILERMessage = "-----BEGIN REMAILER MESSAGE-----";
const AnsiString BeginPGPMessage = "-----BEGIN PGP MESSAGE-----";
const AnsiString EndPGPMessage = "-----END PGP MESSAGE-----";
const AnsiString BeginPGPSignedMessage = "-----BEGIN PGP SIGNED MESSAGE-----";
const AnsiString BeginPGPSignature = "-----BEGIN PGP SIGNATURE-----";
const AnsiString EndPGPSignature = "-----END PGP SIGNATURE-----";
const AnsiString BeginPGPDecryptedVerifiedMessage = "*** BEGIN PGP DECRYPTED/VERIFIED MESSAGE ***";
const AnsiString EndPGPDecryptedVerifiedMessage = "*** END PGP DECRYPTED/VERIFIED MESSAGE ***";
const AnsiString BeginPGPDecryptedMessage = "*** BEGIN PGP DECRYPTED MESSAGE ***";
const AnsiString EndPGPDecryptedMessage = "*** END PGP DECRYPTED MESSAGE ***";
const AnsiString PGPSignatureVerification = "*** PGP SIGNATURE VERIFICATION ***";

const AnsiString GPGCharset = " --display-charset iso-8859-1";
// encryption/decryption
const AnsiString GPGCEncryptCmd =
   GPGCharset+" --status-fd %stat% --batch --no-tty --passphrase-fd %pass% --yes -c -a";
const AnsiString GPGCEncryptToFileCmd =
   GPGCharset+" --status-fd %stat% --batch --no-tty --passphrase-fd %pass% --yes -c -a --output %filename%";
const AnsiString GPGCEncryptCmdNA =
   GPGCharset+" --status-fd %stat% --batch --no-tty --passphrase-fd %pass% --yes -c";
const AnsiString GPGCEncryptToFileCmdNA =
   GPGCharset+" --status-fd %stat% --batch --no-tty --passphrase-fd %pass% --yes -c --output %filename%";
const AnsiString GPGEncryptSignToFileCmd =
   GPGCharset+" --status-fd %stat% --batch --no-tty --passphrase-fd %pass% -u %signkey% --yes --sign -e -r %recipkey% -a -t --output %filename%";
const AnsiString GPGEncryptSignCmd =
   GPGCharset+" --status-fd %stat% --batch --no-tty --passphrase-fd %pass% -u %signkey% --yes --sign -e -r %recipkey% -a -t";
const AnsiString GPGClearSignToFileCmd =
   GPGCharset+" --status-fd %stat% --batch --no-tty --passphrase-fd %pass% -u %signkey% -t --clearsign --output %filename%";
const AnsiString GPGClearSignCmd =
   GPGCharset+" --status-fd %stat% --batch --no-tty --passphrase-fd %pass% -u %signkey% -t --clearsign";
const AnsiString GPGEncryptToFileCmd =
   GPGCharset+" --status-fd %stat% --batch --no-tty -e -r %recipkey% --yes -a -t --output %filename%";
const AnsiString GPGEncryptCmd =
   GPGCharset+" --status-fd %stat% --batch --no-tty -e -r %recipkey% --yes -a -t";
const AnsiString GPGDecryptFromFileCmd =
   GPGCharset+" --status-fd %stat% --batch --no-tty --passphrase-fd %pass% -d --verify-options show-primary-uid-only %filename%";
const AnsiString GPGDecryptCmd =
   GPGCharset+" --status-fd %stat% --batch --no-tty --passphrase-fd %pass% -d --verify-options show-primary-uid-only";
const AnsiString GPGEncryptStringToStreamCmd =
   GPGCharset+" --status-fd %stat% --batch --no-tty --passphrase-fd %pass% -c";
const AnsiString GPGDecryptStreamToStringCmd =
   GPGCharset+" --status-fd %stat% --batch --no-tty --passphrase-fd %pass% -d";
const AnsiString GPGVerifyClearSignedCmd =
   GPGCharset+" --status-fd %stat% -d --verify-options show-primary-uid-only";
// keys/keyrings
const AnsiString GPGListSecretKeysWithColonsCmd =
   GPGCharset+" --status-fd %stat% --list-secret-keys --with-colons";
const AnsiString GPGListSecretKeysWithColonsRingNameCmd =
   GPGCharset+" --status-fd %stat% --secret-keyring %secring% --list-secret-keys --with-colons";
const AnsiString GPGListPublicKeysWithColonsCmd =
   GPGCharset+" --status-fd %stat% --list-public-keys --with-colons";
const AnsiString GPGListPublicKeysWithColonsRingNameCmd =
   GPGCharset+" --status-fd %stat% --keyring %pubring% --list-public-keys --with-colons";
const AnsiString GPGListEncryptionKeysFromFileCmd =
   GPGCharset+" --status-fd %stat% --list-only %filename%";
const AnsiString GPGListEncryptionKeysCmd =
   GPGCharset+" --status-fd %stat% --list-only";
const AnsiString GPGVersionCmd =
   GPGCharset+" --status-fd %stat% --version";
const AnsiString GPGGenRandomCmd =
   GPGCharset+" --status-fd %stat% --gen-random %quality% %count%";

//const bool GPGMaskVersion = true;
const bool GPGWipeEncryptedDB = false;
const int GPGMinPassLength = 6;
const int GPGCacheMinutes = 0;


// ini items of [gpg] section for replacement commandlines, etc
const AnsiString GPGcencrypttofile = "cencrypttofile";
const AnsiString GPGcencrypt = "cencrypt";
const AnsiString GPGcencryptna = "cencryptna";
const AnsiString GPGcencrypttofilena = "cencrypttofilena";
const AnsiString GPGencryptsigntofile = "encryptsigntofile";
const AnsiString GPGencryptsign = "encryptsign";
const AnsiString GPGclearsigntofile = "clearsigntofile";
const AnsiString GPGclearsign = "clearsign";
const AnsiString GPGencrypttofile = "encrypttofile";
const AnsiString GPGencrypt = "encrypt";
const AnsiString GPGdecryptfromfile = "decryptfromfile";
const AnsiString GPGdecrypt = "decrypt";
const AnsiString GPGlistsecretkeyswithcolons = "listsecretkeyswithcolons";
const AnsiString GPGlistsecretkeyswithcolonsringname = "listsecretkeyswithcolonsringnname";
const AnsiString GPGlistpublickeyswithcolons = "listpublickeyswithcolons";
const AnsiString GPGlistpublickeyswithcolonsringname = "listpublickeyswithcolonsringname";
const AnsiString GPGlistencryptionkeysfromfile = "listencryptionkeysfromfile";
const AnsiString GPGlistencryptionkeys = "listencryptionkeys";
const AnsiString GPGencryptstringtostream = "encryptstringtostream";
const AnsiString GPGdecryptstreamtostring = "decryptstreamtostring";
//const AnsiString GPGmaskgpgversion = "maskgpgversion";
const AnsiString GPGwipeencrypteddb = "wipeencrypteddb";
const AnsiString GPGminpasslength = "minpassphraselength";
const AnsiString GPGcacheminutes = "cacheminutes";
const AnsiString GPGgenrandom = "genrandom";
const AnsiString GPGversion = "version";
const AnsiString GPGverifyclearsigned = "verifyclear";

// listkey fields for gpglistkeyswithcolons
const int lkfKeyType = 0;
const int lkfKeyValid = 1;
const int lkfKeyBits = 2;
const int lkfKeyAlgo = 3;
const int lkfKeyID = 4;
const int lkfKeyCreated = 5;
const int lkfKeyExpires = 6;
const int lkfKeyUID = 9;

//
// ini entry to turn on wiping of encrypted db when saving new db
// [advanced]
// wipeencrypteddb=1
//
// ini entry to set minimum password length
// [advanced]
// minpassphraselength=6
//

void __fastcall AnalysePGPHeader(const AnsiString &header, bool &enc, bool &sig,
   AnsiString &enckey, AnsiString &sigkey);
AnsiString __fastcall getcommandline(const AnsiString &item, const AnsiString defvalue);
void __fastcall getheaders(TStrings *headers, TStrings *message);
void __fastcall putheaders(TStrings *message, TStrings *headers);
void __fastcall assemblemessage(TStrings *message, TStream *output,
   TStrings *headers, bool maskversion);
void __fastcall FormatSigVerification(TStrings *verified);
void __fastcall trimbodybottom(TStrings *msgbody);
void __fastcall trimbodyright(TStrings *msgbody);
bool __fastcall setfield(const AnsiString &field, const AnsiString &value,
   AnsiString &line);
void __fastcall ShowGPGError(TForm *Owner, TStrings *ErrorStatus);
void __fastcall ShowGPGError(TForm *Owner, TStrings *Error, TStrings *Status);
void __fastcall DecodeQuotedPrintableInArmor(TStrings *msg);
AnsiString __fastcall recipkeystr(AnsiString CommaKeyIDs);
//---------------------------------------------------------------------------
extern PACKAGE TGPGerrDlg *GPGerrDlg;
extern AnsiString gpgdatadir;
extern AnsiString gpgdir;
//---------------------------------------------------------------------------
#endif
