/*
*  gpg2.cpp
*
*  Copyright (C) 2010 Richard Christman
*
*  This file is part of QuickSilver Lite.
*
*  QuickSilver Lite is free software; you can redistribute it and/or modify
*  it under the terms of the GNU General Public License as published by the
*  Free Software Foundation; either version 3 of the License, or (at your
*  option) any later version.
*
*  This program is distributed in the hope that it will be useful, but
*  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
*  or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
*  for more details.
*
*  You should have received a copy of the GNU General Public License along
*  with this program; if not, see <http://www.gnu.org/licenses>.
*
*  Additional permission under GNU GPL version 3 section 7
*
*  If you modify QuickSilver Lite, or any covered work, by linking or
*  combining it with CodeGear VCL (or a modified version of that library),
*  containing parts covered by the terms of CodeGear's license, the licensors
*  of QuickSilver Lite grant you additional permission to convey the
*  resulting work.
*
*/

//---------------------------------------------------------------------------
#include <vcl.h>
#pragma hdrstop
#include "gpg2.h"
#include "gpg.h"
#include "global.h"
//---------------------------------------------------------------------------
#pragma package(smart_init)
const int bufsize = 32768;
//---------------------------------------------------------------------------

void __fastcall assemblemessagesig(TStrings *message, TStream *output,
   TStrings *headers, TStrings *usenetsig, bool maskversion)
{
   message->LoadFromStream(output);

   // mask gog version
   if (maskversion) // do the masking of the Version Line gpg emits
   {
	  for (int i = 0; i < message->Count; i++)
	  {
		 if (message->Strings[i].Pos("Version:") == 1)
		 {
			message->Strings[i] = "Version: N/A";
			break;
		 }
	  }
   }

   // message headers
   for (int i = 0; i < headers->Count; i++)
	  message->Insert(i, headers->Strings[i]);

   // usenet signature
   if (usenetsig->Count)
   {
	  message->Add("");
	  for (int i = 0; i < usenetsig->Count; i++)
		 message->Add(usenetsig->Strings[i]);
   }
}
//---------------------------------------------------------------------------

bool __fastcall handlemaskversion(TStrings *msg)
{
   bool masked = false;
   for (int i = 0; i < msg->Count; i++)
   {
	  if (msg->Strings[i].Pos("Version:") == 1)
	  {
		 msg->Strings[i] = "Version: N/A";
		 masked = true;
		 break;
	  }
   }
   return masked;
}
//---------------------------------------------------------------------------

void __fastcall getusenetsig(TStrings *usenetsig, TStrings *msg)
{
   int x = msg->IndexOf("-- ");
   if (x > 0)
   {
	  while (msg->Count > x)
	  {
		 usenetsig->Add(msg->Strings[x]);
		 msg->Delete(x);
	  }
   }
}
//---------------------------------------------------------------------------

bool __fastcall GPGClearSignMessage2(TForm *Owner, TStrings *Message,
   const AnsiString &SignKey, AnsiString &Passphrase, bool MaskVersion)
{
   Owner->Perform(WM_SHOWSTATUS, 0, (LPARAM)"PGP signing...");
   AnsiString line, cmdline = getcommandline(GPGclearsign, GPGClearSignCmd);

   TMemoryStream *input = new TMemoryStream();
   TMemoryStream *output = new TMemoryStream();
   TStringList *headers = new TStringList();
   TStringList *usenetsig = new TStringList();

   getusenetsig(usenetsig, Message);
   getheaders(headers, Message);
   trimbodybottom(Message);
   trimbodyright(Message);

   setfield("%signkey%", SignKey, cmdline);

   Message->SaveToStream(input);
   TGPGConsole4 *con = new TGPGConsole4(Owner);
   bool ok = con->Run(cmdline, Passphrase.c_str(), input, output);
   if (ok)
   {
	  ok = con->status->Text.Pos("SIG_CREATED");
	  if (!ok)
	  {
		 if (con->status->Text.Pos("BAD_PASSPHRASE"))
		 {
			Sleep(300);
			MsgBox(Owner, "Bad passphrase!   ", MT_INFO, MB_OK);
		 }
		 else
		 {
			Sleep(300);
			ShowGPGError(Owner, con->error, con->status);
		 }
	  }
	  assemblemessagesig(Message, output, headers, usenetsig, MaskVersion);
   }
   delete con;
   if (ok && (Message->IndexOf(BeginPGPSignedMessage) < 0 ||
			  Message->IndexOf(BeginPGPSignature) < 0))
   {
	   MsgBox(Owner,
		  "The signing appears to have failed! I can't find the   \n" +
		   BeginPGPSignedMessage + " and/or " + BeginPGPSignature + "   \n"
		  "markers!",
		   MT_ERROR, MB_OK);
	   ok = false;
   }

   delete usenetsig;
   delete headers;
   delete output;
   delete input;

   Owner->Perform(WM_SHOWSTATUS, 0, (LPARAM)"Done!");
   Snooze(200);
   return ok;
}
//---------------------------------------------------------------------------

bool __fastcall GPGCEncryptMessage2(TForm *Owner, TStrings *Message,
   AnsiString &Passphrase, bool MaskVersion)
{
   Owner->Perform(WM_SHOWSTATUS, 0, (LPARAM)"PGP encrypting...");
   AnsiString cmdline = getcommandline(GPGcencrypt, GPGCEncryptCmd);

   TMemoryStream *input = new TMemoryStream();
   TMemoryStream *output = new TMemoryStream();
   TStringList *headers = new TStringList();
   getheaders(headers, Message);
   trimbodybottom(Message);
   Message->SaveToStream(input);

   TGPGConsole4 *con = new TGPGConsole4(Owner);
   bool ok = con->Run(cmdline, Passphrase.c_str(), input, output);
   if (ok)
   {
	  ok = con->status->Text.Pos("END_ENCRYPTION") > 0;
	  if (ok)
	  {
		 Message->LoadFromStream(output);
		 if (MaskVersion)
			handlemaskversion(Message);
	  }
	  else
	  {
		 Snooze(300);
		 ShowGPGError(Owner, con->error, con->status);
	  }
   }
   delete con;
   if (ok && (Message->IndexOf(BeginPGPMessage) < 0 || Message->IndexOf(EndPGPMessage) < 0))
   {
	   MsgBox(Owner,
		  "The encryption appears to have failed! I can't find the   \n"
		  "-----Begin PGP Message----- and/or -----End PGP Message-----   \n"
		  "markers!",
		   MT_ERROR, MB_OK);
	   ok = false;
   }
   putheaders(Message, headers);
   delete headers;
   delete output;
   delete input;

   Owner->Perform(WM_SHOWSTATUS, 0, (LPARAM)"Done...");
   Snooze(200);
   return ok;
}
//---------------------------------------------------------------------------

bool __fastcall GPGEncryptMessage2(TForm *Owner, TStrings *Message,
   const AnsiString &CommaEncKeys, bool MaskVersion)
{
   Owner->Perform(WM_SHOWSTATUS, 0, (LPARAM)"PGP encrypting...");
   AnsiString cmdline = getcommandline(GPGencrypt, GPGEncryptCmd);

   TMemoryStream *input = new TMemoryStream();
   TMemoryStream *output = new TMemoryStream();
   TStringList *headers = new TStringList();
   getheaders(headers, Message);
   trimbodybottom(Message);
   Message->SaveToStream(input);

   setfield("%recipkey%", recipkeystr(CommaEncKeys), cmdline);

   TGPGConsole4 *con = new TGPGConsole4(Owner);
   bool ok = con->Run(cmdline, NULL, input, output);
   if (ok)
   {
	  ok = con->status->Text.Pos("END_ENCRYPTION") > 0;
	  if (ok)
	  {
		 Message->LoadFromStream(output);
		 if (MaskVersion)
			handlemaskversion(Message);
	  }
	  else
	  {
		 Snooze(300);
		 ShowGPGError(Owner, con->error, con->status);
	  }
   }
   delete con;
   if (ok && (Message->IndexOf(BeginPGPMessage) < 0 || Message->IndexOf(EndPGPMessage) < 0))
   {
	   MsgBox(Owner,
		  "The encryption appears to have failed! I can't find the   \n"
		  "-----Begin PGP Message----- and/or -----End PGP Message-----   \n"
		  "markers!",
		   MT_ERROR, MB_OK);
	   ok = false;
   }
   putheaders(Message, headers);
   delete headers;
   delete output;
   delete input;

   Owner->Perform(WM_SHOWSTATUS, 0, (LPARAM)"Done...");
   Snooze(200);
   return ok;
}
//---------------------------------------------------------------------------

bool __fastcall GPGEncryptSignMessage2(TForm *Owner, TStrings *Message,
   const AnsiString &CommaEncKeys, const AnsiString &SignKey,
   AnsiString &Passphrase, bool MaskVersion)
{
   Owner->Perform(WM_SHOWSTATUS, 0, (LPARAM)"PGP encrypting/signing...");
   AnsiString cmdline = getcommandline(GPGencryptsign, GPGEncryptSignCmd);

   TMemoryStream *input = new TMemoryStream();
   TMemoryStream *output = new TMemoryStream();
   TStringList *headers = new TStringList();
   getheaders(headers, Message);
   trimbodyright(Message);
   trimbodybottom(Message);
   Message->SaveToStream(input);

   setfield("%signkey%", SignKey, cmdline);
   setfield("%recipkey%", recipkeystr(CommaEncKeys), cmdline);

   TGPGConsole4 *con = new TGPGConsole4(Owner);
   bool ok = con->Run(cmdline, Passphrase.c_str(), input, output);
   //ZeroString(Passphrase);  passphrase is zeroed after saving in PGPCache
   //sm(con->error->Text+con->status->Text);
   if (ok)
   {
	  ok = con->status->Text.Pos("SIG_CREATED") > 0 && con->status->Text.Pos("END_ENCRYPTION") > 0;
	  if (ok)
	  {
		 Message->LoadFromStream(output);
		 if (MaskVersion)
			handlemaskversion(Message);
	  }
	  else
	  {
		 Snooze(300);
		 ShowGPGError(Owner, con->error, con->status);
	  }
   }
   delete con;
   if (ok && (Message->IndexOf(BeginPGPMessage) < 0 || Message->IndexOf(EndPGPMessage) < 0))
   {
	   MsgBox(Owner,
		  "The encryption appears to have failed! I can't find the   \n"
		  "-----Begin PGP Message----- and/or -----End PGP Message-----   \n"
		  "markers!",
		   MT_ERROR, MB_OK);
	   ok = false;
   }
   putheaders(Message, headers);
   delete headers;
   delete output;
   delete input;

   Owner->Perform(WM_SHOWSTATUS, 0, (LPARAM)"Done!...");
   Snooze(200);
   return ok;
}
//---------------------------------------------------------------------------

void __fastcall FormatSigVerification(TStrings *rawreport, TStrings *sig)
{
   assert(rawreport->Count);
   while (rawreport->Count && !rawreport->Strings[0].Pos("Signature made"))
	  rawreport->Delete(0);
   if (!rawreport->Count)
	  return;

   sig->Add("*** PGP SIGNATURE VERIFICATION ***");
   for (int i = 0; i < rawreport->Count; i++)
   {
	  AnsiString line = rawreport->Strings[i];
	  if (line.Pos("gpg:") == 1)
		 sig->Add("*** " + line.SubString(6, 128));
   }
   sig->Add("*** Verified " + FormatDateTime("MM/DD/YY HH:NN:SS", Now()));
}
//---------------------------------------------------------------------------

int __fastcall getpgpblock(TStrings *msg, TStrings *blk)
{
   blk->Clear();
   int x = msg->IndexOf(BeginPGPMessage);
   if (x < 0)
	  x = msg->IndexOf(BeginPGPSignedMessage);
   assert(x >= 0);

   while (msg->Count)
   {
	  AnsiString line = msg->Strings[x];
	  msg->Delete(x);
	  if (line == "**")
		 continue;

	  blk->Add(line);
	  if (line == EndPGPMessage || line == EndPGPSignature)
		 break;
   }
   return x;
}
//---------------------------------------------------------------------------

void __fastcall putpgpblock(TStrings *msg, TStrings *blk, int x)
{
   assert(x <= msg->Count);
   TrimNulls(blk);
   for (int i = 0; i < blk->Count; i++)
   {
	  msg->Insert(x + i, blk->Strings[i]);
   }
}
//---------------------------------------------------------------------------

bool __fastcall GPGVerifyClearSigned2(TForm *Owner, TStrings *Message)
{
   assert(Message->IndexOf(BeginPGPSignedMessage) >= 0);
   Owner->Perform(WM_SHOWSTATUS, 0, (LPARAM)"PGP verifying...");
   AnsiString cmdline = getcommandline(GPGverifyclearsigned, GPGVerifyClearSignedCmd);

   TMemoryStream *input = new TMemoryStream();
   TMemoryStream *output = new TMemoryStream();
   TStringList *block = new TStringList();
   int x = getpgpblock(Message, block);
   block->SaveToStream(input);

   TGPGConsole4 *con = new TGPGConsole4(Owner);
   bool ok = con->Run(cmdline, NULL, input, output);
   if (ok)
   {
	  TStrings *sig = new TStringList();
	  FormatSigVerification(con->error, sig);
	  block->LoadFromStream(output);
	  TrimNulls(block);
	  block->Insert(0, "*** BEGIN PGP VERIFIED MESSAGE ***");
	  block->Insert(1, "");
	  for (int i = 0; i < sig->Count; i++)
		 block->Insert(i, sig->Strings[i]);
	  block->Add("");
	  block->Add("*** END PGP VERIFIED MESSAGE ***");
	  delete sig;
   }
   else
   {
	  Snooze(300);
	  ShowGPGError(Owner, con->error, con->status);
   }
   putpgpblock(Message, block, x);
   delete con;

   delete block;
   delete output;
   delete input;

   Owner->Perform(WM_SHOWSTATUS, 0, (LPARAM)"Done!");
   Snooze(200);

   return ok;
}
//---------------------------------------------------------------------------

/*bool __fastcall GPGVerifyClearSigned2(TForm *Owner, TStrings *Message)
{
   assert(Message->IndexOf(BeginPGPSignedMessage) >= 0);
   Owner->Perform(WM_SHOWSTATUS, 0, (LPARAM)"PGP verifying...");
   AnsiString cmdline = getcommandline(GPGverifyclearsigned, GPGVerifyClearSignedCmd);

   TMemoryStream *input = new TMemoryStream();
   TMemoryStream *output = new TMemoryStream();
   TStringList *headers = new TStringList();
   getheaders(headers, Message);
   Message->SaveToStream(input);

   TGPGConsole4 *con = new TGPGConsole4(Owner);
   bool ok = con->Run(cmdline, NULL, input, output);
   if (ok)
   {
	  FormatSigVerification(con->error);
	  Message->LoadFromStream(output);
	  headers->AddStrings(con->error);
	  headers->Add("*** BEGIN PGP VERIFIED MESSAGE ***");
	  headers->Add("");
	  putheaders(Message, headers);
	  Message->Add("*** END PGP VERIFIED MESSAGE ***");
   }
   else
   {
	  Snooze(300);
	  ShowGPGError(Owner, con->error, con->status);
	  putheaders(Message, headers);
   }
   delete con;

   delete headers;
   delete output;
   delete input;

   Owner->Perform(WM_SHOWSTATUS, 0, (LPARAM)"Done!");
   Snooze(200);

   return ok;
}*/
//---------------------------------------------------------------------------

bool __fastcall GPGDecryptMessage2(TForm *Owner, TStrings *Message,
   AnsiString &Passphrase, bool quiet, TStrings *feedback)
{
   assert(Message->IndexOf(BeginPGPMessage) >= 0);
   Owner->Perform(WM_SHOWSTATUS, 0, (LPARAM)"PGP decrypting...");
   AnsiString cmdline = getcommandline(GPGdecrypt, GPGDecryptCmd);
   if (feedback)
	  feedback->Clear();

   TMemoryStream *input = new TMemoryStream();
   TMemoryStream *output = new TMemoryStream();
   TStringList *block = new TStringList();
   int x = getpgpblock(Message, block);
   block->SaveToStream(input);

   TGPGConsole4 *con = new TGPGConsole4(Owner);
   bool ok = con->Run(cmdline, Passphrase.c_str(), input, output);
   if (ok)
   {
	  block->LoadFromStream(output);
	  TrimNulls(block);
	  ok = con->status->Text.Pos("DECRYPTION_OKAY");
	  if (ok)
	  {
		 TStringList *sig = new TStringList();
		 FormatSigVerification(con->error, sig);
		 if (sig->Count)
		 {
			block->Insert(0, "*** BEGIN PGP DECRYPTED/VERIFIED MESSAGE ***");
			block->Insert(1, "");
			for (int i = 0; i < sig->Count; i++)
			   block->Insert(i, sig->Strings[i]);
			block->Add("");
			block->Add("*** END PGP DECRYPTED/VERIFIED MESSAGE ***");
		 }
		 else
		 {
			block->Insert(0, "*** BEGIN PGP DECRYPTED MESSAGE ***");
			block->Insert(1, "");
			block->Add("");
			block->Add("*** END PGP DECRYPTED MESSAGE ***");
		 }
	  }
	  else
	  {
		 if (!quiet)
		 {
			Snooze(300);
			ShowGPGError(Owner, con->error, con->status);
		 }
	  }
   }

   if (feedback)
   {
	  feedback->AddStrings(con->error);
	  feedback->Add("-----");
	  feedback->AddStrings(con->status);
   }
   putpgpblock(Message, block, x);
   delete con;

   delete block;
   delete output;
   delete input;

   Owner->Perform(WM_SHOWSTATUS, 0, (LPARAM)"Done!");
   Snooze(200);
   return ok;
}
//---------------------------------------------------------------------------

/*bool __fastcall GPGDecryptMessage2(TForm *Owner, TStrings *Message,
   AnsiString &Passphrase, bool quiet, TStrings *feedback)
{
   assert(Message->IndexOf(BeginPGPMessage) >= 0);
   Owner->Perform(WM_SHOWSTATUS, 0, (LPARAM)"PGP decrypting...");
   AnsiString cmdline = getcommandline(GPGdecrypt, GPGDecryptCmd);

   TMemoryStream *input = new TMemoryStream();
   TMemoryStream *output = new TMemoryStream();
   TStringList *headers = new TStringList();
   getheaders(headers, Message);
   Message->SaveToStream(input);
   if (feedback)
	  feedback->Clear();

   TGPGConsole4 *con = new TGPGConsole4(Owner);
   bool ok = con->Run(cmdline, Passphrase.c_str(), input, output);
   if (ok)
   {
	  if (feedback)
	  {
		 feedback->AddStrings(con->error);
		 feedback->AddStrings(con->status);
	  }
	  ok = con->status->Text.Pos("DECRYPTION_OKAY");
	  if (ok)
	  {
		 FormatSigVerification(con->error);
		 Message->AddStrings(headers);
		 if (con->error->Count)
		 {
			headers->AddStrings(con->error);
			headers->Add("*** BEGIN PGP DECRYPTED/VERIFIED MESSAGE ***");
			headers->Add("");
		 }
		 else
		 {
			headers->Add("*** BEGIN PGP DECRYPTED MESSAGE ***");
			headers->Add("");
		 }
		 Message->LoadFromStream(output);
		 putheaders(Message, headers);
		 if (con->error->Count)
		 {
			Message->Add("*** END PGP DECRYPTED/VERIFIED MESSAGE ***");
		 }
		 else
		 {
			Message->Add("*** END PGP DECRYPTED MESSAGE ***");
		 }
	  }
	  else
	  {
		 if (!quiet)
		 {
			Snooze(300);
			ShowGPGError(Owner, con->error, con->status);
		 }
		 putheaders(Message, headers);
	  }
   }
   else
	  putheaders(Message, headers);
   delete con;

   delete headers;
   delete output;
   delete input;

   Owner->Perform(WM_SHOWSTATUS, 0, (LPARAM)"Done!");
   Snooze(200);

   return ok;
}*/
//---------------------------------------------------------------------------

void __fastcall fixasciiarmor(TStrings *pgpmsg)
{
   assert(pgpmsg->Count);
   AnsiString line;
   int x, idx = pgpmsg->Count - 1;
   while (idx && pgpmsg->Strings[idx].Pos("-----END PGP") == 0)
	  idx--;
   if (idx > 0)
   {
	  line = pgpmsg->Strings[--idx];
	  if ((x = line.Pos("=3D")) > 0)
	  {
		 line.Delete(x+1, 2);
		 pgpmsg->Strings[idx] = line;
	  }
	  line = pgpmsg->Strings[--idx];
	  if (line.Pos("=3D") > 0)
	  {
		 while ((x = line.Pos("=3D")) > 0)
			line.Delete(x+1, 2);
		 pgpmsg->Strings[idx] = line;
	  }
   }
}
//---------------------------------------------------------------------------

bool __fastcall GPGGetEncryptionKeys2(TStrings *Message, TStrings *Keys)
{
   Keys->Clear();
   fixasciiarmor(Message);
   AnsiString cmdline = getcommandline(GPGlistencryptionkeys, GPGListEncryptionKeysCmd);

   TMemoryStream *input = new TMemoryStream();
   TMemoryStream *output = new TMemoryStream();
   TStringList *headers = new TStringList();
   getheaders(headers, Message);
   Message->SaveToStream(input);

   TGPGConsole4 *con = new TGPGConsole4(NULL);
   bool ok = con->Run(cmdline, NULL, input, output);
//sm(con->error->Text + con->status->Text);
   if (ok)
   {
	  AnsiString id, line;
	  for (int i = 0; i < con->status->Count; i++)
	  {
		 line = con->status->Strings[i];
		 if (line.Pos("ENC_TO") > 0)
		 {
			id = GetTok(2, line, ' ');
			assert(id.Length() == 16);
			Keys->Add("0x" + id.SubString(id.Length() - 7, 8));
		 }
	  }
	  for (int i = 0; i < con->error->Count; i++)
	  {
		 line = con->error->Strings[i];
		 if (line.Pos("passphrase") > 0)
			Keys->Add(line.SubString(20, 64).Trim());
	  }
   }
   else
   {
	  Snooze(300);
	  ShowGPGError(Screen->ActiveForm, con->error, con->status);
   }

   putheaders(Message, headers);

   delete con;
   delete headers;
   delete output;
   delete input;

   return ok;
}

//---------------------------------------------------------------------------
//   GPGConsole4
//---------------------------------------------------------------------------

__fastcall TGPGConsole4::TGPGConsole4(TForm *owner)
   : TObject()
{
   Owner = owner;
   error = new TStringList();
   attrib = new TStringList();
   status = new TStringList();
   statusstream = new TMemoryStream();
   errorstream = new TMemoryStream();
   inputevent = new TEvent(NULL, false, false, "", false);
   log = new TStringList();
}
//---------------------------------------------------------------------------

__fastcall TGPGConsole4::~TGPGConsole4()
{
   delete log;
   delete inputevent;
   delete errorstream;
   delete statusstream;
   delete status;
   delete attrib;
   delete error;
   //DestroyHandles();
}
//---------------------------------------------------------------------------

static void __fastcall CLOSEHandle(HANDLE &handle)
{
   if (handle)
   {
	  CloseHandle(handle);
	  handle = NULL;
   }
}
//---------------------------------------------------------------------------

bool __fastcall TGPGConsole4::CreateHandles(const AnsiString &cmdLine)
{
   TSecurityAttributes sa;
   HANDLE hOutputReadTmp, hErrorReadTmp, hInputWriteTmp, hPassWriteTmp,
	  hAttrReadTmp, hStatReadTmp;

   bool bstatus = cmdLine.Pos("%stat%");
   bool battrib = cmdLine.Pos("%attr%");
   bool bpasswd = cmdLine.Pos("%pass%");

   DestroyHandles();
   sa.nLength = sizeof(sa);
   sa.lpSecurityDescriptor = NULL;
   sa.bInheritHandle = true;

   if (!CreatePipe(&hInputRead, &hInputWriteTmp, &sa, 0))
   {
	  ShowError("Pipe create error: \n");
	  return false;
   }

   if (!CreatePipe(&hOutputReadTmp, &hOutputWrite, &sa, 0))
   {
	  ShowError("Pipe create error: \n");
	  return false;
   }

   if (!CreatePipe(&hErrorReadTmp, &hErrorWrite, &sa, 0))
   {
	  ShowError("Pipe create error: \n");
	  return false;
   }

   if (bstatus && !CreatePipe(&hStatReadTmp, &hStatWrite, &sa, 0))
   {
	  ShowError("Pipe create error: \n");
	  return false;
   }

   if (battrib && !CreatePipe(&hAttrReadTmp, &hAttrWrite, &sa, 0))
   {
	  ShowError("Pipe create error: \n");
	  return false;
   }

   if (bpasswd && !CreatePipe(&hPassRead, &hPassWriteTmp, &sa, 0))
   {
	  ShowError("Pipe create error: \n");
	  return false;
   }

   if (!DuplicateHandle(GetCurrentProcess(), hInputWriteTmp,
	  GetCurrentProcess(), &hInputWrite, 0, FALSE, DUPLICATE_SAME_ACCESS))
   {
	  ShowError("Duplicate handle error: \n");
	  return false;
   }

   if (!DuplicateHandle(GetCurrentProcess(), hOutputReadTmp,
	  GetCurrentProcess(), &hOutputRead, 0, FALSE, DUPLICATE_SAME_ACCESS))
   {
	  ShowError("Duplicate handle error: \n");
	  return false;
   }

   if (!DuplicateHandle(GetCurrentProcess(), hErrorReadTmp,
	  GetCurrentProcess(), &hErrorRead, 0, FALSE, DUPLICATE_SAME_ACCESS))
   {
	  ShowError("Duplicate handle error: \n");
	  return false;
   }

   if (bstatus && !DuplicateHandle(GetCurrentProcess(), hStatReadTmp,
	  GetCurrentProcess(), &hStatRead, 0, FALSE, DUPLICATE_SAME_ACCESS))
   {
	  ShowError("Duplicate handle error: \n");
	  return false;
   }

   if (battrib && !DuplicateHandle(GetCurrentProcess(), hAttrReadTmp,
	  GetCurrentProcess(), &hAttrRead, 0, FALSE, DUPLICATE_SAME_ACCESS))
   {
	  ShowError("Duplicate handle error: \n");
	  return false;
   }

   if (bpasswd && !DuplicateHandle(GetCurrentProcess(), hPassWriteTmp,
	  GetCurrentProcess(), &hPassWrite, 0, FALSE, DUPLICATE_SAME_ACCESS))
   {
	  ShowError("Duplicate handle error: \n");
	  return false;
   }

   if (battrib) CLOSEHandle(hAttrReadTmp);
   if (bstatus) CLOSEHandle(hStatReadTmp);
   if (bpasswd) CLOSEHandle(hPassWriteTmp);
   CLOSEHandle(hOutputReadTmp);
   CLOSEHandle(hErrorReadTmp);
   CLOSEHandle(hInputWriteTmp);

   return true;
}
//---------------------------------------------------------------------------

void __fastcall TGPGConsole4::DestroyHandles()
{
   CLOSEHandle(hInputRead);
   CLOSEHandle(hOutputRead);
   CLOSEHandle(hInputWrite);
   CLOSEHandle(hOutputWrite);
   CLOSEHandle(hErrorRead);
   CLOSEHandle(hErrorWrite);
   CLOSEHandle(hPassRead);
   CLOSEHandle(hPassWrite);
   CLOSEHandle(hAttrRead);
   CLOSEHandle(hAttrWrite);
   CLOSEHandle(hStatRead);
   CLOSEHandle(hStatWrite);
}
//---------------------------------------------------------------------------

void __fastcall TGPGConsole4::SetupFileDescriptors(AnsiString &cmdline)
{
   int p;
   AnsiString fd;

   p = cmdline.Pos("%stat%");
   if (p)
   {
	  fd = AnsiString((int)hStatWrite);
	  setfield("%stat%", fd, cmdline);
   }

   p = cmdline.Pos("%attr%");
   if (p)
   {
	  fd = AnsiString((int)hAttrWrite);
	  setfield("%attr%", fd, cmdline);
   }

   p = cmdline.Pos("%pass%");
   if (p)
   {
	  fd = AnsiString((int)hPassRead);
	  setfield("%pass%", fd, cmdline);
   }
}
//---------------------------------------------------------------------------

void __fastcall TGPGConsole4::PrepareStartupInfo()
{
   ZeroMemory(&startupinfo, sizeof(STARTUPINFO));
   startupinfo.cb = sizeof(STARTUPINFO);
   startupinfo.lpReserved = NULL;
   startupinfo.lpDesktop = NULL;
   startupinfo.lpTitle = NULL;
   startupinfo.lpReserved2 = NULL;
   startupinfo.cbReserved2 = 0;
   startupinfo.dwFlags = STARTF_USESTDHANDLES;
   startupinfo.hStdInput = hInputRead;
   startupinfo.hStdOutput = hOutputWrite;
   startupinfo.hStdError = hErrorWrite;
}
//---------------------------------------------------------------------------

bool __fastcall TGPGConsole4::LaunchChildProcess(const AnsiString &cmdLine)
{
   if (!CreateProcess(
	  NULL,             // pointer to filename
	  cmdLine.c_str(),  // pointer to command line string
	  NULL, 		    // pointer to process security attributes
	  NULL, 		    // pointer to thread security attributes
	  true,  		    // handle inheritance flag
	  CREATE_NO_WINDOW, // creation flags
	  NULL, 		    // pointer to new environment block
	  NULL, 		    // pointer to current directory name
	 &startupinfo,      // pointer to STARTUPINFO
	 &processinfo       // pointer to PROCESS_INFORMATION
   ))
   {
	  ShowError("Error creating GPG process: \n");
	  return false;
   }

   return true;
}
//---------------------------------------------------------------------------

bool __fastcall TGPGConsole4::SendPassphrase(char *passphrase)
{
//sm(passphrase);
   DWORD byteswrote;
   if (!WriteFile(hPassWrite, passphrase, strlen(passphrase), &byteswrote, NULL))
   {
	  ShowError("Error sending the passphrase to GPG: \n");
	  return false;
   }

   CLOSEHandle(hPassWrite);
   return true;
}
//---------------------------------------------------------------------------

void __fastcall TGPGConsole4::Terminate()
{
   if (running)
	  TerminateProcess(processinfo.hProcess, 0);
   terminated = true;
}
//---------------------------------------------------------------------------

void __fastcall TGPGConsole4::ShowError(const AnsiString &errormsg)
{
   MsgBox(Owner, errormsg + GetLastErrorStr(), MT_ERROR, MB_OK);
}
//---------------------------------------------------------------------------

void __fastcall TGPGConsole4::Log(const AnsiString &thread, int bytes)
{
   AnsiString count = bytes == -1 ? AnsiString("eof") : AnsiString(bytes);
   log->Add(AnsiString(GetTickCount())+"  "+thread+"  "+count);
}
//---------------------------------------------------------------------------

void __fastcall TGPGConsole4::Log(const AnsiString &thread, const AnsiString &text)
{
   log->Add(AnsiString(GetTickCount())+"  "+thread+"  "+text);
}
//---------------------------------------------------------------------------

void __fastcall fixucharquotes(TStrings *error)
{
   int x;
   AnsiString err = error->Text;
   while ((x = err.Pos("\\xe2\\x80\\x9c")) > 0)
   {
	  err.Delete(x, 12);
	  err.Insert("\"", x);
   }
   while ((x = err.Pos("\\xe2\\x80\\x9d")) > 0)
   {
	  err.Delete(x, 12);
	  err.Insert("\"", x);
   }
   while ((x = err.Pos("\\xe2\\x80\\x98")) > 0)
   {
	  err.Delete(x, 12);
	  err.Insert("`", x);
   }
   while ((x = err.Pos("\\xe2\\x80\\x99")) > 0)
   {
	  err.Delete(x, 12);
	  err.Insert("'", x);
   }
   if (err != error->Text)
   {
	  error->Text = err;
   }
}
//---------------------------------------------------------------------------

bool __fastcall TGPGConsole4::Run(AnsiString cmdLine, char *Passphrase,
   TStream *Input, TStream *Output)
{
   inputstream = Input;
   outputstream = Output;

   if (!CreateHandles(cmdLine))
	  return false;

   SetupFileDescriptors(cmdLine);
   PrepareStartupInfo();

   if (!LaunchChildProcess(cmdLine))
	  return false;

   if (Passphrase && !SendPassphrase(Passphrase))
	  return false;

   running = true;
   try
   {
	  CLOSEHandle(hOutputWrite);
	  CLOSEHandle(hStatWrite);
	  CLOSEHandle(hErrorWrite);
	  if (inputstream)
	  {
		 inputthread = new TStdIOInputThread4(this);
	  }
	  outputthread = new TStdIOOutputThread4(this);
	  statusthread = new TStdIOStatusThread4(this);
	  errorthread = new TStdIOErrorThread4(this);

	  while (MsgWaitForMultipleObjects(1, &processinfo.hProcess, false,
		 INFINITE, QS_ALLINPUT) == WAIT_OBJECT_0 + 1)
			Application->ProcessMessages();
	  while (!statusthread->eof || !errorthread->eof) // || (inputthread && !inputthread->eof))
		 Application->ProcessMessages();
   }
   __finally
   {
	  if (!GetExitCodeProcess(processinfo.hProcess, &exitcode))
		 ShowError("Error getting GPG exit code: \n");

	  CloseHandle(processinfo.hThread);
	  CloseHandle(processinfo.hProcess);

	  status->LoadFromStream(statusstream);
	  error->LoadFromStream(errorstream);
	  fixucharquotes(error);
	  delete errorthread;
	  delete statusthread;
	  delete outputthread;
	  if (inputthread)
	  {
		 if (!inputthread->eof)
			TerminateThread(inputthread, 0);
		 else
		    delete inputthread;
	  }
	  DestroyHandles();
	  running = false;
	  return true;
   }
}

//---------------------------------------------------------------------------
// TStdIOInputThread
//---------------------------------------------------------------------------

__fastcall TStdIOInputThread4::TStdIOInputThread4(TGPGConsole4 *Console)
	: TThread(true)
{
   con = Console;
   Resume();
}
//---------------------------------------------------------------------------

//__fastcall TStdIOInputThread4::~TStdIOInputThread4()
//{
//   eof = true;
//}
//---------------------------------------------------------------------------

void __fastcall TStdIOInputThread4::Execute()
{
   HANDLE objects[2];
   objects[0] = (void*)con->processinfo.hProcess;
   objects[1] = (void*)con->inputevent->Handle;
   con->inputevent->SetEvent();

   if (WaitForMultipleObjects(2, objects, false, INFINITE) == WAIT_OBJECT_0 + 1)
   {
	  char buffer[bufsize];
	  DWORD bytesWritten, bytesToWrite;
	  TStream *strm = con->inputstream;

	  strm->Seek(0, soFromBeginning);
	  while(strm->Position < strm->Size)
	  {
		 bytesToWrite = strm->Read(buffer, bufsize);
		 if (!WriteFile(con->hInputWrite, buffer, bytesToWrite, &bytesWritten, NULL))
		 {
			con->ShowError("Error sending input to GPG: \n");
			break;
		 }
		 if (con->terminated)
			break;
	  }
	  FlushFileBuffers(con->hInputWrite);
	  Application->ProcessMessages();
   }

   CLOSEHandle(con->hInputWrite);
   eof = true;
}

//---------------------------------------------------------------------------
// TStdIOOutputThread
//---------------------------------------------------------------------------

__fastcall TStdIOOutputThread4::TStdIOOutputThread4(TGPGConsole4 *Console)
	: TThread(true)
{
   con = Console;
   Resume();
}
//---------------------------------------------------------------------------

void __fastcall TStdIOOutputThread4::Execute()
{
   DWORD bytesRead;
   char buffer[bufsize];

   while (!con->terminated && ReadFile(con->hOutputRead, buffer, bufsize, &bytesRead, NULL) && bytesRead > 0)
   {
	  con->outputstream->Write(&buffer[0], bytesRead);
	  //if (logop) con->Log("o", bytesRead);
   }
   con->outputstream->Position = 0;
   //CLOSEHandle(con->hOutputRead);
   //if (logop) con->Log("o", "eof");
   eof = true;
}

//---------------------------------------------------------------------------
// TStdIOStatusThread
//---------------------------------------------------------------------------

__fastcall TStdIOStatusThread4::TStdIOStatusThread4(TGPGConsole4 *Console)
	: TThread(true)
{
   con = Console;
   Resume();
}
//---------------------------------------------------------------------------

void __fastcall TStdIOStatusThread4::Execute()
{
   DWORD bytesRead;
   char buffer[bufsize];

   while (!con->terminated && ReadFile(con->hStatRead, buffer, bufsize, &bytesRead, NULL) && bytesRead > 0)
   {
	  con->statusstream->Write(&buffer[0], bytesRead);
	  //if (logop) con->Log("s", bytesRead);
   }
   con->statusstream->Position = 0;
   //CLOSEHandle(con->hStatRead);
   //if (logop) con->Log("s", "eof");
   eof = true;
}

//---------------------------------------------------------------------------
// TStdIOErrorThread
//---------------------------------------------------------------------------

__fastcall TStdIOErrorThread4::TStdIOErrorThread4(TGPGConsole4 *Console)
	: TThread(true)
{
   con = Console;
   Resume();
}
//---------------------------------------------------------------------------

void __fastcall TStdIOErrorThread4::Execute()
{
   DWORD bytesRead;
   char buffer[bufsize];

   while (!con->terminated && ReadFile(con->hErrorRead, buffer, bufsize, &bytesRead, NULL) && bytesRead > 0)
   {
	  con->errorstream->Write(&buffer[0], bytesRead);
	  //if (logop) con->Log("e", bytesRead);
   }
   con->errorstream->Position = 0;
   //CLOSEHandle(con->hErrorRead);
   //if (logop) con->Log("e", "eof");
   eof = true;
}
//---------------------------------------------------------------------------
