/*
*  inputdlgs.cpp 
*
*  Copyright (C) 2010 Richard Christman 
*
*  This file is part of QuickSilver Lite.
*
*  QuickSilver Lite is free software; you can redistribute it and/or modify 
*  it under the terms of the GNU General Public License as published by the 
*  Free Software Foundation; either version 3 of the License, or (at your 
*  option) any later version.
* 
*  This program is distributed in the hope that it will be useful, but 
*  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
*  or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License 
*  for more details.
* 
*  You should have received a copy of the GNU General Public License along 
*  with this program; if not, see <http://www.gnu.org/licenses>.
* 
*  Additional permission under GNU GPL version 3 section 7
* 
*  If you modify QuickSilver Lite, or any covered work, by linking or 
*  combining it with CodeGear VCL (or a modified version of that library), 
*  containing parts covered by the terms of CodeGear's license, the licensors 
*  of QuickSilver Lite grant you additional permission to convey the 
*  resulting work.
*
*/

//---------------------------------------------------------------------------

#include <vcl.h>
#pragma hdrstop

#include "inputdlgs.h"
#include "utility.h"
#include "global.h"
//---------------------------------------------------------------------------
#pragma package(smart_init)
#pragma resource "*.dfm"
//TInputDlg *InputDlg;
//---------------------------------------------------------------------------

AnsiString __fastcall GetUserInput(TForm *Owner, const AnsiString &DlgCaption,
   const AnsiString &Prompt, const AnsiString &Default, TStrings *existing)
{
   AnsiString input;
   TInputDlg *inputdlg = new TInputDlg(Owner, existing);
   inputdlg->Input->Text = Default;
   inputdlg->orgText = Default;
   inputdlg->Caption = DlgCaption;
   inputdlg->Prompt->Caption = Prompt;
   inputdlg->ShowModal();
   if (inputdlg->ModalResult == mrOk && inputdlg->Input->Text != "")
      input = inputdlg->Input->Text;
   delete inputdlg;
   return input;
}
//---------------------------------------------------------------------------

__fastcall TInputDlg::TInputDlg(TComponent* Owner)
		: TForm(Owner)
{
}
//---------------------------------------------------------------------------

__fastcall TInputDlg::TInputDlg(TComponent* Owner, TStrings *Existing)
		: TForm(Owner)
{
   PopupParent = (TForm*)Owner;
   existing = Existing;
   TIniFile *ini = new TIniFile(AppIni());
   AssignFont(Font, ini->ReadString("settings", "dialogfont", OptsDefDialogFont));
   AssignFont(Input->Font, ini->ReadString("settings", "windowfont",
	  OptsDefWindowFont));
   Input->Clear();
   delete ini;
}
//---------------------------------------------------------------------------

void __fastcall TInputDlg::FormCloseQuery(TObject *Sender, bool &CanClose)
{
   if (ModalResult == mrOk && existing && existing->IndexOf(Input->Text) >= 0)
   {
	  MsgBox((TForm*)Owner,
		 "The value, " + Input->Text + ", already exists!  ",
		  MT_INFO, MB_OK);
	  CanClose = false;
	  return;
   }

   if (ModalResult == mrOk && Input->Text == orgText)
   {
	  MsgBox((TForm*)Owner,
		 "You must provide a new name for your attachment,   \n"
		 "or click Cancel.",
		  MT_INFO, MB_OK);
	  CanClose = false;
   }
}
//---------------------------------------------------------------------------

void __fastcall TInputDlg::InputChange(TObject *Sender)
{
   OkBtn->Enabled = Input->Text != "" && Input->Text != orgText;
}
//---------------------------------------------------------------------------

void __fastcall TInputDlg::FormShow(TObject *Sender)
{
   OkBtn->Enabled = Input->Text != "" && Input->Text != orgText;
}
//---------------------------------------------------------------------------

