/*
*  klendlgs.cpp 
*
*  Copyright (C) 2010 Richard Christman 
*
*  This file is part of QuickSilver Lite.
*
*  QuickSilver Lite is free software; you can redistribute it and/or modify 
*  it under the terms of the GNU General Public License as published by the 
*  Free Software Foundation; either version 3 of the License, or (at your 
*  option) any later version.
* 
*  This program is distributed in the hope that it will be useful, but 
*  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
*  or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License 
*  for more details.
* 
*  You should have received a copy of the GNU General Public License along 
*  with this program; if not, see <http://www.gnu.org/licenses>.
* 
*  Additional permission under GNU GPL version 3 section 7
* 
*  If you modify QuickSilver Lite, or any covered work, by linking or 
*  combining it with CodeGear VCL (or a modified version of that library), 
*  containing parts covered by the terms of CodeGear's license, the licensors 
*  of QuickSilver Lite grant you additional permission to convey the 
*  resulting work.
*
*/

//---------------------------------------------------------------------------
#include <vcl.h>
#pragma hdrstop

#include "klendlgs.h"
#include "mixmaster.h"
#include "utility.h"
#include "help.h"
#include "global.h"
//---------------------------------------------------------------------------
#pragma package(smart_init)
#pragma resource "*.dfm"
//TKlenDlg *KlenDlg;
//---------------------------------------------------------------------------

AnsiString __fastcall GetKLenRemailer(TForm *Owner, TStrings *remailers, int MsgLen)
{
   AnsiString name;
   TKlenDlg *klendlg = new TKlenDlg(Owner);
   klendlg->AddRemailers(remailers);
   klendlg->Label2->Caption = "Your message is roughly:  "+IntToStr(MsgLen/1000)+"K";
   klendlg->ShowModal();
   if (klendlg->ModalResult == mrOk && klendlg->ListView->Selected)
      name = klendlg->ListView->Selected->Caption;
   delete klendlg;
   return name;
}
//---------------------------------------------------------------------------

bool __fastcall CheckFinalRemailerKLen(TForm *Owner, TStrings *Msg)
{
   bool ok = true;
   TMsgManager *msg = new TMsgManager(Msg);
   if (!msg->HasHeader("Chain:"))
   {
	  delete msg;
	  return true;
   }

   TStringList *klens = new TStringList();
   MixGetRemailerNames(Owner, klens);
   MixGetRemailerKLens(Owner, klens);

   int orgcount = klens->Count, msglen = msg->GuestimateFinalMessageSize();
   int smallest = 2147483647;
   for (int i = 0; i < klens->Count; i++)
   {
	  int x = (int)klens->Objects[i];
	  if (x < smallest)
		 smallest = x;
   }
   if (msglen < smallest)
   {
	  delete klens;
	  delete msg;
	  return true;
   }

   int i = 0; while (i < klens->Count)
   {
	  if ((int)klens->Objects[i] < msglen)
	  {
		 klens->Delete(i);
		 continue;
	  }
	  i++;
   }

   int x;
   AnsiString copies, name, chainhdr = msg->GetHeader("Chain:");
   TStringList *chain = new TStringList();
   chain->CommaText = GetTok(0, chainhdr, ';');
   copies = GetTok(1, chainhdr, ';');
   assert(chain->Count > 0);

   name = chain->Strings[chain->Count - 1];
   if (name != "*")
   {
	  if (klens->IndexOf(name) < 0)
	  {
		 switch (MsgBox(Owner,
			"My information says the remailer you've chosen for last \n"
			"(" + name + ") won't deliver a message as large as yours. Possibly \n"
			"you know something I don't. \n\n"
			"I can get a list of good ones. Would you like too choose a \n"
			"different remailer?",
			 MT_WARN, MB_YESNOCANCEL))
		 {
			case IDCANCEL:
			   ok = false;
			   break;
			case IDYES:
			   name = GetKLenRemailer(Owner, klens, msglen);
			   if (name.IsEmpty())
			   {
				  ok = false;
				  break;
			   }
			   msg->Lines->BeginUpdate();
			   chain->Strings[chain->Count - 1] = name;
			   x = msg->DelHeader("Chain:");
			   chainhdr = "Chain: " + chain->CommaText;
			   if (!copies.IsEmpty())
				  chainhdr += "; " + copies +";";
			   msg->Lines->Insert(x, chainhdr);
			   //Snooze(10);
			   msg->Lines->EndUpdate();
			   break;
		 }
	  }
   }
   else
   {
	  if (klens->Count < orgcount)
	  {
		 int bads = orgcount - klens->Count;
		 AnsiString remailersdont = bads > 1 ? "remailers don't" : "remailer doesn't";
		 switch(MsgBox(Owner,
			"You have a randomly selected final remailer. Unfortunately,    \n" +
			 IntToStr(bads) + " exit " + remailersdont + " deliver "
			"messages as large as yours. If \n"
			"one of those is selected, your message won't be delivered.\n\n"
			"I can get a list of good ones. Would you like to choose a \n"
            "suitable exit remailer?",
             MT_WARN, MB_YESNOCANCEL))
         {
            case IDCANCEL:
               ok = false;
               break;
            case IDYES:
               name = GetKLenRemailer(Owner, klens, msglen);
               if (name.IsEmpty())
               {
                  ok = false;
                  break;
               }
               msg->Lines->BeginUpdate();
               chain->Strings[chain->Count - 1] = name;
               x = msg->DelHeader("Chain:");
               chainhdr = "Chain: " + chain->CommaText;
               if (!copies.IsEmpty())
                  chainhdr += "; " + copies +";";
               msg->Lines->Insert(x, chainhdr);
               //Snooze(10);
               msg->Lines->EndUpdate();
               break;
         }
      }
   }

   delete chain;
   delete klens;
   delete msg;
   Snooze(200);
   return ok;
}
//---------------------------------------------------------------------------

__fastcall TKlenDlg::TKlenDlg(TComponent* Owner)
        : TForm(Owner)
{
   PopupParent = (TForm*)Owner;
   TIniFile *ini = new TIniFile(AppIni());
   LoadWindowPosition(this, "klendlg", ini);
   orgwin = IntToStr(Width)+IntToStr(Height)+
      IntToStr(ListView->Column[0]->Width)+
      IntToStr(ListView->Column[1]->Width)+
      IntToStr(ListView->Column[2]->Width)+
      IntToStr(ListView->Column[3]->Width);
   AssignFont(Font, ini->ReadString("settings", "dialogfont", OptsDefDialogFont));
   AssignFont(ListView->Font, ini->ReadString("settings", "windowfont",
      OptsDefWindowFont));
   delete ini;
   //ListView->ShowColumnHeaders = false;
}
//---------------------------------------------------------------------------

__fastcall TKlenDlg::~TKlenDlg()
{
   if (IntToStr(Width)+IntToStr(Height)+
       IntToStr(ListView->Column[0]->Width)+
       IntToStr(ListView->Column[1]->Width)+
       IntToStr(ListView->Column[2]->Width)+
       IntToStr(ListView->Column[3]->Width) != orgwin)
   {
      TIniFile *ini = new TIniFile(AppIni());
      SaveWindowPosition(this, "klendlg", ini);
      delete ini;
   }
}
//---------------------------------------------------------------------------

void __fastcall TKlenDlg::AddRemailers(TStrings *names)
{
   if (FileExists(MixMListTxtFile()))
   {
      TStringList *lines = new TStringList();
      lines->LoadFromFile(MixMListTxtFile());
      for (int i = 0; i < lines->Count; i++)
      {
         AnsiString name, line = lines->Strings[i];
         if (line.Pos("%") == line.Length())
         {
            name = GetTok(0, line, ' ');
            int x = names->IndexOf(name);
            if (x >= 0)
            {
               AnsiString numk;
               int maxlen = (int)names->Objects[x] / 1000;
               if (maxlen == 100000)
                  numk = "---- ";
               else
                  numk = IntToStr(maxlen) + "K ";
               while (numk.Length() < 7)
                  numk = " " + numk;
               line.Delete(16, 7);
               line.Insert(numk, 16);
               //RemList->Items->Add(" " + line);
               ListViewAdd(line);
            }
         }
      }
      delete lines;
   }
}
//---------------------------------------------------------------------------

void __fastcall TKlenDlg::ListViewAdd(const AnsiString &line)
{
   TListItem *item = ListView->Items->Add();
   item->Caption = line.SubString(1, 15). Trim();
   item->SubItems->Add(line.SubString(16, 12));
   item->SubItems->Add(line.SubString(28, 9).Trim());
   item->SubItems->Add(line.SubString(37, 8).Trim());
}
//---------------------------------------------------------------------------

void __fastcall TKlenDlg::Locksize1Click(TObject *Sender)
{
   //ListView->ShowColumnHeaders = false;
   BorderStyle = bsDialog;
}
//---------------------------------------------------------------------------

void __fastcall TKlenDlg::Resizedialog1Click(TObject *Sender)
{
   //ListView->ShowColumnHeaders = true;
   BorderStyle = bsSizeable;
}
//---------------------------------------------------------------------------

void __fastcall TKlenDlg::PopupMenu1Popup(TObject *Sender)
{
   Resizedialog1->Enabled = BorderStyle == bsDialog;
   Locksize1->Enabled = BorderStyle == bsSizeable;
}
//---------------------------------------------------------------------------

void __fastcall TKlenDlg::Legend1Click(TObject *Sender)
{
   DisplayHtmlHelpTopic(this, "::/html/mlistlegend.htm>legendwin");
}
//---------------------------------------------------------------------------

void __fastcall TKlenDlg::ListViewDblClick(TObject *Sender)
{
   ModalResult = mrOk;
}
//---------------------------------------------------------------------------

void __fastcall TKlenDlg::ListViewAdvancedCustomDrawSubItem(
      TCustomListView *Sender, TListItem *Item, int SubItem,
      TCustomDrawState State, TCustomDrawStage Stage, bool &DefaultDraw)
{
   if (SubItem == 1)
      ListView->Canvas->Font->Name = "courier new";
}
//---------------------------------------------------------------------------

