/*
*  mimedecs.h 
*
*  Copyright (C) 2010 Richard Christman 
*
*  This file is part of QuickSilver Lite.
*
*  QuickSilver Lite is free software; you can redistribute it and/or modify 
*  it under the terms of the GNU General Public License as published by the 
*  Free Software Foundation; either version 3 of the License, or (at your 
*  option) any later version.
* 
*  This program is distributed in the hope that it will be useful, but 
*  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
*  or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License 
*  for more details.
* 
*  You should have received a copy of the GNU General Public License along 
*  with this program; if not, see <http://www.gnu.org/licenses>.
* 
*  Additional permission under GNU GPL version 3 section 7
* 
*  If you modify QuickSilver Lite, or any covered work, by linking or 
*  combining it with CodeGear VCL (or a modified version of that library), 
*  containing parts covered by the terms of CodeGear's license, the licensors 
*  of QuickSilver Lite grant you additional permission to convey the 
*  resulting work.
*
*/

//---------------------------------------------------------------------------
#ifndef MIMEdecsH
#define MIMEdecsH
class TMIMEpartsList;
class TMIMEpart;
//---------------------------------------------------------------------------

bool __fastcall DecodeMIMEMessage(TForm *Owner, TStrings *Message,
   const AnsiString &DisposeDir, TStrings *filenames);
AnsiString __fastcall DecodeMIMEHeader(AnsiString Header);
bool __fastcall DecodeMIMEHeaders(TStrings *msg);

class TMIMEdecoder : public TObject
{
public:
   TForm *Owner;
   int status;
   AnsiString content, sender, disposedir;
   TMIMEpartsList *parts;
   TStringList * filenames;
   TStrings *msg;
   bool ok;

   __fastcall TMIMEdecoder(TForm *Owner, TStrings *Message,
      const AnsiString &DisposeDir);
   __fastcall virtual ~TMIMEdecoder();

   AnsiString __fastcall GetMIMEVersion();
   AnsiString __fastcall GetPrimeMIMEContent();
   void __fastcall GetMIMEContent(TStrings *headers);
   int __fastcall eoh();

   bool __fastcall DecodeBody();
   bool __fastcall DecodeMessage();
   void __fastcall ConstructMessage();
   TMIMEpart* __fastcall GetNextPart(int &fromline, const AnsiString &Boundary);
   TMIMEpart* __fastcall GetNextPartBody(int &fromline);
   AnsiString __fastcall GetContent(const AnsiString &field);

   bool __fastcall DecodeBase64Line(AnsiString &line, TMIMEpart *part);
   int  __fastcall DecodeBase64Line(char *line);
   bool __fastcall DecodeUUencodeLine(AnsiString &line, TMIMEpart *part);
   int  __fastcall DecodeUUencodeLine(char *line);
   bool __fastcall DecodeQuotedPrintableLine(AnsiString &line, TMIMEpart *part);
   int  __fastcall DecodeQuotedPrintableLine(char *Line);
   bool __fastcall ProcessUnknownLine(AnsiString &line, TMIMEpart *part);
   bool __fastcall Process7bitLine(AnsiString &line, TMIMEpart *part);

};

class TMIMEpart : public TObject
{
public:
   AnsiString contenttype, charset, contname, encoding,
      disposition, filename, content, disposedir;
   TStringList *headers;                        
   TMemoryStream *data;
   int status;

   __fastcall TMIMEpart(const AnsiString &DisposeDir);
   __fastcall virtual ~TMIMEpart();

   AnsiString __fastcall GetContent(const AnsiString &field);
   AnsiString __fastcall GetCompatibleFilename();
   AnsiString __fastcall Name();
   void __fastcall GetFields();
   bool __fastcall Dispose();
};

class TMIMEpartsList : public TStringList
{
public:
   __fastcall TMIMEpartsList();
   __fastcall virtual ~TMIMEpartsList();

   int __fastcall AddPart(const AnsiString &name, TMIMEpart *part);
   TMIMEpart* __fastcall GetPart(int partno);
   void __fastcall Clear();
};

const int statMimeOk = 0;

AnsiString __fastcall decodeqencoding(AnsiString &word);
//---------------------------------------------------------------------------
#endif
