/*
*  nymstuff.cpp
*                                                                                         
*  Copyright (C) 2012 Richard Christman
*
*  This file is part of QuickSilver Lite.
*
*  QuickSilver Lite is free software; you can redistribute it and/or modify
*  it under the terms of the GNU General Public License as published by the
*  Free Software Foundation; either version 3 of the License, or (at your
*  option) any later version.
*
*  This program is distributed in the hope that it will be useful, but
*  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
*  or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
*  for more details.
*
*  You should have received a copy of the GNU General Public License along
*  with this program; if not, see <http://www.gnu.org/licenses>.
*
*  Additional permission under GNU GPL version 3 section 7
*
*  If you modify QuickSilver Lite, or any covered work, by linking or
*  combining it with CodeGear VCL (or a modified version of that library),
*  containing parts covered by the terms of CodeGear's license, the licensors
*  of QuickSilver Lite grant you additional permission to convey the
*  resulting work.
*
*/

//---------------------------------------------------------------------------
#include <vcl.h>
#pragma hdrstop

#include "nymstuff.h"
#include "msgmans.h"
#include "global.h"
#include "utility.h"
#include "sigfaileddlgs.h"
#include "qsdbs.h"
#include "gpg.h"
#include "templatedlgs.h"
//---------------------------------------------------------------------------
 #pragma package(smart_init)
//---------------------------------------------------------------------------
// this function resides in composewins.cpp because it has dependencies
// bool __fastcall CheckHandleNymMessage(TStrings *message);
//---------------------------------------------------------------------------

bool __fastcall CheckFormatNymMessage(TForm *Owner, TMsgManager *orgmsg,
   TMsgManager *linmsg, bool &badtemplate)
{
   assert(orgmsg != NULL);
   assert(linmsg != NULL);

   badtemplate = false;
   TStrings *lines = linmsg->Lines;

   AnsiString msgnym, tmpnym, nymhdr = orgmsg->GetHeader("Nym:");
   if (nymhdr != "") // is this an incoming nym message
   {
	  msgnym = GetTok(0, nymhdr, ';');
	  tmpnym = linmsg->GetHeader("Nym:"); // could exist but
	  int x  = linmsg->IdxHeader("Nym:"); // will find it even if empty

	  if (x < 0)
	  {
		 if (MsgBox(Owner,
			"This template doesn't contain a Nym: header.   \n\n"
			"Continue anyway?",
			 MT_WARN, MB_OKCANCEL) == IDCANCEL)
		 {
			badtemplate = true;
			return false;
		 }
	  }

	  if (x >= 0 && tmpnym.IsEmpty())
	  {
		 if (MsgBox(Owner,
			"This template has a Nym: header, but it's empty.  \n\n"
			"Continue anyway?",
			 MT_WARN, MB_OKCANCEL) == IDCANCEL)
		 {
			badtemplate = true;
			return false;
		 }
	  }
	  if (x >= 0) lines->Delete(x);
	  if (x < 0) x = linmsg->EOH();

	  if (tmpnym != "" && tmpnym != msgnym)
	  {
		 if (MsgBox(Owner,
			"You have a template conflict!\n\n"
			"The nym in the message is  : " + msgnym + "    \n"
			"The nym in the template is : " + tmpnym + "    \n\n"
			"Continue to overwrite the nym specified in the template?",
			 MT_WARN, MB_OKCANCEL) == IDCANCEL)
		 {
			 badtemplate = true;
			 return false;
		 }
	  }

	  // save nymserver info
	  AnsiString addr = GetEAddress(nymhdr), domain = GetDomain(addr),
		 typstr = GetTok(2, nymhdr, ';').Trim(),
		 type = typstr.SubString(6, 16);
	  if (domain != "" && type != "")
		 DB->SaveNymserver(domain, type);

	  addr = GetTok(0, nymhdr, ';');
	  lines->Insert(x, "Nym: " + addr);
   }

   /*int x = lines->Count - 1; while (x > 0)
   {
	  if (lines->Strings[x].Pos("> *** END PGP DECRYPTED") == 1)
	  {
		 while (lines->Count > x + 1)
			lines->Delete(x);
		 break;
	  }
	  x--;
   }*/
   return true;
}
//---------------------------------------------------------------------------

void __fastcall ProcessNymMessage(TForm *Owner, TStrings *Message, TStrings *sig,
   const AnsiString &nymhdr)
{
   int x;
   AnsiString line, addr, domain;
   x = 0; while (x < Message->Count)
   {
	  AnsiString line = Message->Strings[x];
	  if (sig->Count && line.Trim().IsEmpty())
		 break;
	  if (line.Pos("***") == 1)
	  {
		 Message->Delete(x);
		 sig->Add(line);
		 continue;
	  }
	  x++;
   }
   while (Message->Count && Message->Strings[1].Trim() != "")
	  Message->Delete(1);
   while (Message->Count && Message->Strings[1].Trim() == "")
	  Message->Delete(1);

   CheckShowNymserverSignatureFailed(Owner, sig, nymhdr);
}
//---------------------------------------------------------------------------

/*void __fastcall WrapSig(TStrings *sig, int cols)
{
   int x;
   AnsiString line;
   TStringList *tmp = new TStringList();
   for (int i = 0; i < sig->Count; i++)
   {
	  line = sig->Strings[i];

	  x = line.Pos("\"Bananasplit Pseudonym Server (Bananasplit Pseudonymous Email Server) <send@nym.mixmin.net>");
	  if (x > 0)
	  {
		 line.Delete(x, 256);
		 line += "Nymserver <send@nym.mixmin.net>";
		 goto skipper;
	  }
	  x = line.Pos("Dizum Nymserv (Configure Nym) <config@dizum.nl>");
	  if (x > 0)
	  {
		 line.Delete(x, 256);
		 line += "Nymserver <config@dizum.nl>";
		 goto skipper;
	  }
	  x = line.Pos("Config mixnym <config@mixnym.net>");
	  if (x > 0)
	  {
		 line.Delete(x, 256);
		 line += "Nymserver <config@mixnym.net>";
		 goto skipper;
	  }
	  x = line.Pos("Nymserver <config@nymph.paranoici.org>");
	  if (x > 0)
	  {
		 line.Delete(x, 256);
		 line += "Nymserver <config@nymph.paranoici.org>";
		 goto skipper;
	  }

	  skipper:
	  if (line.Pos("*** ") == 1)
		 line = line.SubString(5, line.Length());
	  tmp->Add(line);
   }
   WrapText(tmp, cols - 4);
   sig->Clear();
   for (int i = 0; i < tmp->Count; i++)
	  sig->Add("*** " + tmp->Strings[i]);
   delete tmp;
}*/
//---------------------------------------------------------------------------

/*void __fastcall InsertNymserverVerification(TStrings *lines, TStrings *sig)
{
   if (!sig->Count)
	  return;

   TIniFile *ini = new TIniFile(AppIni());
   int cols = ini->ReadInteger("settings", "wordwraponsend", OptsDefMessageWrap);
   delete ini;

   WrapSig(sig, cols);
   assert(sig->Count);

   //lines->Insert(0, BeginPGPDecryptedVerifiedMessage);
   int i = 1; while (i < sig->Count - 1)
   {
	  AnsiString line = sig->Strings[i];
	  if (line.Pos("*** Verified ") == 1 || line.Pos("*** binary "))
		 break;

	  lines->Insert(i - 1, line);
	  i++;
   }
}*/
//---------------------------------------------------------------------------

AnsiString __fastcall GetKeyID8(const AnsiString &pubkeystr)
{
   AnsiString keyid16 = GetTok(4, pubkeystr, ':');
   return "0x"+keyid16.SubString(keyid16.Length()-7, 8);
}
//---------------------------------------------------------------------------

AnsiString __fastcall GetNymKeyID8(const AnsiString &nymhdr)
{
   AnsiString keystr, addr = GetEAddress(nymhdr).Trim();
   keystr = GPGGetPublicKeyStr(addr);
   return GetKeyID8(keystr);
}
//---------------------------------------------------------------------------

/*bool __fastcall CheckNymKeyIDs(TForm *Owner, const AnsiString &nymhdr, const AnsiString &pgphdr)
{
   bool sig, enc;
   AnsiString adr, dom, sigkey, enckey, nymkey, serkey, keystr;
   adr = GetEAddress(nymhdr);
   dom = GetDomain(adr);
   AnalysePGPHeader(pgphdr, enc, sig, enckey, sigkey);
   if (!enc)
   {
	  MsgBox(Owner,
		"Your Pgp: header does not have an encrypt key. It's requitred!   ",
		 MT_INFO, MB_OK);
	  return false;
   }
   if (!sig)
   {
	  MsgBox(Owner,
		"Your Pgp: header does not have a signing key. It's requitred!   ",
		 MT_INFO, MB_OK);
	  return false;
   }

   keystr = GPGGetPublicKeyStr("config@"+dom);
   if (keystr.IsEmpty())
	  keystr = GPGGetPublicKeyStr("send@"+dom);
   assert(keystr != "");
   serkey = GetKeyID8(keystr);

   keystr = GPGGetPublicKeyStr(adr);
   assert(keystr != NULL);
   nymkey = GetKeyID8(keystr);

 //nymkey = GetNymKeyID8(nymhdr);
 //sm(enckey + "=" +serkey + "\n" + sigkey + "=" +nymkey);
   if (enckey != serkey)
   {
	  MsgBox(Owner,
		 "Error! The encryption key in your Pgp: header is the wrong key!   \n"
		 "You'll have to fix that. The correct key is: "+serkey,
		  MT_ERROR, MB_OK);
	  return false;
   }
   if (sigkey != nymkey)
   {
	  MsgBox(Owner,
		 "Error! The signing key in your Pgp: header is the wrong key!   \n"
		 "You'll have to fix that. The correct key is: "+nymkey,
		  MT_ERROR, MB_OK);
	  return false;
   }
   return true;
}*/
//---------------------------------------------------------------------------

bool __fastcall GetNymKeyIDs(TForm *Owner, const AnsiString &nymadr, AnsiString &pgphdr,
   AnsiString &enckey, AnsiString &sigkey, AnsiString &nymkey, AnsiString &serkey)
{
   int x;
   bool enc, sig;
   AnsiString adr, dom, keystr;

   adr = GetEAddress(nymadr);
   dom = GetDomain(adr);

   keystr = GPGGetPublicKeyStr("config@"+dom);
   if (keystr == "")
   {
	  MsgBox(Owner,
		"PROBLEM: I cannot find the PGP key for: config@"+dom,
		 MT_ERROR, MB_OK
	  );
	  return false;
   }
   assert(keystr != "");
   serkey = GetKeyID8(keystr);

   keystr = GPGGetPublicKeyStr(adr);
   assert(keystr != NULL);
   nymkey = GetKeyID8(keystr);

 //nymkey = GetNymKeyID8(nymadr);
   x = pgphdr.LowerCase().Pos("sign=;");
   if (x)
   {
	  pgphdr.Insert(nymkey, x + 5);
   }
   x = pgphdr.LowerCase().Pos("encrypt=;");
   if (x)
   {
	  pgphdr.Insert(serkey, x + 8);
   }

   AnalysePGPHeader(pgphdr, enc, sig, enckey, sigkey);
   if (enckey == "" || sigkey == "")
	  return false;

   return true;
}
//---------------------------------------------------------------------------

bool __fastcall needsemptyline(TStrings *Template)
{
   bool needs = true; // in case the template is headers only.
   for (int i = 0; i < Template->Count; i++)
   {
	  if (Template->Strings[i].Trim().IsEmpty())
	  {
		 needs = false;
		 break;
	  }
   }
   return needs;
}
//---------------------------------------------------------------------------

bool __fastcall ProcessNymCfg(TForm *Owner, TStrings *Message,
   AnsiString &TemplateName, TStrings *NymCfg, bool &badtemplate)
{
   AnsiString nymhdr, hdr, thdr, ohdr, sub, mID, pststr, replyto, blank;
   NymCfg->Clear();

   // get the template and test for problems
   TStringList *Template = new TStringList();
   DB->LoadTemplate(TemplateName, Template);
   // in case the template is empty
   if (!Template->Count)
   {
	  MsgBox(Owner,
		 "Problem! The template you selected is empty! There aren't   \n"
		 "any lines of text at all! You'll need to fix that.   ",
		  MT_INFO, MB_OK);
	  delete Template;
	  return false;
   }
   // in case the template is headers only (no empty line after headers)
   // add an empty line
   if (needsemptyline(Template))
	  Template->Add("");

   TStringList *lines = new TStringList();
   TMsgManager *orgmsg = new TMsgManager(Message);
   TMsgManager *tmpmsg = new TMsgManager(Template);
   TMsgManager *linmsg = new TMsgManager(lines);

   // get special header positions from template if there
   while ((thdr = tmpmsg->PopHeader().Trim()) != "")
   {
	  int x = thdr.Pos(":");
	  assert(x > 0);
	  hdr = thdr.SubString(1, x);
	  thdr = thdr.SubString(x+1, thdr.Length()).Trim();
	  ohdr = orgmsg->GetHeader(hdr).Trim();
	  if (thdr.IsEmpty() && ohdr.IsEmpty())
		 continue;

	  if (!ohdr.IsEmpty())
		 orgmsg->DelHeader(hdr);
	  lines->Add(hdr+" " + (ohdr.IsEmpty() ? thdr : ohdr));
   }

   while ((ohdr = orgmsg->PopHeader()) != "")
   {
	  lines->Add(ohdr);
   }
   lines->Add("");

   TrimNulls(Message);
   NymCfg->AddStrings(lines);
   NymCfg->AddStrings(Message);

   delete linmsg;
   delete tmpmsg;
   delete orgmsg;
   delete lines;
   delete Template;
   return true;
}
//---------------------------------------------------------------------------

bool __fastcall NymCfgProcess(TForm *Owner, TStrings *Msg, LPSTR Message)
{
   TStringList *message = new TStringList();
   message->SetText(Message);

   AnsiString TemplateName;
   TStringList *temps = new TStringList();
   temps->CaseSensitive = false;
   DB->LoadTemplates(temps);
   if (temps->IndexOf("nymconfig") >= 0)
	  TemplateName = "nymconfig";
   delete temps;

   bool ok = true, badtemplate = true;
   while (badtemplate)
   {
	  if (TemplateName.IsEmpty() &&
		 (TemplateName = SelectTemplate(Owner)) == "")
	  {
		 delete message;
		 return false;
	  }

	  ok = ProcessNymCfg(Owner, message, TemplateName, Msg, badtemplate);
	  if (ok)
		 break;

	  TemplateName = "";
   }

   delete message;
   return ok;
}
//---------------------------------------------------------------------------
