/*
*  pgpcpassdlgs.cpp 
*
*  Copyright (C) 2010 Richard Christman 
*
*  This file is part of QuickSilver Lite.
*
*  QuickSilver Lite is free software; you can redistribute it and/or modify 
*  it under the terms of the GNU General Public License as published by the 
*  Free Software Foundation; either version 3 of the License, or (at your 
*  option) any later version.
* 
*  This program is distributed in the hope that it will be useful, but 
*  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
*  or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
*  for more details.
*
*  You should have received a copy of the GNU General Public License along
*  with this program; if not, see <http://www.gnu.org/licenses>.
*
*  Additional permission under GNU GPL version 3 section 7
*
*  If you modify QuickSilver Lite, or any covered work, by linking or
*  combining it with CodeGear VCL (or a modified version of that library),
*  containing parts covered by the terms of CodeGear's license, the licensors
*  of QuickSilver Lite grant you additional permission to convey the
*  resulting work.
*
*/

//---------------------------------------------------------------------------

#include <vcl.h>
#pragma hdrstop

#include "pgpcpassdlgs.h"
#include "utility.h"
#include "global.h"
#include "qsdbs.h"
#include "gpg.h"
//---------------------------------------------------------------------------
#pragma package(smart_init)
#pragma resource "*.dfm"
//TPGPcpassDlg *PGPcpassDlg;
//---------------------------------------------------------------------------

AnsiString __fastcall GetCEncryptionPassword(TForm *Owner)
{
   AnsiString phrase;
   TPGPcpassDlg *pgpcpassdlg = new TPGPcpassDlg(Owner);
   if (pgpcpassdlg->ShowModal() == mrOk)
   {
	  phrase = pgpcpassdlg->Phrase->Text;
   }
   delete pgpcpassdlg;
   return phrase;
}
//---------------------------------------------------------------------------

__fastcall TPGPcpassDlg::TPGPcpassDlg(TComponent* Owner)
		: TForm(Owner)
{
   PopupParent = (TForm*)Owner;
   Phrase->Clear();
   Confirm->Clear();
   if (Owner)
	  Position = poOwnerFormCenter;

   TIniFile *ini = new TIniFile(AppIni());
   AssignFont(Font, ini->ReadString("settings", "dialogfont", OptsDefDialogFont));
   AssignFont(Phrase->Font, ini->ReadString("settings", "windowfont",
      OptsDefWindowFont));
   Confirm->Font->Assign(Phrase->Font);
   delete ini;

   minpasslen = GetAdvancedConfigItem(GPGminpasslength, GPGMinPassLength);
   EditColor = Phrase->Font->Color;
   HideBoxClick(HideBox);
}
//---------------------------------------------------------------------------

__fastcall TPGPcpassDlg::~TPGPcpassDlg()
{
   WipeEditBox(Confirm);
   WipeEditBox(Phrase);
}
//---------------------------------------------------------------------------

void __fastcall TPGPcpassDlg::HideBoxClick(TObject *Sender)
{
   if (!HideBox->Checked)
   {
      Confirm->Font->Color = EditColor;
      Phrase->Font->Color = EditColor;
   }
   else
   {
      Confirm->Font->Color = Confirm->Color; //clWindow;
      Phrase->Font->Color = Phrase->Color; //clWindow;
   }
}
//---------------------------------------------------------------------------

void __fastcall TPGPcpassDlg::FormCloseQuery(TObject *Sender, bool &CanClose)
{
   if (ModalResult == mrOk)// && mode == psmEncrypt)
   {
      if (Phrase->Text.IsEmpty())
      {
         MsgBox(this,
            "You must enter a passphrase or click Cancel!   ",
             MT_INFO, MB_OK);
         CanClose = false;
         return;
      }

      if (mode == pmEncrypt && Phrase->Text.Length() < minpasslen)
      {
         MsgBox(this,
            "Your passphrase must be at least " + IntToStr(minpasslen) +
            " characters long!   ",
             MT_INFO, MB_OK);
         CanClose = false;
         return;
      }

	  if (mode == pmEncrypt && Confirm->Text != Phrase->Text)
      {
         MsgBox(this,
            "Your confirm phrase doesn't match your first phrase!   ",
             MT_INFO, MB_OK);
         CanClose = false;
         return;
      }
   }
}
//---------------------------------------------------------------------------

void __fastcall TPGPcpassDlg::PhraseChange(TObject *Sender)
{
   ProgressBar1->Position = (WORD)PGPcomdlgPassphraseQuality(Phrase->Text.c_str());
}
//---------------------------------------------------------------------------

