/*
*  pgprecipdlgs.cpp 
*
*  Copyright (C) 2010 Richard Christman 
*
*  This file is part of QuickSilver Lite.
*
*  QuickSilver Lite is free software; you can redistribute it and/or modify 
*  it under the terms of the GNU General Public License as published by the 
*  Free Software Foundation; either version 3 of the License, or (at your 
*  option) any later version.
* 
*  This program is distributed in the hope that it will be useful, but 
*  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
*  or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License 
*  for more details.
* 
*  You should have received a copy of the GNU General Public License along 
*  with this program; if not, see <http://www.gnu.org/licenses>.
* 
*  Additional permission under GNU GPL version 3 section 7
* 
*  If you modify QuickSilver Lite, or any covered work, by linking or 
*  combining it with CodeGear VCL (or a modified version of that library), 
*  containing parts covered by the terms of CodeGear's license, the licensors 
*  of QuickSilver Lite grant you additional permission to convey the 
*  resulting work.
*
*/

//---------------------------------------------------------------------------

#include <vcl.h>
#pragma hdrstop

#include "pgprecipdlgs.h"
#include "pgpcpassdlgs.h"
#include "gpg.h"
#include "utility.h"
#include "global.h"
#include "msgmans.h"
//---------------------------------------------------------------------------
#pragma package(smart_init)
#pragma resource "*.dfm"
//TPGPrecipDlg *PGPrecipDlg;
//---------------------------------------------------------------------------

bool __fastcall GetRecipientPGPPublicKeys(TForm *Owner, TStrings *msg,
   AnsiString &encryptkeys, bool &symmetric, bool sign2)
{
   TPGPrecipDlg *pgprecipdlg = new TPGPrecipDlg(Owner, msg);
   if (sign2)
      pgprecipdlg->Symmetric->Enabled = false;
   bool ok = pgprecipdlg->ShowModal() == mrOk;
   if (ok)
   {
      encryptkeys = pgprecipdlg->GetEncryptKeys();
      symmetric = pgprecipdlg->Symmetric->Checked;
   }
   delete pgprecipdlg;
   return ok;
}
//---------------------------------------------------------------------------

__fastcall TPGPrecipDlg::TPGPrecipDlg(TComponent* Owner)
        : TForm(Owner)
{
}
//---------------------------------------------------------------------------

__fastcall TPGPrecipDlg::TPGPrecipDlg(TComponent* Owner, TStrings *Msg)
        : TForm(Owner)
{
   PopupParent = (TForm*)Owner;
   RecipList->Columns->Items[0]->Width = RecipList->Width - 123;
   KeyList->Columns->Items[0]->Width = KeyList->Width - 123;

   TIniFile *ini = new TIniFile(AppIni());
   LoadWindowPosition(this, "recipdlg", KeyList, ini);
   RecipList->Columns->Items[0]->Width = KeyList->Columns->Items[0]->Width;
   orgwin = IntToStr(Width) + IntToStr(Height) +
      IntToStr(KeyList->Columns->Items[0]->Width);
   AssignFont(Font, ini->ReadString("settings", "dialogfont", OptsDefDialogFont));
   AssignFont(KeyList->Font, ini->ReadString("settings", "windowfont", OptsDefWindowFont));
   RecipList->Font->Assign(KeyList->Font);
   windowfontcolor = KeyList->Font->Color;
   delete ini;
   Primary->Width = Canvas->TextWidth(Primary->Caption) + 45;
   Primary->Left = Width - Primary->Width;

   LoadKeys();

   TMsgManager *msg = new TMsgManager(Msg);
   TStringList *recips = new TStringList();
   msg->GetFullAddresses(recips);

   bads = 0;
   orgrecips = "";
   for (int i = 0; i < recips->Count; i++)
   {
      AnsiString keyid, recip = recips->Strings[i];
      TListItem *kitem = KeyList->FindCaption(0, recip, false, true, false);
      TListItem *ritem = RecipList->Items->Add();
      ritem->Caption = recip;
      keyid = kitem ? kitem->SubItems->Strings[0] : AnsiString("");
      if (keyid.IsEmpty())
         bads++;
      ritem->SubItems->Add(keyid);
      if (kitem)
         kitem->Delete();
   }

   if (bads)
   {
      Label2->Caption = "You have " + IntToStr(bads) + " addresses I can't "
         "find PGP keys for. Double-click items in the top list to move to "
         "the lower list.";
   }

   delete recips;
   delete msg;
}
//---------------------------------------------------------------------------

__fastcall TPGPrecipDlg::~TPGPrecipDlg()
{
   if (IntToStr(Width)+ IntToStr(Height) +
      IntToStr(KeyList->Columns->Items[0]->Width) != orgwin)
   {
      TIniFile *ini = new TIniFile(AppIni());
      SaveWindowPosition(this, "recipdlg", KeyList, ini);
      delete ini;
   }
}
//---------------------------------------------------------------------------

void __fastcall TPGPrecipDlg::LoadKeys()
{
   TStringList *keys = new TStringList();

   TGPG *pgp = new TGPG();
   if (pgp->CheckInitSuccess())
	  pgp->ListPublicKeysWithColons(keys);

   KeyList->Items->BeginUpdate();
   KeyList->Items->Clear();

   AnsiString key, type, keyid, uid;
   for (int i = 0; i < keys->Count; i++)
   {
      key = keys->Strings[i];
      type = pgp->KeyType(key);

      if (pgp->KeyValid(key) == "e")
         continue;

      if (type != "pub" && type != "uid")
         continue;

      if (type != "pub" && Primary->Checked)
         continue;

      if (type == "pub")
         keyid = pgp->KeyID8(key);

      uid = pgp->KeyUID(key);
      if (GetEAddress(uid).IsEmpty())
         continue;

      TListItem *item = KeyList->Items->Add();
      item->Caption = uid;
      item->SubItems->Add("0x" + keyid);
   }

   KeyList->Items->EndUpdate();

   delete pgp;
   delete keys;
}
//---------------------------------------------------------------------------

/*AnsiString __fastcall TPGPrecipDlg::GetRecipients()
{
   AnsiString recips;
   for (int i = 0; i < RecipList->Items->Count; i++)
   {
      TListItem *item = RecipList->Items->Item[i];
      if (item->SubItems->Strings[0].IsEmpty())
         continue;

      if (recips.Length())
         recips += ",\n";

   }
}*/
//---------------------------------------------------------------------------

AnsiString __fastcall TPGPrecipDlg::GetEncryptKeys()
{
   if (Symmetric->Checked)
   {
      return symmetrickey;
   }

   AnsiString keys;
   TStringList *list = new TStringList();
   for (int i = 0; i < RecipList->Items->Count; i++)
   {
      TListItem *item = RecipList->Items->Item[i];
      if (item->SubItems->Strings[0].IsEmpty()) // no keyid
         continue;

      // save keyid
      list->Add(item->SubItems->Strings[0]);
   }
   keys = list->CommaText;
   delete list;
   return keys;
}
//---------------------------------------------------------------------------

void __fastcall TPGPrecipDlg::PrimaryClick(TObject *Sender)
{
   LoadKeys();
}
//---------------------------------------------------------------------------

void __fastcall TPGPrecipDlg::SymmetricClick(TObject *Sender)
{
   if (Symmetric->Checked)
   {
      KeyList->Items->BeginUpdate();
      KeyList->ParentColor = true;
      KeyList->Font->Color = KeyList->Color;
      KeyList->Enabled = false;
      KeyList->Items->EndUpdate();

      RecipList->Items->BeginUpdate();
      RecipList->ParentColor = true;
      RecipList->Font->Color = RecipList->Color;
      RecipList->Enabled = false;
      RecipList->Items->EndUpdate();

      Primary->Enabled = false;
      Label2->Enabled = false;
      usermsg = Label2->Caption;
      Snooze(200);
      OkBtn->Click();

   }
   else
   {
      SetRedraw(this, 0);
      KeyList->Items->BeginUpdate();
      KeyList->Color = clWindow;
      KeyList->Font->Color = windowfontcolor;
      KeyList->Enabled = true;
      KeyList->Items->EndUpdate();

      RecipList->Items->BeginUpdate();
      RecipList->Color = clWindow;
      RecipList->Font->Color = windowfontcolor;
      RecipList->Enabled = true;
      RecipList->Items->EndUpdate();
      SetRedraw(this, 1);

      Primary->Enabled = true;
      Label2->Enabled = true;
      Label2->Caption = usermsg;
   }
}
//---------------------------------------------------------------------------

void __fastcall TPGPrecipDlg::FormCloseQuery(TObject *Sender,
      bool &CanClose)
{
   if (ModalResult == mrOk)
   {
      int c = 0;
      for (int i = 0; i < RecipList->Items->Count; i++)
         if (RecipList->Items->Item[i]->SubItems->Strings[0] != "")
            c++;
      if (!Symmetric->Checked && !c)
      {
         MsgBox(this,
            "You must either select at least one valid recipient   \n"
            "or choose the Conventional Encryption option.   \n",
             MT_INFO, MB_OK);
         CanClose = false;
         return;
      }

      if (Symmetric->Checked)
      {
         symmetrickey = GetCEncryptionPassword(this);
         if (symmetrickey.IsEmpty())
         {
            Symmetric->Checked = false;
            CanClose = false;
            return;
         }
         Snooze(100);
      }
   }
}
//---------------------------------------------------------------------------

void __fastcall TPGPrecipDlg::RecipListAdvancedCustomDrawItem(
      TCustomListView *Sender, TListItem *Item, TCustomDrawState State,
      TCustomDrawStage Stage, bool &DefaultDraw)
{
   if (Item->SubItems->Strings[0].IsEmpty())
   {
      RecipList->Canvas->Font->Style = TFontStyles() << fsStrikeOut;
   }
}
//---------------------------------------------------------------------------

void __fastcall TPGPrecipDlg::KeyListDblClick(TObject *Sender)
{
   TListItem *kitem = KeyList->Selected;
   TListItem *ritem = RecipList->Items->Add();
   ritem->Caption = kitem->Caption;
   ritem->SubItems->Add(kitem->SubItems->Strings[0]);
   kitem->Delete();
}
//---------------------------------------------------------------------------

void __fastcall TPGPrecipDlg::RecipListDblClick(TObject *Sender)
{
   TListItem *ritem = RecipList->Selected;
   if (ritem->SubItems->Strings[0] != "")
   {
      TListItem *kitem = KeyList->Items->Add();
      kitem->Caption = ritem->Caption;
      kitem->SubItems->Add(ritem->SubItems->Strings[0]);
   }
   ritem->Delete();
}
//---------------------------------------------------------------------------

