/*
*  prevdlgs.cpp 
*
*  Copyright (C) 2010 Richard Christman 
*
*  This file is part of QuickSilver Lite.
*
*  QuickSilver Lite is free software; you can redistribute it and/or modify 
*  it under the terms of the GNU General Public License as published by the 
*  Free Software Foundation; either version 3 of the License, or (at your 
*  option) any later version.
* 
*  This program is distributed in the hope that it will be useful, but 
*  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
*  or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License 
*  for more details.
* 
*  You should have received a copy of the GNU General Public License along 
*  with this program; if not, see <http://www.gnu.org/licenses>.
* 
*  Additional permission under GNU GPL version 3 section 7
* 
*  If you modify QuickSilver Lite, or any covered work, by linking or 
*  combining it with CodeGear VCL (or a modified version of that library), 
*  containing parts covered by the terms of CodeGear's license, the licensors 
*  of QuickSilver Lite grant you additional permission to convey the 
*  resulting work.
*
*/

//---------------------------------------------------------------------------

#include <vcl.h>
#pragma hdrstop

#include "prevdlgs.h"
#include "utility.h"
#include "global.h"
//---------------------------------------------------------------------------
#pragma package(smart_init)
#pragma resource "*.dfm"
//TPrevDlg *PrevDlg;
//---------------------------------------------------------------------------

int __fastcall PreviewMessage(TForm *Owner, const AnsiString &Caption, TStrings *message)
{
   TPrevDlg *prevdlg = new TPrevDlg(Owner);
   prevdlg->Display->Lines->AddStrings(message);
   prevdlg->Caption = " Preview -- " + Caption;
   prevdlg->Display->SelStart = 0;
   prevdlg->Display->Perform(EM_SCROLLCARET, 0, 0);
   prevdlg->ShowModal();
   int result = prevdlg->ModalResult;
   delete prevdlg;
   return result == mrOk ? IDOK : IDCANCEL;
}
//---------------------------------------------------------------------------

__fastcall TPrevDlg::TPrevDlg(TComponent* Owner)
        : TForm(Owner)
{
   PopupParent = (TForm*)Owner;
   Display->Clear();
   TIniFile *ini = new TIniFile(AppIni());
   LoadWindowPosition(this, "prevdlg", ini);
   orgwin = IntToStr(Width)+IntToStr(Height);
   AssignFont(Font, ini->ReadString("settings", "dialogfont", OptsDefDialogFont));
   AssignFont(Display->Font, ini->ReadString("settings", "editorfont", OptsDefEditorFont));
   delete ini;
}
//---------------------------------------------------------------------------

__fastcall TPrevDlg::~TPrevDlg()
{
   if (IntToStr(Width)+IntToStr(Height) != orgwin)
   {
      TIniFile *ini = new TIniFile(AppIni());
      SaveWindowPosition(this, "prevdlg", ini);
      delete ini;
   }
}
//---------------------------------------------------------------------------

void __fastcall TPrevDlg::FormCloseQuery(TObject *Sender, bool &CanClose)
{
   if (ModalResult == mrCancel)
   {
      if (MsgBox(this, "Close and cancel the current send process?   ",
         MT_CONFIRM, MB_YESNO) == IDNO)
      {
         CanClose = false;
         return;
      }
   }
}
//---------------------------------------------------------------------------

void __fastcall TPrevDlg::FormResize(TObject *Sender)
{
   TRect Rct = Rect(5, 0, Display->ClientWidth-4, Display->ClientHeight-1);
   Display->Perform(EM_SETRECT,0,long(&Rct));
}
//---------------------------------------------------------------------------

void __fastcall TPrevDlg::PopupMenuPopup(TObject *Sender)
{
   Copy1->Enabled = Display->SelLength > 0;
   Selectall1->Enabled = Display->Text.Length() > 0;
   Resizedialog1->Enabled = BorderStyle == bsDialog;
   Locksize1->Enabled = BorderStyle == bsSizeable;
}
//---------------------------------------------------------------------------

void __fastcall TPrevDlg::Copy1Click(TObject *Sender)
{
   Display->CopyToClipboard();
}
//---------------------------------------------------------------------------

void __fastcall TPrevDlg::Resizedialog1Click(TObject *Sender)
{
   BorderStyle = bsSizeable;
}
//---------------------------------------------------------------------------

void __fastcall TPrevDlg::Locksize1Click(TObject *Sender)
{
   BorderStyle = bsDialog;
}
//---------------------------------------------------------------------------

void __fastcall TPrevDlg::Selectall1Click(TObject *Sender)
{
   Display->SelectAll();        
}
//---------------------------------------------------------------------------

