/*
*  smtpmandlgs.cpp 
*
*  Copyright (C) 2010 Richard Christman 
*
*  This file is part of QuickSilver Lite.
*
*  QuickSilver Lite is free software; you can redistribute it and/or modify 
*  it under the terms of the GNU General Public License as published by the 
*  Free Software Foundation; either version 3 of the License, or (at your 
*  option) any later version.
* 
*  This program is distributed in the hope that it will be useful, but 
*  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
*  or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License 
*  for more details.
* 
*  You should have received a copy of the GNU General Public License along 
*  with this program; if not, see <http://www.gnu.org/licenses>.
* 
*  Additional permission under GNU GPL version 3 section 7
* 
*  If you modify QuickSilver Lite, or any covered work, by linking or 
*  combining it with CodeGear VCL (or a modified version of that library), 
*  containing parts covered by the terms of CodeGear's license, the licensors 
*  of QuickSilver Lite grant you additional permission to convey the 
*  resulting work.
*
*/

//---------------------------------------------------------------------------
#include <vcl.h>
#pragma hdrstop

#include "smtpmandlgs.h"
#include "smtphostdlgs.h"
#include "global.h"
#include "utility.h"
#include "msgmans.h"
#include "qsdbs.h"
#include "help.h"
//---------------------------------------------------------------------------
#pragma package(smart_init)
#pragma resource "*.dfm"
//TSMTPManDlg *SMTPManDlg;
//---------------------------------------------------------------------------

bool __fastcall ManageSMTPServers(TForm *Owner)
{
   TSMTPManDlg *smtpdlg = new TSMTPManDlg(Owner);
   smtpdlg->ShowModal();
   bool ok = smtpdlg->ModalResult == mrOk;
   delete smtpdlg;
   return ok;
}
//---------------------------------------------------------------------------

__fastcall TSMTPManDlg::TSMTPManDlg(TComponent* Owner)
        : TForm(Owner)
{
   PopupParent = (TForm*)Owner;
   TIniFile *ini = new TIniFile(AppIni());
   LoadWindowPosition(this, "smtpmandlg", ListView, ini);
   orgwin = IntToStr(Width)+IntToStr(Height) +
      IntToStr(ListView->Columns->Items[0]->Width) +
      IntToStr(ListView->Columns->Items[1]->Width) +
      IntToStr(ListView->Columns->Items[2]->Width);
   AssignFont(Font, ini->ReadString("settings", "dialogfont", OptsDefDialogFont));
   AssignFont(ListView->Font, ini->ReadString("settings", "windowfont",
      OptsDefWindowFont));
   delete ini;


   hosts = new TStringList();
   DB->LoadSMTP(hosts);
   TStringList *cfg = new TStringList();
   for (int i = 0; i < hosts->Count; i++)
   {
      cfg->CommaText = hosts->Values[hosts->Names[i]];
      TListItem *item = ListView->Items->Add();
      item->Caption = cfg->Values["key"];
      item->SubItems->Add(cfg->Values["host"]);
      item->SubItems->Add(cfg->Values["userid"]);
      item->SubItems->Add(cfg->Values["port"]);
      item->SubItems->Add(cfg->Values["auth"]);
      item->SubItems->Add(cfg->Values["password"]);
	  item->SubItems->Add(cfg->Values["tls"]);
   }
   modified = false;
}
//---------------------------------------------------------------------------

__fastcall TSMTPManDlg::~TSMTPManDlg()
{
   if (IntToStr(Width)+IntToStr(Height)+
	   IntToStr(ListView->Column[0]->Width)+
	   IntToStr(ListView->Column[1]->Width)+
	   IntToStr(ListView->Column[2]->Width) != orgwin)
   {
	  TIniFile *ini = new TIniFile(AppIni());
	  SaveWindowPosition(this, "smtpmandlg", ListView, ini);
	  delete ini;
   }
   delete hosts;
}
//---------------------------------------------------------------------------

void __fastcall TSMTPManDlg::Resizedialog1Click(TObject *Sender)
{
   BorderStyle = bsSizeable;
}
//---------------------------------------------------------------------------

void __fastcall TSMTPManDlg::Locksize1Click(TObject *Sender)
{
   BorderStyle = bsDialog;
}
//---------------------------------------------------------------------------

void __fastcall TSMTPManDlg::PopupMenu1Popup(TObject *Sender)
{
   Edit1->Enabled = ListView->Selected;
   Delete1->Enabled = Edit1->Enabled;
   Resizedialog1->Enabled = BorderStyle == bsDialog;
   Locksize1->Enabled = BorderStyle == bsSizeable;
}
//---------------------------------------------------------------------------

void __fastcall TSMTPManDlg::Delete1Click(TObject *Sender)
{
   assert(ListView->Selected != NULL);
   if (MsgBox(this,
	  "Delete the smtp host: " + ListView->Selected->Caption + "?   ",
	  MT_CONFIRM, MB_OKCANCEL) == IDOK)
   {
	  ListView->Selected->Delete();
	  modified = true;
   }
}
//---------------------------------------------------------------------------

void __fastcall TSMTPManDlg::Add1Click(TObject *Sender)
{
   TStringList *host = new TStringList();
   TStringList *existing = new TStringList();
   for (int i = 0; i < ListView->Items->Count; i++)
	  existing->Add(ListView->Items->Item[i]->Caption);

   if (AddSMTPHost(this, host, existing))
   {
	  TListItem *item = ListView->Items->Add();
	  item->Caption = host->Values["key"];
	  item->SubItems->Add(host->Values["host"]);
	  item->SubItems->Add(host->Values["userid"]);
	  item->SubItems->Add(host->Values["port"]);
	  item->SubItems->Add(host->Values["auth"]);
	  item->SubItems->Add(host->Values["password"]);
	  item->SubItems->Add(host->Values["tls"]);
	  modified = true;
   }
   delete existing;
   delete host;
}
//---------------------------------------------------------------------------

void __fastcall TSMTPManDlg::Edit1Click(TObject *Sender)
{
   assert(ListView->Selected != NULL);
   TStringList *host = new TStringList();
   TStringList *existing = new TStringList();
   for (int i = 0; i < ListView->Items->Count; i++)
   {
	  TListItem *item = ListView->Items->Item[i];
	  if (item == ListView->Selected)
	  {
		 assert(item->SubItems->Count == 6);
		 host->Values["key"] = item->Caption;
		 host->Values["host"] = item->SubItems->Strings[0];
		 host->Values["userid"] = item->SubItems->Strings[1];
		 host->Values["port"] = item->SubItems->Strings[2];
		 host->Values["auth"] = item->SubItems->Strings[3];
		 host->Values["password"] = item->SubItems->Strings[4];
		 host->Values["tls"] = item->SubItems->Strings[5];
	  }
	  else
		 existing->Add(item->Caption);
   }
   if (EditSMTPHost(this, host, existing))
   {
	  TListItem *item = ListView->Selected;
	  item->Caption = host->Values["key"];
	  item->SubItems->Strings[0] = host->Values["host"];
	  item->SubItems->Strings[1] = host->Values["userid"];
	  item->SubItems->Strings[2] = host->Values["port"];
	  item->SubItems->Strings[3] = host->Values["auth"];
	  item->SubItems->Strings[4] = host->Values["password"];
	  item->SubItems->Strings[5] = host->Values["tls"];
	  modified = true;
   }
   delete existing;
   delete host;
}
//---------------------------------------------------------------------------

void __fastcall TSMTPManDlg::FormCloseQuery(TObject *Sender,
	  bool &CanClose)
{
   if (modified)
   {
	  if (ModalResult == mrCancel)
	  {
		 switch (MsgBox(this,
			"Your SMTP hosts info is modified. Save your changes?   ",
			 MT_CONFIRM, MB_YESNOCANCEL))
		 {
			case IDCANCEL:
			   CanClose = false;
			case IDNO:
			   return;
		 }
		 ModalResult = mrOk;
	  }

	  hosts->Clear();
	  TStringList *host = new TStringList();
	  for (int i = 0; i < ListView->Items->Count; i++)
	  {
		 host->Clear();
		 TListItem *item = ListView->Items->Item[i];
		 assert(item->SubItems->Count == 6);

		 host->Values["key"] = item->Caption;
		 host->Values["host"] = item->SubItems->Strings[0];
		 host->Values["userid"] = item->SubItems->Strings[1];
		 host->Values["port"] = item->SubItems->Strings[2];
		 host->Values["auth"] = item->SubItems->Strings[3];
		 host->Values["password"] = item->SubItems->Strings[4];
		 host->Values["tls"] = item->SubItems->Strings[5];

		 hosts->Add(item->Caption + "=" + host->CommaText);
	  }
	  DB->SaveSMTP(hosts);
   }
}
//---------------------------------------------------------------------------

void __fastcall TSMTPManDlg::HelpBtnClick(TObject *Sender)
{
   DisplayHtmlHelpTopic(this, "::/html/smtphosts.htm>smtphosts");
}
//---------------------------------------------------------------------------

void __fastcall TSMTPManDlg::ListViewClick(TObject *Sender)
{
   EditBtn->Enabled = ListView->Selected;
   DeleteBtn->Enabled = EditBtn->Enabled;
}
//---------------------------------------------------------------------------

